--------------------------------------------------------------------------------
-- CERN BE-CO-HT
-- GN4124 core for PCIe FMC carrier
-- http://www.ohwr.org/projects/gn4124-core
--------------------------------------------------------------------------------
--
-- unit name:   gn4124_axi4_pkg
--
-- description: Package for components declaration and core wide constants.
-- Spartan6 FPGAs version.
--
--------------------------------------------------------------------------------
-- This source describes Open Hardware and is licensed under the CERN-OHL-W v2.
-- You may redistribute and modify this source and make products using it
-- under the terms of the CERN-OHL-W v2 or future versions (https://ohwr.org/cern_ohl_w_v2.txt).
--
-- This source is distributed WITHOUT ANY EXPRESS OR IMPLIED
-- WARRANTY, INCLUDING OF MERCHANTABILITY, SATISFACTORY
-- QUALITY AND FITNESS FOR A PARTICULAR PURPOSE. Please see
-- the CERN-OHL-W v2 for applicable conditions.
--
--------------------------------------------------------------------------------
-- Copyright CERN 2021
--------------------------------------------------------------------------------

--! Standard library
library IEEE;
--! Standard packages
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;


--==============================================================================
--! Package declaration
--==============================================================================
package gn4124_axi4_pkg is


--==============================================================================
--! Constants declaration
--==============================================================================
  constant c_GN4124_AXI4_BURST_TYPE_INCR : std_logic_vector(1 downto 0) := "01";

--==============================================================================
--! Records declaration
--==============================================================================


  -- these records definine an "AXI4 memory slave"
  ---    see AXI4 spec: (ARM IHI 0022H   ID040120
  ---    BUT we want to be a peripheral slave as we wont support all operation types etc...
  -- hence not everything is uncommented....

  -- Covers AW and AR types; NOTE: ID is excluded as this is per Read and Write port..
  type t_gn4124_axi4_ax_chnl is record
    addr   : std_logic_vector(63 downto 0);
    len    : std_logic_vector( 7 downto 0);
    size   : std_logic_vector( 2 downto 0);
    burst  : std_logic_vector( 1 downto 0);
    cache  : std_logic_vector( 3 downto 0);
    qos    : std_logic_vector( 3 downto 0);
--    prot   : std_logic_vector( 2 downto 0);
--    lock   : std_logic;
--    region : std_logic_vector;
--    user   : std_logic_vector;  -- I can't think of a use for this, yet...
  end record t_gn4124_axi4_ax_chnl;



  type t_aw_preprocessed is record
    err         : std_logic;            -- if '1' causes data to be flushed (num words as per len)
    addr        : std_logic_vector(63 downto 2);
    addr_is_64b : std_logic;
    tc          : std_logic_vector(2 downto 0);
    nosnoop     : std_logic;
    vc          : std_logic;
    len         : unsigned(8 downto 0);   -- NOTE the max transfer size is limited due to
                                          --  AXI len being 8 bits (256 transfers in a transaction)
    fwbe        : std_logic_vector(3 downto 0);
    lwbe        : std_logic_vector(3 downto 0);
    send_b      : std_logic;                       -- if '1' causes B_channel to send a response...
    b_resp      : std_logic_vector( 1 downto 0);
  end record t_aw_preprocessed;


  type t_ar_preprocessed is record
    err         : std_logic;           -- if '1' causes data to be flushed (num words as per len)
    addr        : std_logic_vector(63 downto 2);
    addr_is_64b : std_logic;
    tc          : std_logic_vector(2 downto 0);
    nosnoop     : std_logic;
    vc          : std_logic;
    len         : unsigned(8 downto 0);   -- NOTE the max transfer size is limited due to
                                          --  AXI len being 8 bits (256 transfers in a transaction)
    fwbe        : std_logic_vector(3 downto 0);
    lwbe        : std_logic_vector(3 downto 0);  -- might be extraneous ?
    r_last      : std_logic;                     -- ithis is the 'last' bit i.e. when we have the
                                                 -- appropriate length, this then defines if it is
                                                 --  the last burst from the AR_channel operation
    r_resp      : std_logic_vector( 1 downto 0);
  end record t_ar_preprocessed;



  -- read data channel
  type t_gn4124_axi4_r_chnl is record
    data : std_logic_vector(31 downto 0);
    resp : std_logic_vector( 1 downto 0);
    last : std_logic;
--    user :   no current use for this signal..
  end record t_gn4124_axi4_r_chnl;

  -- Write data channel
  type t_gn4124_axi4_w_chnl is record
    data  : std_logic_vector(31 downto 0);
    strb  : std_logic_vector( 3 downto 0);
    last  : std_logic;
--    id    -- AXI3 only
--    user  -- cant think of a use for now!!
  end record t_gn4124_axi4_w_chnl;

  -- write response channel
  type t_gn4124_axi4_b_chnl is record
    resp : std_logic_vector( 1 downto 0);
--    user :   no use for this signal..
  end record t_gn4124_axi4_b_chnl;




  -- Function to test if the cache value for the AR channel is legal
  function f_ar_cache_is_legal(arcache : std_logic_vector(3 downto 0))
  return std_logic;

  -- Function to give us a lookup based on the AR_CACHE signal as to if we want the PCIE read
  -- command to have the NO_SNOOP bit set (i.e. can snoop from CPU cache(s)). See the package body
  -- for a further description.
  function f_ar_cache_2_nosnoop(arcache    : std_logic_vector(3 downto 0);
                                en_nosnoop : std_logic)
  return std_logic;


  -- Function to test if the cache value for the AW channel is legal
  function f_aw_cache_is_legal(awcache : std_logic_vector(3 downto 0))
  return std_logic;


  -- Function to give us a lookup based on the AW_CACHE signal as to if we want the PCIE read
  -- command to have the NO_SNOOP bit set (i.e. can snoop from CPU cache(s)). See the package body
  -- for a further description.
  function f_aw_cache_2_nosnoop(awcache    : std_logic_vector(3 downto 0);
                                en_nosnoop : std_logic)
  return std_logic;



  constant c_REC_AW_PREPROCESSED_SLV_WIDTH : positive :=
      1   -- err         : std_logic;
   + 62   -- addr        : std_logic_vector(63 downto 2);
   +  1   -- addr_is_64b : std_logic;
   +  3   -- tc          : std_logic_vector(2 downto 0);
   +  1   -- nosnoop     : std_logic;
   +  1   -- vc          : std_logic;
   +  9   -- len         : unsigned(8 downto 0);
   +  4   -- fwbe        : std_logic_vector(3 downto 0);
   +  4   -- lwbe        : std_logic_vector(3 downto 0);
   +  1   -- send_b      : std_logic;
   +  2   -- b_resp      : std_logic_vector( 1 downto 0);
  ;


  constant c_REC_AR_PREPROCESSED_SLV_WIDTH : positive :=
      1  -- err         : std_logic;
   + 62  -- addr        : std_logic_vector(63 downto 2);
   +  1  -- addr_is_64b : std_logic;
   +  3  -- tc          : std_logic_vector(2 downto 0);
   +  1  -- nosnoop     : std_logic;
   +  1  -- vc          : std_logic;
   +  9  -- len         : unsigned(8 downto 0);
   +  4  -- fwbe        : std_logic_vector(3 downto 0);
   +  4  -- --lwbe        : std_logic_vector(3 downto 0);
   +  1  -- r_last      : std_logic;
   +  2  -- r_resp      : std_logic_vector( 1 downto 0);
  ;



  -- Functions to pack records into SLV and unpack them to records again;
  function to_slv(in_rec : t_aw_preprocessed)
  return std_logic_vector;


  function to_t_aw_preprocessed(slv_in : std_logic_vector(c_REC_AW_PREPROCESSED_SLV_WIDTH -1 downto 0))
  return t_aw_preprocessed;


  function to_slv(in_rec : t_ar_preprocessed)
  return std_logic_vector;


  function to_t_ar_preprocessed(slv_in : std_logic_vector(c_REC_AR_PREPROCESSED_SLV_WIDTH -1 downto 0))
  return t_ar_preprocessed;

end gn4124_axi4_pkg;



package body gn4124_axi4_pkg is

-- Notes on truning the AXI4 cache meanings into no_snoop values and legality checks:
---  despite all the decoding below;  the best description) is in the AMBA AXI spec from ARM
--    section A4.3 (ARM ID012621 Issue H.c 26 Jan 2021  page A4-64)
--
-- A4.3 defines the changes names of the bits in AxCACHE from their names in AIX3 to be
--  AxCACHE[0]   Bufferable (B) bit
--  AxCACHE[1]   Modifiable Bit    (AXI3 was cacheable (C) bit)  <no change to functionality>
--  AxCACHE[2]   Read-Allocate (RA) Bit
--  AxCACHE[3]   Write-Allocate (WA) Bit
--
--
--  The definition of AxCACHE[3:2] is that they must not be asserted if the C bit is deasserted.
--    this forms the basis of the legality checks (as per section A4.2)
--
--  AXI4 also updates the "meaning" of Read-Allocate and Write-Allocate
--    It seems that the most appropriate behaviour is to use the following definition from A4.3.2
--      A Transaction must be looked up in cache is the value of AxCACHE[3:2] is not 0b00
--      A transaction does not need to be looked up in cache if the value of AxCAche[3:2] is 0b00
--  so let us set the PCIE-no-snoop so it is '1' when AxCAche[3:2] is 0b00 else return '0'
--  BUT only if the no-snoop behaviour is enabled, then when we read the descriptions of:
--  "Normal Non-cacheable Bufferable" and "Device bufferable" it seems as though in these cases a
--  A cache snoop may be advisable.  
--
--  It is reasonable to assume that this may need updating at a future point to behave in a more
--  similar manner to other AXI4 slave ports
  function f_ar_cache_is_legal(arcache : std_logic_vector(3 downto 0))
  return std_logic is
    variable retval : std_logic;
  begin
    case arcache is                 -- ARCACHE meaning from A4-5 table in AXI SPEC
      when "0000" => retval := '1'; -- Device non-bufferable
      when "0001" => retval := '1'; -- Device bufferable
      when "0010" => retval := '1'; -- Normal Non-cacheable Non-bufferable
      when "0011" => retval := '1'; -- Normal Non-cacheable Bufferable
      when "0100" => retval := '0';
      when "0101" => retval := '0';
      when "0110" => retval := '1'; -- Write-Through Read-Allocate  (AXI3 compliance)
      when "0111" => retval := '1'; -- Write-Back Read-Allocate     (AXI3 compliance)
      when "1000" => retval := '0';
      when "1001" => retval := '0';
      when "1010" => retval := '1'; -- Write-Through No-Allocate   && Write-Through Write-Allocate
      when "1011" => retval := '1'; -- Write-Back No-Allocate      && Write-Back Write-Allocate
      when "1100" => retval := '0';
      when "1101" => retval := '0';
      when "1110" => retval := '1'; -- Write-Through Read-Allocate  && Write-Through Read and Write-Allocate
      when "1111" => retval := '1'; -- Write-Back Read-Allocate     && Write-Back Read and Write-Allocate
      when others => retval := '0';
    end case;
    return retval;
  end function f_ar_cache_is_legal;


  function f_ar_cache_2_nosnoop(arcache    : std_logic_vector(3 downto 0);
                                en_nosnoop : std_logic)
  return std_logic is
    variable retval : std_logic;
  begin
    if en_nosnoop = '1' then
      case arcache is                 -- ARCACHE meaning from A4-5 table in AXI SPEC
        when "0000" => retval := '1'; -- Device non-bufferable
        when "0001" => retval := '0'; -- Device bufferable
        when "0010" => retval := '1'; -- Normal Non-cacheable Non-bufferable
        when "0011" => retval := '0'; -- Normal Non-cacheable Bufferable
                                           -- can get rdata from final dest or a write that is progressing there.
                                           -- HAve a look at table A4.4 for bit 0
        when "0110" => retval := '0'; -- Write-Through Read-Allocate  (AXI3 compliance)
        when "0111" => retval := '0'; -- Write-Back Read-Allocate     (AXI3 compliance)
        when "1010" => retval := '0'; -- Write-Through No-Allocate   && Write-Through Write-Allocate
        when "1011" => retval := '0'; -- Write-Back No-Allocate      && Write-Back Write-Allocate
                                       -- no allocate hints we dont want it in the read-cache therfore dont snoop!
        when "1110" => retval := '0'; -- Write-Through Read-Allocate  && Write-Through Read and Write-Allocate
        when "1111" => retval := '0'; -- Write-Back Read-Allocate     && Write-Back Read and Write-Allocate
        when others => retval := '0';
      end case;
    else
      retval := '0';
    end if;

    return retval;
  end function f_ar_cache_2_nosnoop;


  function f_aw_cache_is_legal(awcache : std_logic_vector(3 downto 0))
  return std_logic is
    variable retval : std_logic;
  begin
    case awcache is                 -- AWCACHE meaning from A4-5 table in AXI SPEC
      when "0000" => retval := '1'; -- Device non-bufferable
      when "0001" => retval := '1'; -- Device bufferable
      when "0010" => retval := '1'; -- Normal Non-cacheable Non-bufferable
      when "0011" => retval := '1'; -- Normal Non-cacheable Bufferable
      when "0100" => retval := '0';
      when "0101" => retval := '0';
      when "0110" => retval := '1'; -- Write-Through No-Allocate    && Write-Through Read-Allocate
      when "0111" => retval := '1'; -- Write-Back No-Allocate       && Write-Back Read-Allocate
      when "1000" => retval := '0';
      when "1001" => retval := '0';
      when "1010" => retval := '1'; -- Write-Through Write-Allocate (AXI3)
      when "1011" => retval := '1'; -- Write-Back Write-Allocate (AXI3)
      when "1100" => retval := '0';
      when "1101" => retval := '0';
      when "1110" => retval := '1'; -- Write-Through Write-Allocate  && Write-Through Read and Write-Allocate
      when "1111" => retval := '1'; -- Write-Back Write-Allocate     && Write-Back Read and Write-Allocate
      when others => retval := '0';
    end case;
    return retval;
  end function f_aw_cache_is_legal;



  function f_aw_cache_2_nosnoop(awcache    : std_logic_vector(3 downto 0);
                                en_nosnoop : std_logic)
  return std_logic is
    variable retval : std_logic;
  begin
    if en_nosnoop = '1' then
      case awcache is                 -- AWCACHE menaing from A4-5 table in AXI SPEC
        when "0000" => retval := '1'; -- Device non-bufferable
        when "0001" => retval := '0'; -- Device bufferable
        when "0010" => retval := '1'; -- Normal Non-cacheable Non-bufferable
        when "0011" => retval := '0'; -- Normal Non-cacheable Bufferable
        when "0110" => retval := '0'; -- Write-Through No-Allocate    && Write-Through Read-Allocate
        when "0111" => retval := '0'; -- Write-Back No-Allocate       && Write-Back Read-Allocate
        when "1010" => retval := '0'; -- Write-Through Write-Allocate (AXI3)
        when "1011" => retval := '0'; -- Write-Back Write-Allocate (AXI3)
        when "1110" => retval := '0'; -- Write-Through Write-Allocate  && Write-Through Read and Write-Allocate
        when "1111" => retval := '0'; -- Write-Back Write-Allocate     && Write-Back Read and Write-Allocate
        when others => retval := '0';
      end case;
    else
      retval := '0';
    end if;
    return retval;
  end function f_aw_cache_2_nosnoop;


  function to_slv(in_rec : t_aw_preprocessed)
  return std_logic_vector is
    variable retval : std_logic_Vector(c_REC_AW_PREPROCESSED_SLV_WIDTH-1 downto 0);
  begin
    retval(0)            := in_rec.err;
    retval(1)            := in_rec.addr_is_64b;
    retval(63 downto  2) := in_rec.addr;
    retval(66 downto 64) := in_rec.tc;
    retval(67)           := in_rec.nosnoop;
    retval(68)           := in_rec.vc;
    retval(77 downto 69) := std_logic_vector(in_rec.len);
    retval(81 downto 78) := in_rec.fwbe;
    retval(85 downto 82) := in_rec.lwbe;
    retval(86)           := in_rec.send_b;
    retval(88 downto 87) := in_rec.b_resp;
    return retval;
  end function to_slv;


  function to_t_aw_preprocessed(slv_in : std_logic_vector(c_REC_AW_PREPROCESSED_SLV_WIDTH -1 downto 0))
  return t_aw_preprocessed is
    variable retval : t_aw_preprocessed;
  begin
    retval.err         := slv_in(0);
    retval.addr_is_64b := slv_in(1);
    retval.addr        := slv_in(63 downto  2);
    retval.tc          := slv_in(66 downto 64);
    retval.nosnoop     := slv_in(67);
    retval.vc          := slv_in(68);
    retval.len         := unsigned(slv_in(77 downto 69));
    retval.fwbe        := slv_in(81 downto 78);
    retval.lwbe        := slv_in(85 downto 82);
    retval.send_b      := slv_in(86);
    retval.b_resp      := slv_in(88 downto 87);
    return retval;
  end function to_t_aw_preprocessed;


  function to_slv(in_rec : t_ar_preprocessed)
  return std_logic_vector is
    variable retval : std_logic_Vector(c_REC_AR_PREPROCESSED_SLV_WIDTH-1 downto 0);
  begin
    retval(0)            := in_rec.err;
    retval(62 downto 1)  := in_rec.addr;
    retval(63)           := in_rec.addr_is_64b;
    retval(66 downto 64) := in_rec.tc;
    retval(67)           := in_rec.nosnoop;
    retval(68)           := in_rec.vc;
    retval(77 downto 69) := std_logic_vector(in_rec.len);
    retval(81 downto 78) := in_rec.fwbe;
    retval(85 downto 82) := in_rec.lwbe;
    retval(86)           := in_rec.r_last;
    retval(88 downto 87) := in_rec.r_resp;
    return retval;
  end function to_slv;


  function to_t_ar_preprocessed(slv_in : std_logic_vector(c_REC_AR_PREPROCESSED_SLV_WIDTH -1 downto 0))
  return t_ar_preprocessed is
    variable retval : t_ar_preprocessed;
  begin
    retval.err         := slv_in(0);
    retval.addr        := slv_in(62 downto 1);
    retval.addr_is_64b := slv_in(63);
    retval.tc          := slv_in(66 downto 64);
    retval.nosnoop     := slv_in(67);
    retval.vc          := slv_in(68);
    retval.len         := unsigned(slv_in(77 downto 69));
    retval.fwbe        := slv_in(81 downto 78);
    retval.lwbe        := slv_in(85 downto 82);
    retval.r_last      := slv_in(86);
    retval.r_resp      := slv_in(88 downto 87);
    return retval;
  end function to_t_ar_preprocessed;


end gn4124_axi4_pkg;
