--------------------------------------------------------------------------------
-- GN4124 core for PCIe FMC carrier
-- http://www.ohwr.org/projects/gn4124-core
--------------------------------------------------------------------------------
--
-- unit name:   gn4124_axi_stream_dcfifo
--
-- description:  wraps the CERN generic (Dual clock) DC_FIFO to provide AXI_Stream Style 
--               signalling Additionally for the dc_fifo running in showahead mode, it is 
--               intended to mask any W -> R races and resultant metastability on the data 
--               output by controlling sampling based on empty
--
--------------------------------------------------------------------------------
--
-- This source describes Open Hardware and is licensed under the CERN-OHL-W v2.
-- You may redistribute and modify this source and make products using it
-- under the terms of the CERN-OHL-W v2 or future versions (https://ohwr.org/cern_ohl_w_v2.txt).
--
-- This source is distributed WITHOUT ANY EXPRESS OR IMPLIED
-- WARRANTY, INCLUDING OF MERCHANTABILITY, SATISFACTORY
-- QUALITY AND FITNESS FOR A PARTICULAR PURPOSE. Please see
-- the CERN-OHL-W v2 for applicable conditions.
--
--------------------------------------------------------------------------------
-- Copyright CERN 2021
--------------------------------------------------------------------------------
--
-- NOTES:
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;

use work.genram_pkg.all;

entity gn4124_axi_stream_dcfifo is
  generic (
    g_DATA_WIDTH    : positive := 32;
    g_NUM_WORDS     : positive := 8;
    g_RESET_SRC_DAT : boolean  := true
  );
  port (
      snk_clk_i     : in  std_logic;
      snk_rst_n_i   : in  std_logic;
      snk_rdy_o     : out std_logic;
      snk_vld_i     : in  std_logic;
      snk_dat_i     : in  std_logic_vector(g_DATA_WIDTH -1 downto 0);
      src_clk_i     : in  std_logic;
      src_rst_n_i   : in  std_logic;
      src_rdy_i     : in  std_logic;
      src_vld_o     : out std_logic;
      src_dat_o     : out std_logic_vector(g_DATA_WIDTH -1 downto 0)
   );
end entity gn4124_axi_stream_dcfifo;



architecture rtl of gn4124_axi_stream_dcfifo is

  signal wr_full        : std_logic;
  signal wr_write       : std_logic;

  signal rd_empty       : std_logic;
  signal src_valid_int  : std_logic;
  signal src_dat_clk_en : std_logic;
      -- TBD: Apply attributes on this to make it a clk_en for the regs
  signal rd_dat         : std_logic_vector(g_DATA_WIDTH -1 downto 0);
begin


  snk_rdy_o <= not wr_full;
  wr_write  <= snk_vld_i and not wr_full;


  inst_dcfifo : entity work.generic_async_fifo_dual_rst
  generic map(
     g_data_width             => g_DATA_WIDTH,
     g_size                   => g_NUM_WORDS,
     g_show_ahead             => true,
     g_with_rd_empty          => true,   -- with empty flag
     g_with_rd_full           => false,
     g_with_rd_almost_empty   => false,
     g_with_rd_almost_full    => false,
     g_with_rd_count          => false,
     g_with_wr_empty          => false,
     g_with_wr_full           => true,   -- with full flag
     g_with_wr_almost_empty   => false,
     g_with_wr_almost_full    => false,
     g_with_wr_count          => false,  -- with words counter
     g_almost_empty_threshold => 0,  -- threshold for almost empty flag
     g_almost_full_threshold  => 0  -- threshold for almost full flag
  ) port map (
     rst_wr_n_i               => snk_rst_n_i,
     clk_wr_i                 => snk_clk_i,
     d_i                      => snk_dat_i,
     we_i                     => wr_write,
     wr_full_o                => wr_full,
     wr_empty_o               => open,
     wr_almost_empty_o        => open,
     wr_almost_full_o         => open,
     wr_count_o               => open,
     rst_rd_n_i               => src_rst_n_i,
     clk_rd_i                 => src_clk_i,
     q_o                      => rd_dat,
     rd_i                     => src_dat_clk_en,
     rd_empty_o               => rd_empty,
     rd_full_o                => open,
     rd_almost_empty_o        => open,
     rd_almost_full_o         => open,
     rd_count_o               => open
  );


  p_dat : process (src_clk_i)
  begin
    if rising_edge(src_clk_i) then
      if src_dat_clk_en = '1' then
        src_dat_o <= rd_dat;
      end if;
      if g_RESET_SRC_DAT and src_rst_n_i = '0' then
        src_dat_o <= (others => '0');
      end if;
    end if;
  end process p_dat;


  p_valid : process (src_clk_i)
  begin
    if rising_edge(src_clk_i) then
      if src_dat_clk_en = '1' then
        src_valid_int <= '1';
      elsif src_rdy_i = '1' and src_valid_int = '1' then
        src_valid_int <= '0';
      end if;
      if src_rst_n_i = '0' then
        src_valid_int <= '0';
      end if;
    end if;
  end process p_valid;


  src_vld_o <= src_valid_int;


  p_int_clken : process (src_valid_int, src_rdy_i, rd_empty)
  begin
    if src_valid_int = '1' and src_rdy_i = '1' and rd_empty = '0' then
      src_dat_clk_en <= '1';
    elsif src_valid_int = '0' and rd_empty = '0' then
      src_dat_clk_en <= '1';
    else
      src_dat_clk_en <= '0';
    end if;
  end process p_int_clken;


end architecture rtl;