--------------------------------------------------------------------------------
-- CERN BE-CO-HT
-- GN4124 core for PCIe FMC carrier
-- http://www.ohwr.org/projects/gn4124-core
--------------------------------------------------------------------------------
--
-- unit name:   l2p_arbiter
--
-- description: Arbitrates PCIe accesses between Wishbone master,
--              L2P DMA master and P2L DMA master
--
--------------------------------------------------------------------------------
-- Copyright CERN 2010-2021
--------------------------------------------------------------------------------
-- Copyright and related rights are licensed under the Solderpad Hardware
-- License, Version 2.0 (the "License"); you may not use this file except
-- in compliance with the License. You may obtain a copy of the License at
-- http://solderpad.org/licenses/SHL-2.0.
-- Unless required by applicable law or agreed to in writing, software,
-- hardware and materials distributed under this License is distributed on an
-- "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
-- or implied. See the License for the specific language governing permissions
-- and limitations under the License.
--------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;
use work.gn4124_core_pkg.all;


entity l2p_arbiter is
  generic (
     g_NUM_PORTS : positive := 4
  );
  port (
      ---------------------------------------------------------
      -- GN4124 core clock and reset
      clk_i        : in  std_logic;
      rst_n_i      : in  std_logic;

      ---------------------------------------------------------
      -- From initiators to arbiter (arb)
      --   indexing is deliberatley from n downto 1 not 0!
      dat_i        : in  a_arbiter_rec(   g_NUM_PORTS downto 1);
      req_i        : in  std_logic_vector(g_NUM_PORTS downto 1);
      gnt_o        : out std_logic_vector(g_NUM_PORTS downto 1);

      ---------------------------------------------------------
      -- From arbiter (arb) to serializer (ser)
      arb_2_ser_o  : out t_arbiter_rec
      );
end l2p_arbiter;


architecture rtl of l2p_arbiter is


  ------------------------------------------------------------------------------
  -- Signals declaration
  ------------------------------------------------------------------------------
  signal req_valid          : std_logic_vector(g_NUM_PORTS downto 1);

  signal granted_interface  : integer range 0 to g_NUM_PORTS;
  -- encoding:  '0' is no ports active, any other value is the selected port's index


  signal padded_input_array : a_arbiter_rec(g_NUM_PORTS downto 0);
  signal selected_idat      : t_arbiter_rec;
  signal data_pipeline      : t_arbiter_rec;

  constant c_CONST_NO_VALID_REQUESTS : std_logic_vector(g_NUM_PORTS downto 1) := (others => '0');
begin



  p_req_valid : process (granted_interface, req_i)
  begin
    for i in req_valid'range loop
      if req_i(i) = '1' and (granted_interface = 0 or granted_interface = i) then
        req_valid(i) <= '1';
      else
        req_valid(i) <= '0';
      end if;
    end loop;
  end process p_req_valid;



  padded_input_array(g_NUM_PORTS downto 1) <= dat_i;
  padded_input_array(0)                    <= c_DEFAULT_ARBITER_REC;

  selected_idat <= padded_input_array(granted_interface);


  p_grant_generation : process (clk_i)
  begin
    if rising_edge(clk_i) then
      if rst_n_i = '0' then
        gnt_o             <= (others => '0');
        granted_interface <= 0;
      else
        if req_valid /= c_CONST_NO_VALID_REQUESTS then -- if there is >= 1 valid_request
          for i in g_NUM_PORTS downto 1 loop           -- scan through all ports &
            if req_valid(i) = '1' then                 -- priority encode which valid gets used.
              gnt_o             <= (others => '0');
              gnt_o(i)          <= '1';
              granted_interface <= i;
            end if;
          end loop;
        elsif selected_idat.valid = '1' and selected_idat.dframe = '0' then -- last beat of transfer
          gnt_o             <= (others => '0');
          granted_interface <= 0;
        end if;
      end if;
    end if;
  end process p_grant_generation;



  p_selected_src_register : process (clk_i)
  begin
    if rising_edge(clk_i) then
      data_pipeline  <= selected_idat;
      -- abnormal style to make sure that arb_ser_data_t does not use rst_n_i as a clk_en
      if rst_n_i = '0' then
        data_pipeline.valid  <= '0';
        data_pipeline.dframe <= '0';
      end if;
    end if;
  end process p_selected_src_register;


  p_output_pipeline : process (clk_i)
  begin
    if rising_edge(clk_i) then
      arb_2_ser_o  <= data_pipeline;
      -- as arb_ser_data_o is not reset, do not use the "else statement"; we dont want a clk_en
      if rst_n_i = '0' then
        arb_2_ser_o.valid  <= '0';
        arb_2_ser_o.dframe <= '0';
      end if;
    end if;
  end process p_output_pipeline;




--   Implement some code to show unused data trasnfer cycles while something
--    is granted (help optimise *_req_i) this is to provide a little help for system optimisation.
--    we do not want to synthesise it
--synthesis translate_off
  b_SIMONLY_unused_granted_bus_cycles : block     -- use a block so the signal are invisible and self-contained
    signal gnt_was_unused       : std_logic_vector(g_NUM_PORTS downto 1);
  begin
    p_optimisation_helper : process (clk_i)
    begin
      if rising_edge(clk_i) then
        for i in gnt_was_unused'range loop
          if i = granted_interface then
            gnt_was_unused(i) <= not(padded_input_array(i).dframe or padded_input_array(i).valid);
          else
            gnt_was_unused(i) <= '0';
          end if;
        end loop;
      end if;
    end process p_optimisation_helper;
  end block b_SIMONLY_unused_granted_bus_cycles;

    -- TBD: assert to make sure only a single bit in  req_valid is asserted if granted_interface/= 0
--synthesis translate_on


end rtl;
