--------------------------------------------------------------------------------
-- GN4124 core for PCIe FMC carrier
-- http://www.ohwr.org/projects/gn4124-core
--------------------------------------------------------------------------------
--
-- unit name:   l2p_axi4_wr_dc
--
-- description: Wrapper for the modules that make the AXI4 write interface for the gennum phy
--               This component exposes and AXI4 Peripheral Slave port and connects into the
--               existing IP core L2P infrastructure to master the PCIE bus.
--
--------------------------------------------------------------------------------
--
-- This source describes Open Hardware and is licensed under the CERN-OHL-W v2.
-- You may redistribute and modify this source and make products using it
-- under the terms of the CERN-OHL-W v2 or future versions (https://ohwr.org/cern_ohl_w_v2.txt).
--
-- This source is distributed WITHOUT ANY EXPRESS OR IMPLIED
-- WARRANTY, INCLUDING OF MERCHANTABILITY, SATISFACTORY
-- QUALITY AND FITNESS FOR A PARTICULAR PURPOSE. Please see
-- the CERN-OHL-W v2 for applicable conditions.
--
--------------------------------------------------------------------------------
-- Copyright CERN 2021
--------------------------------------------------------------------------------
--
-- NOTES: Implemented as a wrapper to make deploying this as a sync or async block simpler...
--             in the async case, doing the preprocesing in a faster clk domain will reduce latency
--             to initially send the first bytes of the PCIE packet.
--
-- NOT fully COMPLIANT with the AXI4 SPEC as a memory slave.....
--       Hence this component exposes an AXI4 peripheral slave.
--             a/ Incomplete support for AXI4 burst Types
--                    FIXED   NO    -- I dont see a PC needing a FIFO style write target
--                    INCR    YES
--                    WRAP    NO    -- this interface is not intended for CPU cache line population
--                  i/ I dont expect they will be needed
--                  ii/ they could be implemented by splitting a burst into multiple writes
--
--             b/ cache -- this should probably get mapped to the snoop control bit in the packet
--                   header. The AXI signal is present but may want "refinement"
--                   Because of the potential for MODIFIABLE vs NON_Modifiable transactions with the
--                   use of strobes etc, it seems sensible that this core Treats everything as
--                   non-modificable, and thus AXI masters / initiators are responsible for PCIE
--                   transaction size limits.  (Possible future enhancement)
--
--             c/ SIZE -- will error for anything other than 3'h2 (32bit)
--             b/ write_strobes : this needs to map to PCIE, so they should be contiguously asserted
--             e/ PCIE TC is currently hardccoded but QOS is plumbed in to allow extending the code
--                 to make QOS -> TC programmable
--             f/ WRITE ACKNOWLEDGEMENTS ARE SENT UPON POSTING A PCIE TLP TO THE GENNUM PHY NOT
--                COMPLETION FROM THE HOST SYSTEM
--------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;

use work.gn4124_axi4_pkg.all;
use work.genram_pkg.all;


entity l2p_axi4_wr_dc is
  generic (
    g_IMPLEMENT_AXI          : boolean                      := false;
    g_ID_WIDTH               : positive                     := 2;
    g_DATA_BYTE_WIDTH        : positive                     := 4;
    g_AXI_W_FIFO_WRDS        : positive                     := 512;
    g_AXI_AW_FIFO_WRDS       : positive                     := 16;
    g_AXI_B_FIFO_WRDS        : positive                     := 8;
    g_EN_AXI_QOS_2_TC_LOOKUP : boolean                      := false;
    g_PCIE_DEAGULT_TC        : std_logic_vector(2 downto 0) := "000"
  );
  port (
    ---------------------------------------------------------
    -- GN4124 core clock and reset
    clk_i                    : in  std_logic;
    rst_n_i                  : in  std_logic;  -- assume synchronous to clk

    ---------------------------------------------------------
    -- To the P2L Interface (send the DMA Master Read request)  arm = Axi Read Master
    ---------------------------------------------------------
    -- req and gnt to get access from the arbitrator
    awm_arb_req_o            : out std_logic;
    arb_awm_gnt_i            : in  std_logic;
    -- data pipe to push data from this Axi Read Master to the arbitrator..
    awm_arb_valid_o          : out std_logic;  -- Read completion signals
    awm_arb_dframe_o         : out std_logic;  -- Toward the arbiter
    awm_arb_data_o           : out std_logic_vector(31 downto 0);

    -- L2P channel control
    l2p_edb_o                : out std_logic;  -- Asserted when transfer is aborted
    l_wr_rdy_i               : in  std_logic_vector(1 downto 0);
                                        -- Asserted when GN4124 is ready to receive master write
    l2p_rdy_i                : in  std_logic; -- De-asserted to pause transfer
    tx_error_i               : in  std_logic; -- Asserted when unexpected or malformed paket received

    ---------------------------------------------------------
    -- AXI4 slave interface (WRITE):   3 channels AW, W and B
    ---------------------------------------------------------

    clk_axi_i                : in  std_logic;
    rst_axi_n_i              : in  std_logic;

    aw_ready_o               : out std_logic;
    aw_valid_i               : in  std_logic;
    aw_chnl_id_i             : in  std_logic_vector(g_ID_WIDTH-1 downto 0);
    aw_chnl_addr_i           : in  std_logic_vector(63 downto 0);
    aw_chnl_len_i            : in  std_logic_vector(7 downto 0);
    aw_chnl_size_i           : in  std_logic_vector(2 downto 0);
    aw_chnl_burst_i          : in  std_logic_vector(1 downto 0);
    aw_chnl_cache_i          : in  std_logic_vector(3 downto 0);
    aw_chnl_qos_i            : in  std_logic_vector(3 downto 0);

    w_ready_o                : out std_logic;
    w_valid_i                : in  std_logic;
    w_chnl_data_i            : in  std_logic_vector(g_DATA_BYTE_WIDTH*8 -1 downto 0);
    w_chnl_strobe_i          : in  std_logic_vector(g_DATA_BYTE_WIDTH   -1 downto 0);
    w_chnl_last_i            : in  std_logic;

    b_ready_i                : in  std_logic;
    b_valid_o                : out std_logic;
    b_chnl_resp_o            : out std_logic_vector(1 downto 0);
    b_chnl_id_o              : out std_logic_vector(g_ID_WIDTH-1 downto 0);

    ---------------------------------------------------------
    -- AXI4 preprocessor control / status signals
    ---------------------------------------------------------
    preproc_seen_err_o       : out std_logic;
    preproc_err_code_o       : out std_logic_vector(3 downto 0);

    preproc_en_nosnoop_i     : in  std_logic;
    preproc_en_mods_i        : in  std_logic;
    preproc_mod_max_sz_i     : in  unsigned(7 downto 0)
  );
end l2p_axi4_wr_dc;


architecture struct of l2p_axi4_wr_dc is

-- All singals are in the relevant block/generate: then we have no warnings for unconnected/unused singals.

begin


  gen_without_axi_wr : if not g_IMPLEMENT_AXI generate
    awm_arb_req_o      <= '0';
    awm_arb_valid_o    <= '0';
    awm_arb_dframe_o   <= '0';
    awm_arb_data_o     <= X"00000000";
    l2p_edb_o          <= '0';
    aw_ready_o         <= '0';
    w_ready_o          <= '0';
    b_valid_o          <= '0';
    preproc_seen_err_o <= '0';
    preproc_err_code_o <= "0000";
    b_chnl_resp_o      <= (others => '0');
    b_chnl_id_o        <= (others => '0');
  end generate gen_without_axi_wr;

  gen_axi_wr_subsystem : if g_IMPLEMENT_AXI generate
    signal aw_rec               : t_gn4124_axi4_ax_chnl;

    signal pp_wo_rdy            : std_logic;
    signal pp_wo_vld            : std_logic;
    signal pp_wo_dat            : std_logic_vector(g_DATA_BYTE_WIDTH*8 downto 0);

    signal wdat_fifo_full       : std_logic;
    signal wdat_fifo_write      : std_logic;

    signal pp_ctrl_rdy          : std_logic;
    signal pp_ctrl_vld          : std_logic;
    signal pp_ctrl_rec          : t_aw_preprocessed;
    signal pp_ctrl_rec_id       : std_logic_vector(g_ID_WIDTH-1 downto 0);
    signal pp_ctrl_rec_slv      : std_logic_vector(   g_ID_WIDTH
                                                   + c_REC_AW_PREPROCESSED_SLV_WIDTH-1 downto 0);

    signal cmd_fifo_full        : std_logic;
    signal cmd_fifo_write       : std_logic;

    signal wdat_fifo_dout       : std_logic_vector(32 downto 0);
    signal wdat_fifo_empty      : std_logic;
    signal wdat_fifo_rd         : std_logic;

    signal cmd_fifo_dout_slv    : std_logic_vector(   g_ID_WIDTH
                                                   + c_REC_AW_PREPROCESSED_SLV_WIDTH-1 downto 0);
    signal cmd_fifo_dout_rec    : t_aw_preprocessed;
    signal cmd_fifo_dout_rec_id : std_logic_vector(g_ID_WIDTH-1 downto 0);
    signal cmd_fifo_empty       : std_logic;
    signal cmd_fifo_rd          : std_logic;

    -- axi gennum convertor signals
    signal agc_aw_rdy           : std_logic;
    signal agc_aw_vld           : std_logic;
    signal agc_w_rdy            : std_logic;
    signal agc_w_vld            : std_logic;

    signal int_b_rdy            : std_logic;
    signal int_b_vld            : std_logic;
    signal int_b_resp           : std_logic_vector(1 downto 0);
    signal int_b_id             : std_logic_vector(g_ID_WIDTH-1 downto 0);

    signal b_fifo_wdat  : std_logic_Vector(g_ID_WIDTH+1 downto 0);
    signal b_fifo_rdat  : std_logic_Vector(g_ID_WIDTH+1 downto 0);

  begin

    aw_rec <= (addr  => aw_chnl_addr_i,
               len   => aw_chnl_len_i,
               size  => aw_chnl_size_i,
               burst => aw_chnl_burst_i,
               cache => aw_chnl_cache_i,
               qos   => aw_chnl_qos_i
              );


    preproc_inst : entity work.l2p_axi4_wr_preprocessor
    generic map (
      g_EN_AXI_QOS_2_TC_LOOKUP  => g_EN_AXI_QOS_2_TC_LOOKUP,
      G_PCIE_DEAGULT_QOS        => "000",
      g_DATA_BYTE_WIDTH         => g_DATA_BYTE_WIDTH,
      g_ID_WIDTH                => g_ID_WIDTH,
      g_EN_DEBUG_ASSERTS        => false
    ) port map (
      clk_i                     => clk_axi_i,
      rst_n_i                   => rst_axi_n_i,
      seen_err_o                => preproc_seen_err_o,
      err_code_o                => preproc_err_code_o,
      enable_no_snoop_i         => preproc_en_nosnoop_i,
      vc_i                      => '0',
      enable_modifications_i    => preproc_en_mods_i,
      -- modified_write_max_size   => preproc_mod_max_sz,
      aw_chnl_ready_o           => aw_ready_o,
      aw_chnl_valid_i           => aw_valid_i,
      aw_chnl_i                 => aw_rec,
      aw_chnl_id_i              => aw_chnl_id_i,
      w_chnl_ready_o            => w_ready_o,
      w_chnl_valid_i            => w_valid_i,
      w_chnl_data_i             => w_chnl_data_i,
      w_chnl_strb_i             => w_chnl_strobe_i,
      w_chnl_last_i             => w_chnl_last_i,
      wo_rdy_i                  => pp_wo_rdy,
      wo_vld_o                  => pp_wo_vld,
      wo_dat_o                  => pp_wo_dat(31 downto 0),
      wo_dat_eop_o              => pp_wo_dat(32),
      ctrl_rdy_i                => pp_ctrl_rdy,
      ctrl_vld_o                => pp_ctrl_vld,
      ctrl_rec_o                => pp_ctrl_rec,
      ctrl_rec_id_o             => pp_ctrl_rec_id
    );


    -- Data CDC fifo; takes raw data and a LAST flag
    dat_fifo : entity work.gn4124_axi_stream_dcfifo
    generic map(
      g_DATA_WIDTH => 1 + g_DATA_BYTE_WIDTH*8,
      g_NUM_WORDS  => g_AXI_W_FIFO_WRDS
    ) port map (
      snk_clk_i    => clk_axi_i,
      snk_rst_n_i  => rst_axi_n_i,
      snk_rdy_o    => pp_wo_rdy,
      snk_vld_i    => pp_wo_vld,
      snk_dat_i    => pp_wo_dat,
      src_clk_i    => clk_i,
      src_rst_n_i  => rst_n_i,
      src_rdy_i    => agc_w_rdy,
      src_vld_o    => agc_w_vld,
      src_dat_o    => wdat_fifo_dout
    );


    --pack CMD_FIFO control signals   (Note ID is separate to the record: parameterisation)
    pp_ctrl_rec_slv <= pp_ctrl_rec_id & to_slv(pp_ctrl_rec);


    -- this assertion is to ensure that the packing and unpacking functions "match",
    -- without having to debug on either side of a fifo...
    --synthesis translate_off
    proc_rec_pack_unpack_checking : process (clk_axi_i)
      variable v_decoded_from_slv : t_aw_preprocessed;
    begin
      if rising_edge(clk_axi_i) then
        v_decoded_from_slv := to_t_aw_preprocessed(pp_ctrl_rec_slv(c_REC_AW_PREPROCESSED_SLV_WIDTH-1 downto 0));
        assert v_decoded_from_slv = pp_ctrl_rec
          report "RECORD pack and unpack error" severity ERROR;
      end if;
    end process proc_rec_pack_unpack_checking;
    --synthesis translate_on


    -- Command dc_Fifo
    cmd_dcfifo : entity work.gn4124_axi_stream_dcfifo
    generic map(
       g_DATA_WIDTH => c_REC_AW_PREPROCESSED_SLV_WIDTH + g_ID_WIDTH,
       g_NUM_WORDS  => g_AXI_AW_FIFO_WRDS
    ) port map (
       snk_clk_i    => clk_axi_i,
       snk_rst_n_i  => rst_axi_n_i,
       snk_rdy_o    => pp_ctrl_rdy,
       snk_vld_i    => pp_ctrl_vld,
       snk_dat_i    => pp_ctrl_rec_slv,
       src_clk_i    => clk_i,
       src_rst_n_i  => rst_n_i,
       src_rdy_i    => agc_aw_rdy,
       src_vld_o    => agc_aw_vld,
       src_dat_o    => cmd_fifo_dout_slv
    );


    -- unpack command dcfifo outputs
    cmd_fifo_dout_rec    <= to_t_aw_preprocessed(cmd_fifo_dout_slv(c_REC_AW_PREPROCESSED_SLV_WIDTH-1 downto 0));
    cmd_fifo_dout_rec_id <= cmd_fifo_dout_slv(cmd_fifo_dout_slv'high downto c_REC_AW_PREPROCESSED_SLV_WIDTH);


    gennum_master_axi_wr_init : entity work.l2p_axi4_wr_initiator
    generic map (
      g_ID_WIDTH           => g_ID_WIDTH,
      g_DATA_BYTE_WIDTH    => g_DATA_BYTE_WIDTH
    ) port map (
      clk_i                  => clk_i,
      rst_n_i                => rst_n_i,
      --enable_bus_mastering => '1',
      seen_err_o             => open,
      err_code_o             => open,
      --wr_ipg               => "0000",
      aw_ready_o             => agc_aw_rdy,
      aw_valid_i             => agc_aw_vld,
      aw_rec_i               => cmd_fifo_dout_rec,
      aw_rec_id_i            => cmd_fifo_dout_rec_id,
      w_ready_o              => agc_w_rdy,
      w_valid_i              => agc_w_vld,
      w_dat_i                => wdat_fifo_dout(g_DATA_BYTE_WIDTH *8 -1 downto 0),
      w_last_i               => wdat_fifo_dout(g_DATA_BYTE_WIDTH *8),
      b_chnl_ready_i         => int_b_rdy,
      b_chnl_valid_o         => int_b_vld,
      b_chnl_id_o            => int_b_id,
      b_chnl_resp_o          => int_b_resp,
      arb_axw_gnt_i          => arb_awm_gnt_i,
      axw_arb_req_o          => awm_arb_req_o,
      axw_arb_valid_o        => awm_arb_valid_o,
      axw_arb_dframe_o       => awm_arb_dframe_o,
      axw_arb_data_o         => awm_arb_data_o,
      l2p_edb_o              => l2p_edb_o,
      l_wr_rdy_i             => l_wr_rdy_i,
      l2p_rdy_i              => l2p_rdy_i,
      tx_error_i             => tx_error_i
    );


       -- Pack  SLV for the fifo
    b_fifo_wdat   <= int_b_id & int_b_resp;

    b_chl_dcfifo : entity work.gn4124_axi_stream_dcfifo
    generic map(
      g_DATA_WIDTH    => 2+g_ID_WIDTH,
      g_NUM_WORDS     => 8,
      g_RESET_SRC_DAT => TRUE
    ) port map (
      snk_clk_i       => clk_i,
      snk_rst_n_i     => rst_n_i,
      snk_rdy_o       => int_b_rdy,
      snk_vld_i       => int_b_vld,
      snk_dat_i       => b_fifo_wdat,
      src_clk_i       => clk_axi_i,
      src_rst_n_i     => rst_axi_n_i,
      src_rdy_i       => b_ready_i,
      src_vld_o       => b_valid_o,
      src_dat_o       => b_fifo_rdat
    );

    --unpack Bresp signals
    b_chnl_resp_o <= b_fifo_rdat(1 downto 0);
    b_chnl_id_o   <= b_fifo_rdat(b_fifo_rdat'high downto 2);

  end generate gen_axi_wr_subsystem;

end architecture struct;
