--------------------------------------------------------------------------------
-- GN4124 core for PCIe FMC carrier
-- http://www.ohwr.org/projects/gn4124-core
--------------------------------------------------------------------------------
--
-- unit name:   p2l_axi4_wr_initiator
--
-- description: converts a local represnetation of the axi command to the format the gennum phy uses
--              it then generates the B_CHANNEL "write completion" transfers for AXI4
--
--------------------------------------------------------------------------------
--
-- This source describes Open Hardware and is licensed under the CERN-OHL-W v2.
-- You may redistribute and modify this source and make products using it
-- under the terms of the CERN-OHL-W v2 or future versions (https://ohwr.org/cern_ohl_w_v2.txt).
--
-- This source is distributed WITHOUT ANY EXPRESS OR IMPLIED
-- WARRANTY, INCLUDING OF MERCHANTABILITY, SATISFACTORY
-- QUALITY AND FITNESS FOR A PARTICULAR PURPOSE. Please see
-- the CERN-OHL-W v2 for applicable conditions.
--------------------------------------------------------------------------------
-- Copyright CERN 2021
--------------------------------------------------------------------------------
--
-- NOTES:
--       Assumption that the preprocessing module has seen and processed all write data before
--       issuing the command, thus we should never underrun the write data..
--
--
--------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;

use work.gn4124_core_pkg.all;
use work.gn4124_axi4_pkg.all;


entity l2p_axi4_wr_initiator is
  generic (
    --g_USE_AW_REGISTER_PL : boolean  := false;
    g_ID_WIDTH         : positive := 2;
    g_DATA_BYTE_WIDTH  : positive := 4
  );
  port (
    ---------------------------------------------------------
    -- GN4124 core clock and reset
    clk_i              : in  std_logic;
    rst_n_i            : in  std_logic;

    seen_err_o         : out std_logic;
    err_code_o         : out std_logic_vector(3 downto 0);

    ---------------------------------------------------------
    -- AXI4 write channels:
    ---------------------------------------------------------
    -- Write address channel
    aw_ready_o         : out std_logic;
    aw_valid_i         : in  std_logic;
    aw_rec_i           : in  t_aw_preprocessed;
    aw_rec_id_i        : in  std_logic_Vector(g_ID_WIDTH-1 downto 0);

    -- write address channel
    w_ready_o          : out std_logic;
    w_valid_i          : in  std_logic;
    w_dat_i            : in  std_logic_vector(g_DATA_BYTE_WIDTH *8 -1 downto 0);
    w_last_i           : in  std_logic;

    -- write response channel
    b_chnl_ready_i     : in  std_logic;
    b_chnl_valid_o     : out std_logic;
    b_chnl_resp_o      : out std_logic_vector(1 downto 0);
    b_chnl_id_o        : out std_logic_Vector(g_ID_WIDTH-1 downto 0);

    ---------------------------------------------------------
    -- To the P2L Interface (send the DMA Master Read request)  arm = Axi Read Master
    ---------------------------------------------------------
    -- req and gnt to get access from the arbitrator
    arb_axw_gnt_i      : in  std_logic;
    axw_arb_req_o      : out std_logic;
    axw_arb_valid_o    : out std_logic;  -- Read completion signals
    axw_arb_dframe_o   : out std_logic;  -- Toward the arbiter
    axw_arb_data_o     : out std_logic_vector(31 downto 0);

    -- L2P channel control
    l2p_edb_o          : out std_logic;  -- Asserted when transfer is aborted
    l_wr_rdy_i         : in  std_logic_vector(1 downto 0); -- GN4124 is ready for master wr_data
    l2p_rdy_i          : in  std_logic;  -- De-asserted to pause transfer already in progress
    tx_error_i         : in  std_logic   -- Asserted when unexpected or malformed paket received
  );
end l2p_axi4_wr_initiator;


architecture rtl of l2p_axi4_wr_initiator is

  -- P2L DMA read request FSM
  type t_dma_state is (
     S_IDLE,
     S_SEND_HDR,
     S_SEND_ADDR_H,
     S_SEND_ADDR_L,
     S_SEND_WDAT,
     S_ERR,                 -- drains ammount of wdat,
     S_WAIT_B_ACCEPTED
  );

  signal dma_state : t_dma_state;

  signal remaining_axi_wdat_words : unsigned(8 downto 0);
  signal last_axi_wdat_word       : std_logic;

  signal w_ready_int : std_logic;

  signal sm_wdat_scfifo_rd_en : std_logic;

  type t_first_err_seen is (
     NO_ERR,
     ERR_AWLEN_VS_WLAST_EARLY,
     ERR_AWLEN_VS_WLAST_MISSING,
     ERR_RCVD_TXERR
  );
  signal err_detected   : t_first_err_seen;
  signal int_seen_error : std_logic;

  signal gennum_ready_for_wdat : std_logic;

  signal good_l2p_rdy_cnt_vc0      : unsigned(3 downto 0);
  signal good_l2p_rdy_cnt_vc1      : unsigned(3 downto 0);

  signal b_pending_error : std_logic;

  signal op_length : unsigned(9 downto 0);

  signal aw_ready_int : std_logic;

  signal sm_wdat_fifo_drain : std_logic;


begin

  w_ready_o   <= w_ready_int;

  w_ready_int <=   (sm_wdat_fifo_drain and w_valid_i)
                or (sm_wdat_scfifo_rd_en and w_valid_i and l2p_rdy_i);

  l2p_edb_o   <= '0';
  -- TBD work out if we actually need to drive this (e.g. wr implement response to tx_abort...)

  p_err_holding : process (clk_i)
  begin
    if rising_edge(clk_i) then
      if int_seen_error = '0' then
        if err_detected /= NO_ERR then
          int_seen_error <= '1';
          case err_detected is
            when ERR_AWLEN_VS_WLAST_EARLY   => err_code_o <= "0001";
            when ERR_AWLEN_VS_WLAST_MISSING => err_code_o <= "0010";
            when ERR_RCVD_TXERR             => err_code_o <= "0011";
            when others                     => err_code_o <= "0000";
          end case;
        end if;
      end if;

      if rst_n_i = '0' then
        int_seen_error <= '0';
        err_code_o     <= "0000";
      end if;
    end if;
  end process p_err_holding;

  seen_err_o <= int_seen_error;

-- because we can send data to either of the VC, we need to know which one(s) can take data..
  proc_gennum_ready_for_wdat_0 : process (clk_i)
  begin
    if rising_edge(clk_i) then
      if l_wr_rdy_i(0) = '1' and l2p_rdy_i  = '1'  then
        if good_l2p_rdy_cnt_vc0(3) = '0' then
          good_l2p_rdy_cnt_vc0 <= good_l2p_rdy_cnt_vc0 + 1;
        end if;
      else
        good_l2p_rdy_cnt_vc0 <= to_unsigned(0,4);
      end if;
    end if;
  end process proc_gennum_ready_for_wdat_0;


  proc_gennum_ready_for_wdat_1 : process (clk_i)
  begin
    if rising_edge(clk_i) then
      if l_wr_rdy_i(1) = '1' and l2p_rdy_i  = '1'  then
        if good_l2p_rdy_cnt_vc1(3) = '0' then
          good_l2p_rdy_cnt_vc1 <= good_l2p_rdy_cnt_vc0 + 1;
        end if;
      else
        good_l2p_rdy_cnt_vc1 <= to_unsigned(0,4);
      end if;
    end if;
  end process proc_gennum_ready_for_wdat_1;


  proc_vc_sel : process (good_l2p_rdy_cnt_vc1, good_l2p_rdy_cnt_vc0, aw_rec_i.vc)
  begin
    if aw_rec_i.vc = '1' then
      gennum_ready_for_wdat <= good_l2p_rdy_cnt_vc1(3);
    else
      gennum_ready_for_wdat <= good_l2p_rdy_cnt_vc0(3);
    end if;
  end process proc_vc_sel;


  op_length   <= ("0" & aw_rec_i.len);

  aw_ready_o  <= aw_ready_int;

  --  -----------------------------------------------------------------------------
  --  -- State machine
  --  -----------------------------------------------------------------------------
  p_fsm : process (clk_i)
  begin
    if rising_edge(clk_i) then
      err_detected         <= NO_ERR;
      axw_arb_req_o        <= '0';
      aw_ready_int         <= '0';
      sm_wdat_fifo_drain   <= '0';
      sm_wdat_scfifo_rd_en <= '0';
      b_chnl_valid_o       <= '0';

      case dma_state is
        when S_IDLE =>
          if aw_ready_int = '0' then

          -- Ideally a number of these values could be outside the "if";
          -- but after having seen instabilaties from the output fo the fwft fifo, the code was
          -- put in here. Moving some of these assignments out of the if may help timing/area.
            remaining_axi_wdat_words <= unsigned(aw_rec_i.len) -1;
            if aw_rec_i.len(8 downto 1) = "00000000" then
              last_axi_wdat_word <= '1';
            else
              last_axi_wdat_word <= '0';
            end if;

            b_chnl_resp_o <= aw_rec_i.b_resp;
            b_chnl_id_o   <= aw_rec_id_i;

            if aw_valid_i = '1' and ( aw_rec_i.err = '1' or b_pending_error = '1') then
              dma_state       <= S_ERR;
              b_pending_error <= '1';
            elsif aw_valid_i = '1' and w_valid_i = '1' and gennum_ready_for_wdat = '1' then
              dma_state     <= S_SEND_HDR;
              axw_arb_req_o <= '1';
            end if;
          end if;


        when S_SEND_HDR =>
          axw_arb_data_o  <= f_pack_l2p_header_write(tc              => aw_rec_i.tc,
                                                     snoop           => aw_rec_i.nosnoop,
                                                     addr_is_64b     => aw_rec_i.addr_is_64b,
                                                     frst_wrd_be     => aw_rec_i.fwbe,
                                                     last_wrd_be     => aw_rec_i.lwbe,
                                                     vc              => aw_rec_i.vc,
                                                     length_in_words => op_length);

          if arb_axw_gnt_i = '1' then
            axw_arb_req_o    <= '0';
            axw_arb_valid_o  <= '1';
            axw_arb_dframe_o <= '1';

            if aw_rec_i.addr_is_64b = '1' then
              dma_state <= S_SEND_ADDR_H;
            else
              dma_state <= S_SEND_ADDR_L;
            end if;
          else
            axw_arb_req_o    <= '1';
          end if;


        when S_SEND_ADDR_H =>
          axw_arb_data_o   <= aw_rec_i.addr(63 downto 32);
          axw_arb_valid_o  <= '1';
          dma_state        <= S_SEND_ADDR_L;


        when S_SEND_ADDR_L =>
          axw_arb_data_o       <= aw_rec_i.addr(31 downto 2) & "00";
          axw_arb_valid_o      <= '1';
          dma_state            <= S_SEND_WDAT;
          sm_wdat_scfifo_rd_en <= '1';


        when S_SEND_WDAT =>
          sm_wdat_scfifo_rd_en <= '1';
          axw_arb_valid_o      <= w_ready_int;

          if w_ready_int = '1' then
            axw_arb_data_o   <= w_dat_i;
                    -- if have sent all data
            if remaining_axi_wdat_words(8 downto 1) = "00000000" then
              last_axi_wdat_word <= '1';
            end if;

            -- Something has gone seriousley wrong. as we have an unexpected ammount of data...
            -- unlikey but maybe possible if axi clk domain reset is asserted at exactly the
            -- wrong time...
            assert w_last_i = last_axi_wdat_word report "Unhandeled Error condition"
              severity FAILURE;

            if last_axi_wdat_word = '0' then
              remaining_axi_wdat_words <= remaining_axi_wdat_words - 1;
              if w_last_i = '1' then
                err_detected <= ERR_AWLEN_VS_WLAST_EARLY;
              end if;
            else
              if w_last_i = '0' then
                err_detected <= ERR_AWLEN_VS_WLAST_MISSING;
              end if;
              axw_arb_dframe_o <= '0';
              if aw_rec_i.send_b = '1' then
                dma_state      <= S_WAIT_B_ACCEPTED;
                b_chnl_valid_o <= '1';
              else
                dma_state    <= S_IDLE;
                aw_ready_int <= '1';
              end if;
              sm_wdat_scfifo_rd_en <= '0';
            end if;
          end if;

          -- TBD: w_ready_int timeout -- (wdata underflow )

          if tx_error_i = '1' then
            axw_arb_dframe_o     <= '0';
            axw_arb_valid_o      <= '1';
            dma_state            <= S_ERR;
            err_detected         <= ERR_RCVD_TXERR;
            b_pending_error      <= '1';
            sm_wdat_scfifo_rd_en <= '0';
            b_chnl_resp_o        <= "10"; -- SLVERR
            -- If/when the hardware supports splitting a single AXI transaction
            -- into two (or more) on the PCIE bus. we will need to add logic to remember
            --   e.g. one AXI xaction gets split into three PCIE writes.
            --          Then if first of 3 PCIE ops gets an error, we need to remember to flag
            --          an error on BRESP for the final packet we see...
            --    32bit AXI data & 8 bit length gives us 1KB
            --             (larger than a 128 or 256B PCIE_MAX_XACTION_SIZE
            -- BUT, as the "contract" for using this interface is that the initiator is aware of
            --    PCIE 4K boundaries and transaction sizes this shoudl be a NO-OP....
          end if;


        when S_WAIT_B_ACCEPTED =>
          axw_arb_valid_o      <= '0';
          axw_arb_dframe_o     <= '0';

          if b_chnl_ready_i = '1' then
            b_chnl_valid_o  <= '0';
            dma_state       <= S_IDLE;
            aw_ready_int    <= '1';
            b_pending_error <= '0';
          else
            b_chnl_valid_o  <= '1';
          end if;

          -- TBD BRESP Timout..


        when S_ERR =>
          -- read all data from fifo than send B_resp if appropriate...
          axw_arb_dframe_o   <= '0';
          axw_arb_valid_o    <= '0';
          sm_wdat_fifo_drain <= '1';
          b_chnl_resp_o      <= "10"; -- SLVERR
          if w_ready_int = '1' then
            if remaining_axi_wdat_words(8 downto 1) = "00000000" then
              last_axi_wdat_word <= '1';
            end if;

            if last_axi_wdat_word = '0' then
              remaining_axi_wdat_words <= remaining_axi_wdat_words - 1;
            else
              sm_wdat_fifo_drain <= '0';
              if aw_rec_i.send_b = '1' then
                dma_state      <= S_WAIT_B_ACCEPTED;
                b_chnl_valid_o <= '1';
              else
                dma_state    <= S_IDLE;
                aw_ready_int <= '1';
              end if;
            end if;
          end if;


        when others =>
          dma_state          <= S_IDLE;
          b_pending_error    <= '0';
          axw_arb_valid_o    <= '0';
          axw_arb_dframe_o   <= '0';
      end case;

      -- Note: slightly odd way of doing this: but it means we dont get a rst acting as a clk_enable
      if rst_n_i = '0' then
        dma_state                   <= S_IDLE;
        err_detected                <= NO_ERR;
        b_pending_error             <= '0';
        axw_arb_valid_o             <= '0';
        axw_arb_dframe_o            <= '0';
      end if;
    end if;
  end process p_fsm;

end architecture rtl;
