--------------------------------------------------------------------------------
-- GN4124 core for PCIe FMC carrier
-- http://www.ohwr.org/projects/gn4124-core
--------------------------------------------------------------------------------
--
-- unit name:   l2p_axi4_wr_preprocessor
--
-- description: Implements a AXI4 peripheral slave write port, its purpose is to format and buffer
--              AXI write transactions ready for the L2P module..
--
--------------------------------------------------------------------------------
--
-- This source describes Open Hardware and is licensed under the CERN-OHL-W v2.
-- You may redistribute and modify this source and make products using it
-- under the terms of the CERN-OHL-W v2 or future versions (https://ohwr.org/cern_ohl_w_v2.txt).
--
-- This source is distributed WITHOUT ANY EXPRESS OR IMPLIED
-- WARRANTY, INCLUDING OF MERCHANTABILITY, SATISFACTORY
-- QUALITY AND FITNESS FOR A PARTICULAR PURPOSE. Please see
-- the CERN-OHL-W v2 for applicable conditions.
--
--------------------------------------------------------------------------------
-- Copyright CERN 2021
--------------------------------------------------------------------------------
--
-- NOTES:     Why axi4-slave-peripheral instead of axi-4-memory?
--                    a/ because it is impossible to implement non-modifiable transactions with
--                      strobes deasserted in the middle of a xfer over PCIE as a single transaction
--
--            AXI transactions are not allowed to cross a 4KB address boundary..  (PCIE and AXI)
--
--
-- NOT fully COMPLIANT with the AXI4 SPEC.....
--             a/ Incomplete support for AXI4 burst Types
--                    FIXED   NO     -- I dont see a PC needing a FIFO style write target
--                                    -- could support by splitting into multiple transactions,
--                                        but only if "non-modifiable"
--                    INCR    YES
--                    WRAP    NO     -- this interface is not intended for CPU cache line population
--
--                  i/ I dont expect FIXED or WRAP will be needed...
--                           a FIXED op of SIZE 64 len 3    on a 32bit bus
--                        is arguably modifiable back to    FIXED , len 1 on a 64bit bus....
--                     --- is it usefull?? it doesn't appear to be legal but I can see a use for it...
--
--             b/ AWCAHCE -- this maps to two functionalities...
--                       NOSNOOP in the L2p HEader / PCIE packet (function of AWCACHE(3,2,0)
--                       AWCACHE(1) defines if the treasaction is "modifiable"
--                             non MODIFIABLE transactions may be too large for PCIE or the GENNUM PHY
--
--             c/ SIZE -- Narrow transfers > 1 word are probably not going to be supported...
--                     will cause   RESP = slv_error.
--                     Reasoning: it hurst bandwidth,
--                     NOTE:UNLAIGNED TRANSFETRS!!!!!!!
--             d/ AXI write strobes:
--                      the core assumes that they are asserted in one single contiguous group.
--                      There is >= 1 valid byte enable in the first transfer and last transfer
--                      in a transaction
--             e/ AXI QOS => PCIE TC mapping....    TBD
--             f/ PCIE_VC number.... spec isn't too clear.
--                            will allow programmability but suggest leaving at '0'
--                            This probably wante reviewing!!!
--
--             g/ validate LEN vs Xaction type... e.g. burst & wrap have a max_length of 16;
--                                                only incr allows a 256 data transfer transaction.
--                         IF a transaction is INCR & (Len > 16) (even if non-modifiable is set)
--                          then it can be split to smaller biursts  NOTES: A3.4.1
--                           This is future work!!!
--------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;

use work.gn4124_axi4_pkg.all;

entity l2p_axi4_wr_preprocessor is
  generic (
    g_EN_AXI_QOS_2_TC_LOOKUP : boolean                      := false;
    g_PCIE_DEAGULT_QOS       : std_logic_vector(2 downto 0) := "000";
    g_DATA_BYTE_WIDTH        : positive                     := 4;
    g_ID_WIDTH               : positive                     := 2;
    g_EN_DEBUG_ASSERTS       : boolean                      := false
  );
  port (
    ---------------------------------------------------------
    clk_i                   : in  std_logic;
    rst_n_i                 : in  std_logic;      -- synch to clk, clock must toggle while asserted.

    seen_err_o              : out std_logic;
    err_code_o              : out std_logic_vector(3 downto 0);

    enable_no_snoop_i       : in  std_logic;
    vc_i                    : in  std_logic             := '0';
    enable_modifications_i  : in  std_logic             := '0'; -- currently unused

    ---------------------------------------------------------
    -- AXI4 input interfaces:
    ---------------------------------------------------------
    -- Write address channel
    aw_chnl_ready_o         : out std_logic;
    aw_chnl_valid_i         : in  std_logic;
    aw_chnl_i               : in  t_gn4124_axi4_ax_chnl;
    aw_chnl_id_i            : in  std_logic_vector(g_ID_WIDTH -1 downto 0);

    -- Write channel
    w_chnl_ready_o          : out std_logic;
    w_chnl_valid_i          : in  std_logic;
    w_chnl_data_i           : in  std_logic_vector(8 * g_DATA_BYTE_WIDTH - 1 downto 0);
    w_chnl_strb_i           : in  std_logic_vector(    g_DATA_BYTE_WIDTH - 1 downto 0);
    w_chnl_last_i           : in  std_logic;

    ---------------------------------------------------------
    -- Output interfaces to P2L
    ---------------------------------------------------------
    wo_rdy_i                : in  std_logic;
    wo_vld_o                : out std_logic;
    wo_dat_o                : out std_logic_vector(8 * g_DATA_BYTE_WIDTH -1 downto 0);
    wo_dat_eop_o            : out std_logic;                      -- indicate last beat/transfer of packet

    ctrl_rdy_i              : in  std_logic;
    ctrl_vld_o              : out std_logic;
    ctrl_rec_o              : out t_aw_preprocessed;
    ctrl_rec_id_o           : out std_logic_vector(g_ID_WIDTH -1 downto 0)
  );
end l2p_axi4_wr_preprocessor;


architecture rtl of l2p_axi4_wr_preprocessor is

  -- P2L DMA read request FSM
  type t_proc_state is (
     S_IDLE,
     S_WDAT_PLD,
     S_DISCARD_WDAT,
     S_PAD_WDAT,
     S_ERR,
     S_SEND_CMD
  );

  signal state : t_proc_state;

  signal remaining_axi_wdat_words : unsigned(8 downto 0);
  signal last_axi_wdat_word       : std_logic;

  signal aw_chl_sm    : t_gn4124_axi4_ax_chnl;
  signal aw_chl_sm_id : std_logic_vector(g_ID_WIDTH -1 downto 0);
  signal aw_sm_rdy    : std_logic;
  signal aw_sm_vld    : std_logic;

  signal sm_fwbe          : std_logic_vector(g_DATA_BYTE_WIDTH-1 downto 0);
  signal comb_fwbe        : std_logic_vector(g_DATA_BYTE_WIDTH-1 downto 0);
  signal bad_first_strobe : std_logic_vector(g_DATA_BYTE_WIDTH-1 downto 0);

  signal sm_first_word : std_logic;

  signal ctrl_vld_int  : std_logic;

  signal w_chl_en          : std_logic;
  signal sm_insert_padding : std_logic;
  signal sm_discard_dat    : std_logic;


  type t_err_code is (
      E_ALL_OK,                 -- no errors
      E_AW_CACHE_UNSUPPORTED,   -- AW_CACHE is not a supported value
      E_AW_BURST_UNSUPPORTED,   -- AW_BURST is not a supported value
      E_AW_SIZE_UNSUPPORTED,    -- AW_SIZE  is not a supported value
      E_W_UNDERRUN,             -- Insufficient data on W channel (as delimited by W_LAST)
      E_W_OVERRUN,              -- Too much data on W channel  (as delimited by W_LAST)
      E_W_FIRST_STROBE_NONE,    -- between AXI_ADDR[1:0] and strobes, FWBE would have been 0!
      E_W_FIRST_STROBE_U_ADDR,  -- A strobe that is below the index of AXI_ADDR[1:0] was asserted
      E_W_FIRST_STROBE_CONTIG,  -- For this transaction strobes must be contiguous!!
      E_W_LAST_STROBE,          -- for this transaction the last strobes were not contiguous.
      E_W_STROBES               -- becasue we are targeting PCIE,
                                --    the "middle" strobes should have been contiguous
      );

  signal sm_errcode : t_err_code;

  signal int_seen_error : std_logic;

  type t_strobe_checking_mode is (
    CHK_MD_SINGLE_DW,
    CHK_MD_2DW_UNALIGNED,
    CHK_MD_2DW_ALIGNED,
    CHK_MD_GTE_3_DW
  );

  signal strb_chk_md : t_strobe_checking_mode;


begin

  w_chnl_ready_o  <= (wo_rdy_i and w_chl_en) or sm_discard_dat;

  wo_vld_o        <= (w_chnl_valid_i and w_chl_en) or sm_insert_padding;
  wo_dat_o        <= w_chnl_data_i;
  wo_dat_eop_o    <= last_axi_wdat_word;


  -- catch the first error seen untill we reset the module.
  p_error_hold : process (clk_i)
  begin
    if rising_edge(clk_i) then
      if int_seen_error = '0' then
        if sm_errcode /= E_ALL_OK then
          int_seen_error <= '1';
          case sm_errcode is
            when E_AW_CACHE_UNSUPPORTED  =>  err_code_o <= "0001";
            when E_AW_BURST_UNSUPPORTED  =>  err_code_o <= "0010";
            when E_AW_SIZE_UNSUPPORTED   =>  err_code_o <= "0011";
            when E_W_UNDERRUN            =>  err_code_o <= "0100";
            when E_W_OVERRUN             =>  err_code_o <= "0101";
            when E_W_FIRST_STROBE_NONE   =>  err_code_o <= "0110";
            when E_W_LAST_STROBE         =>  err_code_o <= "0111";
            when E_W_STROBES             =>  err_code_o <= "1000";
            when E_W_FIRST_STROBE_U_ADDR =>  err_code_o <= "1001";
            when E_W_FIRST_STROBE_CONTIG =>  err_code_o <= "1010";
            when others =>                   err_code_o <= "0000";
          end case;
        end if;
      end if;

      if rst_n_i = '0' then
        int_seen_error <= '0';
        err_code_o     <= "0000";
      end if;
    end if;
  end process p_error_hold;


  seen_err_o <= int_seen_error;


  -- Block inside generate to additional signal definitions.
  b_block_aw_in_pl : block
    signal aw_chl_pl      : t_gn4124_axi4_ax_chnl;
    signal ax_chl_pl_id   : std_logic_vector(g_ID_WIDTH -1 downto 0);
    signal clken_aw_plreg : std_logic;
    signal clken_aw_opreg : std_logic;
    signal aw_use_pl      : std_logic;
  begin
    -- logic to give us "standarad AXI" write data control....
    axi_ax_pipeline_control : entity work.rl0_pl_stage_flowcontrol_srst
    port map (
      clk_i                 => clk_i,
      rst_n_i               => rst_n_i,
      in_rdy_o              => aw_chnl_ready_o,
      in_vld_i              => aw_chnl_valid_i,
      out_rdy_i             => aw_sm_rdy,
      out_vld_o             => aw_sm_vld,
      clk_en_pl_reg_o       => clken_aw_plreg,
      clk_en_op_reg_o       => clken_aw_opreg,
      use_dat_from_pl_reg_o => aw_use_pl
    );

    p_aw_register_pipeline : process (clk_i)
    begin
      if rising_edge(clk_i) then
        if clken_aw_plreg = '1' then
          aw_chl_pl    <= aw_chnl_i;
          ax_chl_pl_id <= aw_chnl_id_i;
        end if;

        if clken_aw_opreg = '1' then
          if aw_use_pl = '1' then
            aw_chl_sm    <= aw_chl_pl;
            aw_chl_sm_id <= ax_chl_pl_id;
          else
            aw_chl_sm    <= aw_chnl_i;
            aw_chl_sm_id <= aw_chnl_id_i;
          end if;
        end if;
      end if;
    end process p_aw_register_pipeline;

  end block b_block_aw_in_pl;


  proc_merge_be : process (sm_fwbe, w_chnl_strb_i)
  begin
    looplabel : for bitno in comb_fwbe'range loop
      comb_fwbe(bitno)        <= w_chnl_strb_i(bitno) and     sm_fwbe(bitno);
      bad_first_strobe(bitno) <= w_chnl_strb_i(bitno) and not sm_fwbe(bitno);
    end loop looplabel;
  end process proc_merge_be;



  -----------------------------------------------------------------------------
  -- PCIe read request FSM
  -----------------------------------------------------------------------------
  p_fsm : process (clk_i)
  begin
    if rising_edge(clk_i) then
      aw_sm_rdy         <= '0';
      sm_errcode        <= E_ALL_OK;
      sm_insert_padding <= '0';
      sm_discard_dat    <= '0';

      case state is
        when S_IDLE =>
          ctrl_rec_o.err  <= '0';
          ctrl_rec_o.addr <= aw_chl_sm.addr(63 downto 2);

          if aw_chl_sm.addr(63 downto 32) = x"00000000" then
            ctrl_rec_o.addr_is_64b <= '0';
          else
            ctrl_rec_o.addr_is_64b <= '1';
          end if;

          ctrl_rec_o.tc      <= G_PCIE_DEAGULT_QOS;
          ctrl_rec_o.nosnoop <= f_aw_cache_2_nosnoop(aw_chl_sm.cache, enable_no_snoop_i);
          ctrl_rec_o.vc      <= vc_i;

          -- OK can reassign this later but...
          ctrl_rec_o.len <= ('0' & unsigned(aw_chl_sm.len)) +1;

          case aw_chl_sm.len is
            when X"00" =>   strb_chk_md <= CHK_MD_SINGLE_DW;
            when X"01" => if aw_chl_sm.addr(3) = '0' then
                            strb_chk_md <= CHK_MD_2DW_ALIGNED;
                          else
                            strb_chk_md <= CHK_MD_2DW_UNALIGNED;
                          end if;
            when others =>  strb_chk_md <= CHK_MD_GTE_3_DW;
          end case;


          -- these should be calculable from AR_ADDR, AR_SIZE, AR_LEN
          case aw_chl_sm.addr(1 downto 0) is
            when "00" =>   sm_fwbe <= "1111";
            when "01" =>   sm_fwbe <= "1110";
            when "10" =>   sm_fwbe <= "1100";
            when others => sm_fwbe <= "1000";
          end case;

          ctrl_rec_o.lwbe          <= "1111";
          ctrl_rec_o.send_b        <= '1';
          ctrl_rec_o.b_resp        <= "00";    -- OK
          ctrl_rec_id_o            <= aw_chl_sm_id;
          sm_first_word            <= '1';
          remaining_axi_wdat_words <= ("0" & unsigned(aw_chl_sm.len));

          if aw_chl_sm.len = X"00" then
            last_axi_wdat_word <= '1';
          else
            last_axi_wdat_word <= '0';
          end if;

          if aw_sm_vld = '1' and aw_sm_rdy = '0' then
            w_chl_en          <= '1';
            -- Makes sure the cache bits are legal. - if not error...
            if f_aw_cache_is_legal(aw_chl_sm.cache) /= '1' then
              state              <= S_ERR;
              sm_errcode         <= E_AW_CACHE_UNSUPPORTED;
              last_axi_wdat_word <= '1';
            elsif aw_chl_sm.burst /= c_GN4124_AXI4_BURST_TYPE_INCR then  -- only support INCR
              state              <= S_ERR;
              sm_errcode         <= E_AW_BURST_UNSUPPORTED;
              last_axi_wdat_word <= '1';
            elsif aw_chl_sm.size /= "010" then   -- only support 32bit ops
              state              <= S_ERR;
              sm_errcode         <= E_AW_SIZE_UNSUPPORTED;
              last_axi_wdat_word <= '1';
            else
              state <= S_WDAT_PLD;
            end if;
          end if;


        when S_WDAT_PLD         =>
          w_chl_en     <= '1';

          if w_chnl_valid_i = '1' and w_chl_en = '1' and wo_rdy_i = '1' then
            if remaining_axi_wdat_words < 2 then
              last_axi_wdat_word <= '1';
            end if;

            remaining_axi_wdat_words <= remaining_axi_wdat_words -1;

            sm_first_word <= '0';

            if sm_first_word = '1' then
              ctrl_rec_o.fwbe <= comb_fwbe;
              if comb_fwbe = "0000" then -- bad AXI,
                                        -- Technically PCIE allowed for a single word xfer..
                assert not g_EN_DEBUG_ASSERTS
                  report "no write strobe was asserted for the first word of the transfer" severity error;

                ctrl_rec_o.err     <= '1';
                ctrl_rec_o.b_resp  <= "10";   -- SLV_ERR
                sm_errcode         <= E_W_FIRST_STROBE_NONE;
              end if;
              if bad_first_strobe /= "0000" then  -- STRB under addr bits  bad AXI
                assert not g_EN_DEBUG_ASSERTS
                  report "there was an unexpected write strobe asserted... " severity error;

                ctrl_rec_o.err     <= '1';
                ctrl_rec_o.b_resp  <= "10";   -- SLV_ERR
                sm_errcode         <= E_W_FIRST_STROBE_U_ADDR;
              end if;
              -- Validate no discontinuous bit patterns IF 3DW or 2DW unaligned
              --     deliberate use of spaces with the 2 bit values to graphically highlight
              --     relative alignments...
              if strb_chk_md = CHK_MD_2DW_UNALIGNED or strb_chk_md = CHK_MD_GTE_3_DW then
                if    "01"   = w_chnl_strb_i(3 downto 2)
                   or  "01"  = w_chnl_strb_i(2 downto 1)
                   or   "01" = w_chnl_strb_i(1 downto 0)
                then
                  assert not g_EN_DEBUG_ASSERTS
                     report "there was a discontinuous write strobe asserion pattern... " severity error;

                  ctrl_rec_o.err     <= '1';
                  ctrl_rec_o.b_resp  <= "10";   -- SLV_ERR
                  sm_errcode         <= E_W_FIRST_STROBE_CONTIG;
                end if;
              end if;
            end if;

            if w_chnl_last_i /= last_axi_wdat_word then   -- unexpected end condition
              w_chl_en <= '0';

              if w_chnl_last_i = '1' then  -- early end
                assert not g_EN_DEBUG_ASSERTS
                  report "Wlast bit was asserted, but we were expecting more data in the packet!, transaction will be discarded" severity error;

                state             <= S_PAD_WDAT;
                sm_insert_padding <= '1';
                sm_errcode        <= E_W_UNDERRUN;
              else -- too much data
                assert not g_EN_DEBUG_ASSERTS
                  report "WLAST bit should have been asserted, transaction will be discsardsd" severity error;

                state          <= S_DISCARD_WDAT;
                sm_errcode     <= E_W_OVERRUN;
                sm_discard_dat <= '1';
              end if;

            elsif w_chnl_last_i = '1' then -- good end condition,
              state    <= S_SEND_CMD;
              w_chl_en <= '0';

              if sm_first_word = '1' then  -- a single DWORD transfer requires all bits are 0
                ctrl_rec_o.lwbe <= "0000";
              else
                ctrl_rec_o.lwbe <= w_chnl_strb_i;

                if sm_first_word = '0' and strb_chk_md /= CHK_MD_2DW_ALIGNED then
                  -- check lwbe assertion values match that allowable on PCIE...
                  case w_chnl_strb_i(3 downto 0) is
                    when "1111" | "0111" | "0011" | "0001" =>
                      ctrl_rec_o.lwbe   <= w_chnl_strb_i;

                    when others =>
                      ctrl_rec_o.lwbe   <= "1111";
                      ctrl_rec_o.err    <= '1';
                      ctrl_rec_o.b_resp <= "10";   -- SLV_ERR
                      sm_errcode        <= E_W_LAST_STROBE;
                      assert not g_EN_DEBUG_ASSERTS
                           report "bad Strobe value on WLAST, Xaction will be dropped!!!" severity error;
                  end case;
                end if;
              end if;
            else -- normal data transfer
              -- detect Byte enable errors (e.g. deasserted in the middle of a burst....
              if sm_first_word = '0' and w_chnl_strb_i /= "1111" then
                ctrl_rec_o.err    <= '1';
                ctrl_rec_o.b_resp <= "10";   -- SLV_ERR
                sm_errcode        <= E_W_STROBES;
              end if;
            end if;
          end if;
          -- TBD do we need a timeout if we have not received data in a "reasonable period"??


        when S_ERR =>
          ctrl_rec_o.len    <= to_unsigned(1, ctrl_rec_o.len'length);
          ctrl_rec_o.err    <= '1';
          ctrl_rec_o.b_resp <= "10";   -- SLV_ERR

          --   Take the first wdat word, force it to the "last", and set the cmd to a lenght of 1 for discarding
          if w_chnl_valid_i = '1' and w_chl_en = '1' and wo_rdy_i = '1' then

            w_chl_en        <= '0';

             -- if it happened to be a single beat Xfer then immediatley send the cmd
            if w_chnl_last_i = '1' then
              ctrl_vld_int   <= '1';
              state          <= S_SEND_CMD;
            else  -- Otherwise discard the rest of the data...
              state          <= S_DISCARD_WDAT;
              sm_discard_dat <= '1';
            end if;
          end if;


        when S_DISCARD_WDAT =>
          sm_discard_dat <= '1';
          if w_chnl_valid_i = '1' then
            if w_chnl_last_i = '1' then
              sm_discard_dat <= '0';
              state          <= S_SEND_CMD;
              ctrl_vld_int   <= '1';
            end if;
          end if;

          ctrl_rec_o.err    <= '1';
          ctrl_rec_o.b_resp <= "10";   -- SLV_ERR


        when S_PAD_WDAT =>
          sm_insert_padding <= '1';

          ctrl_rec_o.err    <= '1';
          ctrl_rec_o.b_resp <= "10";   -- SLV_ERR

          if wo_rdy_i = '1' then
            if 2 > remaining_axi_wdat_words  then
              last_axi_wdat_word <= '1';
            end if;

            remaining_axi_wdat_words <= remaining_axi_wdat_words -1;

            if last_axi_wdat_word = '1' then
              sm_insert_padding <= '0';
              state             <= S_SEND_CMD;
              ctrl_vld_int      <= '1';
            end if;
          end if;


        when S_SEND_CMD =>
          ctrl_vld_int <= '1';
          if ctrl_vld_int = '1' and ctrl_rdy_i = '1' then
            ctrl_vld_int <= '0';
            state        <= S_IDLE;
            aw_sm_rdy    <= '1';
          end if;


        when others =>
          state <= S_IDLE;


      end case;

      -- Not implementing RESETas a if elsif end-if      so we dont get a clk_enable inferred!
      if rst_n_i = '0' then
        state         <= S_IDLE;
        w_chl_en      <= '0';
        ctrl_vld_int  <= '0';
      end if;
    end if;
  end process p_fsm;

  ctrl_vld_o <= ctrl_vld_int;

end architecture rtl;
