--------------------------------------------------------------------------------
-- GN4124 core for PCIe FMC carrier
-- http://www.ohwr.org/projects/gn4124-core
--------------------------------------------------------------------------------
--
-- entity name:   p2l_axi4_rd_dc
--
-- description: Wrapper for the modules that make the AXI4 read interface for the gennum phy.
--    Implemented as a wrapper to make deploying this as a Single Clock (sc)
--    or Dual Clock (dc) simpler...
--    This is the dual clock variant (_dc).
--    (Single Clock variants may be written & added to OHWR later).
--
--------------------------------------------------------------------------------
--
-- This source describes Open Hardware and is licensed under the CERN-OHL-W v2.
-- You may redistribute and modify this source and make products using it
-- under the terms of the CERN-OHL-W v2 or future versions (https://ohwr.org/cern_ohl_w_v2.txt).
--
-- This source is distributed WITHOUT ANY EXPRESS OR IMPLIED
-- WARRANTY, INCLUDING OF MERCHANTABILITY, SATISFACTORY
-- QUALITY AND FITNESS FOR A PARTICULAR PURPOSE. Please see
-- the CERN-OHL-W v2 for applicable conditions.
--
--------------------------------------------------------------------------------
-- Copyright CERN 2021
--------------------------------------------------------------------------------
--
-- NOTE this supports parts of the AXI4 SPEC.....
--     This implements a AXI4 Peripheral Slave port (functionally a subset of a "memory slave")
--            limitations:
--                AR_SIZE   is fixed at 2 = 32bit! (to match datapths
--                AR_BURST  only INCR SUPPORTED (WRAP and FIXED not supported)
--                AR_CACHE  a lookup is performed to drive the PCIE nosnoop bit.
--                               TODO: this will want reviewing (gn4124_axi4_pkg.vhd)
--                AR_LEN    (Note: a 32bit datapath gives us max Transaction size of 1KiB)
--                             we may need to implement a wider datapath to overcome this!
--                             Currently length is not modified and is passed through to PCIE:
--                              Initiator beware...
--                AR_ADDR   As the addr is passd through to PCIE
--                               this is the "physical address" a linux driver would use
--                               PCIE matches AXI4: dont cross 4K boundaries.
--                                TO BE ENFORCED BY THE INITIATOR!
--                AR_QOS    can be used as a lookup for the PICE TLP "Traffic Class"
--
--                Transactions are completed and returned in order.  with no modifications
--------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;

use work.gn4124_axi4_pkg.all;
use work.gencores_pkg.all;
use work.genram_pkg.all;

entity p2l_axi4_rd_dc is
  generic (
    g_IMPLEMENT_AXI          : boolean                      := false;
    g_ID_WIDTH               : positive                     := 2;
    g_DATA_WIDTH             : positive range 32 to 32      := 32;    -- fixed for now..
    g_AXI_AR_FIFO_WRDS       : positive                     := 16;    -- AXI4 AR Channel fifo depth
    g_AXI_R_FIFO_WRDS        : positive                     := 32;    -- AXI4  R Channel fifo depth
    g_EN_AXI_QOS_2_TC_LOOKUP : boolean                      := false; -- currently unused!
    g_PCIE_DEFAULT_TC        : std_logic_vector(2 downto 0) := "000"; -- default PCIE TC value
    g_RD_AXI_GN_INIT_IP_PL   : boolean                      := true   -- put a fifo on the input to the intiator form the fifo..
  );
  port (
    -- -------------------------------------------------------
    -- GN4124 core clock and reset
    clk_i                    : in  std_logic;
    rst_n_i                  : in  std_logic;  -- sync reset

    -- -------------------------------------------------------
    -- To the P2L Interface (send the DMA Master Read request)  <arm = Axi Read Master>
    -- -------------------------------------------------------
    -- req and gnt to get access from the arbitrator
    arm_arb_req_o            : out std_logic;
    arb_arm_gnt_i            : in  std_logic;
    -- data pipe to push data from this Axi Read Master to the arbitrator..
    arm_arb_valid_o          : out std_logic;  -- Read completion signals
    arm_arb_dframe_o         : out std_logic;  -- Toward the arbiter
    arm_arb_data_o           : out std_logic_vector(31 downto 0);


    -- -------------------------------------------------------
    -- AXI4 slave interface (read):   2 channels AR and R
    -- -------------------------------------------------------
    clk_axi_i                : in  std_logic;
    rst_axi_n_i              : in  std_logic;

    ar_ready_o               : out std_logic;
    ar_valid_i               : in  std_logic;
    ar_chnl_id_i             : in  std_logic_vector(g_ID_WIDTH-1 downto 0);
    ar_chnl_addr_i           : in  std_logic_vector(63 downto 0);
    ar_chnl_len_i            : in  std_logic_vector(7 downto 0);
    ar_chnl_size_i           : in  std_logic_vector(2 downto 0);
    ar_chnl_burst_i          : in  std_logic_vector(1 downto 0);
    ar_chnl_cache_i          : in  std_logic_vector(3 downto 0);
    ar_chnl_qos_i            : in  std_logic_vector(3 downto 0);

    r_ready_i                : in  std_logic;
    r_valid_o                : out std_logic;
    r_chnl_data_o            : out std_logic_vector(g_DATA_WIDTH -1 downto 0);
    r_chnl_last_o            : out std_logic;
    r_chnl_resp_o            : out std_logic_vector(1 downto 0);
    r_chnl_id_o              : out std_logic_vector(g_ID_WIDTH-1 downto 0);

    -- -------------------------------------------------------
    -- AXI4 preprocessor control / status signals
    -- -------------------------------------------------------
    preproc_seen_err_o       : out std_logic;
    preproc_err_code_o       : out std_logic_vector(2 downto 0);

    preproc_en_nosnoop_i     : in  std_logic;
    preproc_en_mods_i        : in  std_logic;
    preproc_mod_max_sz_i     : in  unsigned(7 downto 0);

    gn_seen_err_o            : out std_logic;
    gn_err_code_o            : out std_logic_vector(2 downto 0);

    -- ------------------------------------------------------
    -- From P2L Decoder (receive the read completion)
    -- ------------------------------------------------------
    -- Header
    pd_pdm_hdr_start_i       : in  std_logic;                     -- Header strobe
    pd_pdm_hdr_length_i      : in  std_logic_vector(9 downto 0);  -- Packet length in 32-bit words multiples
    pd_pdm_hdr_cid_i         : in  std_logic_vector(1 downto 0);  -- Completion ID
    pd_pdm_hdr_last_i        : in  std_logic;
    pd_pdm_hdr_stat_i        : in  std_logic_vector(1 downto 0);
    pd_pdm_master_cpld_i     : in  std_logic;                     -- Master read completion with data
    pd_pdm_master_cpln_i     : in  std_logic;                     -- Master read completion without data
    -- Data
    pd_pdm_data_valid_i      : in  std_logic;                      -- Indicates Data is valid
    pd_pdm_data_last_i       : in  std_logic;                      -- Indicates end of the packet
    pd_pdm_data_i            : in  std_logic_vector(31 downto 0);  -- Data
    pd_pdm_be_i              : in  std_logic_vector(3 downto 0);   -- Byte Enable for data

    -- -------------------------------------------------------
    -- P2L control
    p2l_rdy_o                : out std_logic;  -- De-asserted to pause transfer already in progress
    rx_error_o               : out std_logic   -- Asserted when transfer is aborted
  );
end p2l_axi4_rd_dc;


architecture struct of p2l_axi4_rd_dc is

  constant c_R_FIFO_FILL_LVL_W  : natural  := f_log2_ceil(g_AXI_R_FIFO_WRDS);
  constant c_R_BACKOFF_FILL_LVL : positive := g_AXI_R_FIFO_WRDS - 15;
      -- OK this has a downside in that it enforces a min size to the fifo; but should do for now!
      -- this may need to be tweaked as the AXI-slave is used in anger...
begin


  gen_without_axi_rd : if not g_IMPLEMENT_AXI generate
    arm_arb_req_o      <= '0';
    arm_arb_valid_o    <= '0';
    arm_arb_dframe_o   <= '0';
    arm_arb_data_o     <= (others => '0');
    ar_ready_o         <= '0';
    r_valid_o          <= '0';
    preproc_seen_err_o <= '0';
    preproc_err_code_o <= "000";
    gn_seen_err_o      <= '0';
    gn_err_code_o      <= "000";
    p2l_rdy_o          <= '1';
    rx_error_o         <= '0';
    r_chnl_data_o      <= (others => '0');
    r_chnl_resp_o      <= (others => '0');
    r_chnl_last_o      <= '0';
    r_chnl_id_o        <= (others => '0');
  end generate gen_without_axi_rd;


  gen_axi_rd_subsystem : if g_IMPLEMENT_AXI generate
--    b_axi_rd : block
    -- R channel fifo
    constant c_R_CHL_FIFO_W   : positive := g_DATA_WIDTH + 2 + 1 + g_ID_WIDTH;
    signal   r_fifo_wdat      : std_logic_Vector(c_R_CHL_FIFO_W-1 downto 0);
    signal   r_fifo_rdat      : std_logic_Vector(c_R_CHL_FIFO_W-1 downto 0);

    -- preprocessor => dc_fifo
    signal pp_ctrl_rdy       : std_logic;
    signal pp_ctrl_vld       : std_logic;
    signal pp_ctrl_rec       : t_ar_preprocessed;
    signal pp_ctrl_rec_id    : std_logic_vector(g_ID_WIDTH-1 downto 0);

    -- dc_fifo -> initiator
    signal int_ar_rec        : t_ar_preprocessed;
    signal int_ar_id         : std_logic_vector(g_ID_WIDTH-1 downto 0);
    signal int_ar_rdy        : std_logic;
    signal int_ar_vld        : std_logic;

    -- axi_initiator output (rdata)
    signal int_r_rdy         : std_logic;
    signal int_r_vld         : std_logic;
    signal int_r_rec         : t_gn4124_axi4_r_chnl;
    signal int_r_rec_id      : std_logic_vector(g_ID_WIDTH-1 downto 0);

    signal r_chl_dcfifo_fill : std_logic_vector(c_R_FIFO_FILL_LVL_W-1 downto 0);
    signal ar_chnl           : t_gn4124_axi4_ax_chnl;

    signal pp_ctrl_rec_slv   : std_logic_vector(g_ID_WIDTH + c_REC_AR_PREPROCESSED_SLV_WIDTH-1 downto 0);
    signal cmd_fifo_dout_slv : std_logic_vector(g_ID_WIDTH + c_REC_AR_PREPROCESSED_SLV_WIDTH-1 downto 0);

  begin



    ar_chnl <= (
                 addr  => ar_chnl_addr_i,
                 len   => ar_chnl_len_i,
                 size  => ar_chnl_size_i,
                 burst => ar_chnl_burst_i,
                 cache => ar_chnl_cache_i,
                 qos   => ar_chnl_qos_i
               );


    preproc_inst : entity work.p2l_axi4_rd_preprocessor
    generic map (
       g_EN_AXI_QOS_2_TC_LOOKUP  => g_EN_AXI_QOS_2_TC_LOOKUP,
       g_PCIE_DEAGULT_QOS        => g_PCIE_DEFAULT_TC,
       g_ID_WIDTH                => g_ID_WIDTH
    ) port map (
       clk_i                     => clk_axi_i,
       rst_n_i                   => rst_axi_n_i,
       seen_err_o                => preproc_seen_err_o,
       err_code_o                => preproc_err_code_o,
       enable_no_snoop_i         => preproc_en_nosnoop_i,
       vc_i                      => '0',
       enable_modifications_i    => preproc_en_mods_i,
       modified_max_size_i       => preproc_mod_max_sz_i,
       ar_ready_o                => ar_ready_o,
       ar_valid_i                => ar_valid_i,
       ar_chl_i                  => ar_chnl,
       ar_chl_id_i               => ar_chnl_id_i,
       ctrl_rdy_i                => pp_ctrl_rdy,
       ctrl_vld_o                => pp_ctrl_vld,
       ctrl_rec_o                => pp_ctrl_rec,
       ctrl_rec_id_o             => pp_ctrl_rec_id
    );




    --synthesis translate_off
    proc_rec_pack_unpack_assertion : process (clk_axi_i)
      variable decoded_from_slv : t_ar_preprocessed;
    begin
      if rising_edge(clk_axi_i) then
        decoded_from_slv := to_t_ar_preprocessed(pp_ctrl_rec_slv(c_REC_AR_PREPROCESSED_SLV_WIDTH-1 downto 0));
        assert decoded_from_slv = pp_ctrl_rec
          report "RECORD pack and unpack error" severity ERROR;
      end if;
    end process proc_rec_pack_unpack_assertion;
    --synthesis translate_on


    --CMD_FIFO control signals
    pp_ctrl_rec_slv <= pp_ctrl_rec_id & to_slv(pp_ctrl_rec);

    cmd_fifo : entity work.gn4124_axi_stream_dcfifo
    generic map(
      g_DATA_WIDTH => g_ID_WIDTH + c_REC_AR_PREPROCESSED_SLV_WIDTH,
      g_NUM_WORDS  => g_AXI_AR_FIFO_WRDS
    ) port map (
      snk_clk_i    => clk_axi_i,
      snk_rst_n_i  => rst_axi_n_i,
      snk_rdy_o    => pp_ctrl_rdy,
      snk_vld_i    => pp_ctrl_vld,
      snk_dat_i    => pp_ctrl_rec_slv,
      src_clk_i    => clk_i,
      src_rst_n_i  => rst_n_i,
      src_rdy_i    => int_ar_rdy,
      src_vld_o    => int_ar_vld,
      src_dat_o    => cmd_fifo_dout_slv
    );

    int_ar_rec <= to_t_ar_preprocessed(cmd_fifo_dout_slv(c_REC_AR_PREPROCESSED_SLV_WIDTH-1 downto 0));
    int_ar_id  <= cmd_fifo_dout_slv(cmd_fifo_dout_slv'high downto c_REC_AR_PREPROCESSED_SLV_WIDTH);

    gennum_master_axi_rd_init : entity work.p2l_axi4_rd_initiator
    generic map (
       g_USE_AR_REGISTER_PL    => g_RD_AXI_GN_INIT_IP_PL,
       g_ID_WIDTH              => g_ID_WIDTH,
       g_RD_FIFO_W             => c_R_FIFO_FILL_LVL_W,
       g_RD_FIFO_DEPTH         => g_AXI_R_FIFO_WRDS,
       g_RD_FIFO_SPACE_CMP_VAL => c_R_BACKOFF_FILL_LVL
    ) port map (
       clk_i                   => clk_i,
       rst_n_i                 => rst_n_i,
       seen_err_o              => gn_seen_err_o,
       err_code_o              => gn_err_code_o,
       ar_ready_o              => int_ar_rdy,
       ar_valid_i              => int_ar_vld,
       ar_rec_i                => int_ar_rec,
       ar_rec_id_i             => int_ar_id,
       r_ready_i               => int_r_rdy,
       r_valid_o               => int_r_vld,
       r_rec_o                 => int_r_rec,
       r_rec_id_o              => int_r_rec_id,
       r_chl_dcfifo_fill_lvl_i => r_chl_dcfifo_fill,
       tid_gnt_i               => '1',
       tid_req_o               => open,
       tid_val_i               => "10",
       cid_val_o               => open,
       cid_fin_o               => open,
       arb_axr_gnt_i           => arb_arm_gnt_i,
       axr_arb_req_o           => arm_arb_req_o,
       axr_arb_valid_o         => arm_arb_valid_o,
       axr_arb_dframe_o        => arm_arb_dframe_o,
       axr_arb_data_o          => arm_arb_data_o,
       pd_pdm_hdr_start_i      => pd_pdm_hdr_start_i,
       pd_pdm_hdr_length_i     => pd_pdm_hdr_length_i,
       pd_pdm_hdr_cid_i        => pd_pdm_hdr_cid_i,
       pd_pdm_hdr_last_i       => pd_pdm_hdr_last_i,
       pd_pdm_hdr_stat_i       => pd_pdm_hdr_stat_i,
       pd_pdm_master_cpld_i    => pd_pdm_master_cpld_i,
       pd_pdm_master_cpln_i    => pd_pdm_master_cpln_i,
       pd_pdm_data_valid_i     => pd_pdm_data_valid_i,
       pd_pdm_data_last_i      => pd_pdm_data_last_i,
       pd_pdm_data_i           => pd_pdm_data_i,
       pd_pdm_be_i             => pd_pdm_be_i,
       p2l_rdy_o               => p2l_rdy_o,
       rx_error_o              => rx_error_o
    );


        -- Pack and unpack SLV for the fifo
    r_fifo_wdat   <= int_r_rec_id & int_r_rec.last & int_r_rec.resp & int_r_rec.data;

    r_chnl_data_o <= r_fifo_rdat(g_DATA_WIDTH -1 downto 0);
    r_chnl_resp_o <= r_fifo_rdat(g_DATA_WIDTH+1 downto g_DATA_WIDTH);
    r_chnl_last_o <= r_fifo_rdat(g_DATA_WIDTH+2);
    r_chnl_id_o   <= r_fifo_rdat(r_fifo_rdat'high downto g_DATA_WIDTH+3);


    r_chl_dcfifo : entity work.gn4124_axi_r_chl_dcfifo
    generic map(
      g_DATA_WIDTH            => c_R_CHL_FIFO_W,
      g_NUM_WORDS             => g_AXI_R_FIFO_WRDS,
      g_ALMOST_FULL_THRESHOLD => c_R_BACKOFF_FILL_LVL,
      g_RESET_SRC_DAT         => TRUE
    ) port map (
      snk_clk_i               => clk_i,
      snk_rst_n_i             => rst_n_i,
      snk_rdy_o               => int_r_rdy,
      snk_vld_i               => int_r_vld,
      snk_dat_i               => r_fifo_wdat,
      snk_lvl_o               => r_chl_dcfifo_fill,
      snk_almost_full_o       => open,    -- here as an alternative to using the level myself
      src_clk_i               => clk_axi_i,
      src_rst_n_i             => rst_axi_n_i,
      src_rdy_i               => r_ready_i,
      src_vld_o               => r_valid_o,
      src_dat_o               => r_fifo_rdat
    );

  end generate gen_axi_rd_subsystem;
end architecture struct;
