--------------------------------------------------------------------------------
-- GN4124 core for PCIe FMC carrier
-- http://www.ohwr.org/projects/gn4124-core
--------------------------------------------------------------------------------
--
-- unit name:   p2l_axi4_rd_initiator
--
-- description: Interfaces into the existing P2L and L2p infrastructure of the gennum gn4124 IP core
--               as part of p2l_axi4_rd_dc
--
--------------------------------------------------------------------------------
--
-- This source describes Open Hardware and is licensed under the CERN-OHL-W v2.
-- You may redistribute and modify this source and make products using it
-- under the terms of the CERN-OHL-W v2 or future versions (https://ohwr.org/cern_ohl_w_v2.txt).
--
-- This source is distributed WITHOUT ANY EXPRESS OR IMPLIED
-- WARRANTY, INCLUDING OF MERCHANTABILITY, SATISFACTORY
-- QUALITY AND FITNESS FOR A PARTICULAR PURPOSE. Please see
-- the CERN-OHL-W v2 for applicable conditions.
--
--------------------------------------------------------------------------------
-- Copyright CERN 2021
--------------------------------------------------------------------------------
--
-- NOTES:  this does not currently use the existing "abort" infrastrucutre is one needed?
--
--
--
--
--------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;

use work.gn4124_core_pkg.all;
use work.gn4124_axi4_pkg.all;


entity p2l_axi4_rd_initiator is
  generic (
    g_USE_AR_REGISTER_PL    : boolean  := false;
    g_ID_WIDTH              : positive := 2;
    g_RD_FIFO_W             : natural;
    g_RD_FIFO_DEPTH         : positive;
    g_RD_FIFO_SPACE_CMP_VAL : positive
  );
  port (
    ---------------------------------------------------------
    -- GN4124 core clock and reset
    clk_i                   : in  std_logic;
    rst_n_i                 : in  std_logic;

    -- error reporting / debugging interface
    seen_err_o              : out std_logic;
    err_code_o              : out std_logic_vector(2 downto 0);
    ---------------------------------------------------------
    -- AXI4 read interface:
    ---------------------------------------------------------
    -- Read address channel
    ar_ready_o              : out std_logic;
    ar_valid_i              : in  std_logic;
    ar_rec_i                : in  t_ar_preprocessed;
    ar_rec_id_i             : in  std_logic_vector(g_ID_WIDTH-1 downto 0);

    -- read data channel
    r_ready_i               : in  std_logic;
    r_valid_o               : out std_logic;
    r_rec_o                 : out t_gn4124_axi4_r_chnl;
    r_rec_id_o              : out std_logic_vector(g_ID_WIDTH-1 downto 0);

    -- interface to extend the module(s) allowing multiple transactions in flight at once...
    --     but to do that we need to manage the TID/CID values in use across all modules
    -- currently not used..
    tid_gnt_i               : in  std_logic;
    tid_req_o               : out std_logic;
    tid_val_i               : in  std_logic_vector(1 downto 0); -- := g_TID_CID;
    cid_val_o               : out std_logic_vector(1 downto 0);
    cid_fin_o               : out std_logic;

    r_chl_dcfifo_fill_lvl_i : in  std_logic_vector(g_RD_FIFO_W-1 downto 0);
    -- TODO: should make this generic based on the fifo size, or "merge" the fifo into this module...

    -- data pipe to push data from this Axi Read Master to the arbitrator..
    axr_arb_valid_o         : out std_logic;  -- Read completion signals
    axr_arb_dframe_o        : out std_logic;  -- Toward the arbiter
    axr_arb_data_o          : out std_logic_vector(31 downto 0);
    -- req and gnt to get access from the arbitrator to send a command through the phy
    axr_arb_req_o           : out std_logic;
    arb_axr_gnt_i           : in  std_logic;

    ---------------------------------------------------------
    -- From P2L Decoder (receive the read completion)
    --
    -- Header
    pd_pdm_hdr_start_i      : in  std_logic;  -- Header strobe
    pd_pdm_hdr_length_i     : in  std_logic_vector(9 downto 0);  -- Packet length in 32-bit words
    pd_pdm_hdr_cid_i        : in  std_logic_vector(1 downto 0);  -- Completion ID
    pd_pdm_hdr_last_i       : in  std_logic;                     -- last completion for header
    pd_pdm_master_cpld_i    : in  std_logic;  -- Master read completion with data
    pd_pdm_master_cpln_i    : in  std_logic;  -- Master read completion without data
    pd_pdm_hdr_stat_i       : in  std_logic_vector(1 downto 0);  -- "00" = OK,
                                                                 -- "01" = unsupported_req,
                                                                 -- "10" = completere_abort,
                                                                 -- "11" = timeout
    -- Data
    pd_pdm_data_valid_i     : in  std_logic;  -- Indicates Data is valid
    pd_pdm_data_last_i      : in  std_logic;  -- Indicates end of the packet
    pd_pdm_data_i           : in  std_logic_vector(31 downto 0);  -- Data
    pd_pdm_be_i             : in  std_logic_vector(3 downto 0);  -- Byte Enable for data

    -- P2L control
    p2l_rdy_o               : out std_logic;  -- De-asserted to pause transfer already in progress
    rx_error_o              : out std_logic   -- Asserted when transfer is aborted

    --TBD: do I add any AXI stats.... e.g.
    --                AR_accept_2_R_end
    --                AR_Backpressured_cycles
    --                R_backpressured_cycles
  );
end p2l_axi4_rd_initiator;


architecture rtl of p2l_axi4_rd_initiator is

  -- P2L DMA read request FSM
  type t_cmd_state is (
     ST_IDLE,
     ST_GOT_OP,
     ST_SEND_HDR,
     ST_SEND_ADDR_H,
     ST_SEND_ADDR_L,
     ST_WAIT_DAT,
     ST_ERR
  );

  -- signals to allow optional ingress pipeline.
  signal ar_chl_sm            : t_ar_preprocessed;
  signal ar_chl_id_sm         : std_logic_vector(g_ID_WIDTH-1 downto 0);
  signal ar_sm_rdy            : std_logic;
  signal ar_sm_vld            : std_logic;

  signal cmd_state            : t_cmd_state;
  signal lcl_cid              : std_logic_vector(1 downto 0);
  signal err_push_last        : std_logic;

  signal cmd_length_in_words  : unsigned(9 downto 0);

  signal outstanding_data     : unsigned (8 downto 0);
  -- TBD to we add a outstanding_data_is_0 signal to help timing?

  signal rdata_fifo_has_space : std_logic;

  signal no_data_timeout      : unsigned(25 downto 0);

  signal rd_fifo_empty_cycle_count : unsigned(2 downto 0);
  signal rd_fifo_empty_duration_ok : std_logic;

  signal r_valid                   : std_logic;

  type t_err is (
     ERR_RXD_NSPRTD_REQ,     -- Completion, no data, and stat indicates: unsupported_req,
     ERR_RXD_CMPLT_ABRT,     -- Completion, no data, and stat indicates: completere_abort,
     ERR_RXD_TIMEOUT,        -- Completion, no data, and stat indicates: timeout
     ERR_RXD_OK,             -- Completion, no data, and stat indicates: OK  ? HOW did this happen?
     ERR_EXPECTED_HDR_LST,   -- This was expected to be the last packet of the completions received.
     ERR_EXPECTED_MORE_DATA, -- Probably (Hopefully) an unreachable error!
     NO_ERR                  -- no error, normal operation
  );

  signal sm_err_decode  : t_err;
  signal int_seen_error : std_logic;

  signal fifo_space_bodge_count : unsigned(3 downto 0);

  signal potential_op_fifo_depth : unsigned(g_RD_FIFO_W downto 0);


  signal op_fifo_has_space : std_logic;

--synthesis translate_off
  signal sm_active_cycles          : natural;
  signal sm_idle_backed_off_cycles : natural;
--synthesis translate_on

begin


  p_error_hold : process (clk_i)
  begin
    if rising_edge(clk_i) then
      if int_seen_error = '0' then
        if sm_err_decode /= NO_ERR then
          int_seen_error <= '1';
          case sm_err_decode is
            when ERR_RXD_NSPRTD_REQ     => err_code_o <= "001";
            when ERR_RXD_CMPLT_ABRT     => err_code_o <= "010";
            when ERR_RXD_TIMEOUT        => err_code_o <= "011";
            when ERR_RXD_OK             => err_code_o <= "100";
            when ERR_EXPECTED_HDR_LST   => err_code_o <= "101";
            when ERR_EXPECTED_MORE_DATA => err_code_o <= "111";
            when others                 => err_code_o <= "000";
          end case;
        end if;
      end if;

      if rst_n_i = '0' then
        int_seen_error <= '0';
        err_code_o     <= "000";
      end if;
    end if;
  end process p_error_hold;


  seen_err_o <= int_seen_error;

  r_valid_o <= r_valid;

  -- Block inside generate to additional signal definitions.
  gen_ar_pipeline : if g_USE_AR_REGISTER_PL generate
    signal ar_chl_pl      : t_ar_preprocessed;
    signal ar_chl_id_pl   : std_logic_vector(g_ID_WIDTH-1 downto 0);
    signal clken_ar_plreg : std_logic;
    signal clken_ar_opreg : std_logic;
    signal ar_use_pl      : std_logic;
  begin
    -- logic to give us "standarad AXI" write data control....
    axi_ax_pipeline_control : entity work.rl0_pl_stage_flowcontrol_srst
    port map (
      clk_i                 => clk_i,
      rst_n_i               => rst_n_i,
      in_rdy_o              => ar_ready_o,
      in_vld_i              => ar_valid_i,
      out_rdy_i             => ar_sm_rdy,
      out_vld_o             => ar_sm_vld,
      clk_en_pl_reg_o       => clken_ar_plreg,
      clk_en_op_reg_o       => clken_ar_opreg,
      use_dat_from_pl_reg_o => ar_use_pl
    );

    p_ar_register_pipeline : process (clk_i)
    begin
      if rising_edge(clk_i) then
        if clken_ar_plreg  = '1' then
          ar_chl_pl    <= ar_rec_i;
          ar_chl_id_pl <= ar_rec_id_i;
        end if;

        if clken_ar_opreg = '1' then
          if ar_use_pl = '1' then
            ar_chl_sm    <= ar_chl_pl;
            ar_chl_id_sm <= ar_chl_id_pl;
          else
            ar_chl_sm    <= ar_rec_i;
            ar_chl_id_sm <= ar_rec_id_i;
          end if;
        end if;
      end if;
    end process p_ar_register_pipeline;

  end generate gen_ar_pipeline;

  gen_no_ar_pipeline : if not g_USE_AR_REGISTER_PL generate
    ar_chl_sm    <= ar_rec_i;
    ar_chl_id_sm <= ar_rec_id_i;
    ar_sm_vld    <= ar_valid_i;
    ar_ready_o   <= ar_sm_rdy;
  end generate gen_no_ar_pipeline;


  cmd_length_in_words <= "0" & ar_chl_sm.len;

  -----------------------------------------------------------------------------
  -- PCIe read request FSM
  -----------------------------------------------------------------------------
  p_fsm : process (clk_i)
  begin
    if rising_edge(clk_i) then
      ar_sm_rdy     <= '0';
      sm_err_decode <= NO_ERR;

      case cmd_state is
        when ST_IDLE =>
          --if ar_sm_vld = '1' and ar_sm_rdy = '0' and rd_fifo_empty_duration_ok = '1' then  -- PC 1 MAY 2023 perf incr??
          if ar_sm_vld = '1' and ar_sm_rdy = '0' and op_fifo_has_space = '1' then
            if ar_chl_sm.err = '1' then
              cmd_state     <= ST_ERR;
              err_push_last <= ar_chl_sm.r_last;
            else
              cmd_state <= ST_GOT_OP;
              --tid_req_o   <= '1';
            end if;
            outstanding_data <= ar_chl_sm.len -1;
          end if;


        when ST_GOT_OP =>
          -- preparations for handeling multiple TID/CID...
          --tid_req_o   <= '1';
          --if '1' = tid_gnt then
          tid_req_o       <= '0';
          cmd_state     <= ST_SEND_HDR;
          axr_arb_req_o <= '1';
          lcl_cid       <= tid_val_i;
          --end if;

        when ST_SEND_HDR =>
          axr_arb_data_o   <= f_pack_l2p_header_read(tc              => ar_chl_sm.tc,
                                                     snoop           => ar_chl_sm.nosnoop,
                                                     addr_is_64b     => ar_chl_sm.addr_is_64b,
                                                     frst_wrd_be     => ar_chl_sm.fwbe,
                                                     last_wrd_be     => ar_chl_sm.lwbe,
                                                     vc              => ar_chl_sm.vc,
                                                     tid_cid         => lcl_cid,
                                                     length_in_words => cmd_length_in_words);

          if arb_axr_gnt_i = '1' then
            axr_arb_req_o    <= '0';
            axr_arb_valid_o  <= '1';
            axr_arb_dframe_o <= '1';

            if ar_chl_sm.addr_is_64b = '1' then
              cmd_state <= ST_SEND_ADDR_H;
            else
              cmd_state <= ST_SEND_ADDR_L;
            end if;
          end if;


        when ST_SEND_ADDR_H =>
          axr_arb_data_o <= ar_chl_sm.addr(63 downto 32);
          cmd_state      <= ST_SEND_ADDR_L;


        when ST_SEND_ADDR_L =>
          axr_arb_data_o(31 downto 2) <= ar_chl_sm.addr(31 downto 2);
          axr_arb_data_o(1 downto 0)  <= "00";
          axr_arb_dframe_o            <= '0';
          cmd_state                   <= ST_WAIT_DAT;


        when ST_WAIT_DAT =>
          -- TODO: I should use "stat" from the P2L header,
          --  and use this to flag completion error information!
          axr_arb_valid_o <= '0';

          if pd_pdm_hdr_cid_i = lcl_cid and pd_pdm_data_valid_i = '1' then
            -- read response destined for me!
            if pd_pdm_master_cpln_i = '1' then
              cmd_state <= ST_ERR;

              case pd_pdm_hdr_stat_i is
                when "01"   => sm_err_decode <= ERR_RXD_NSPRTD_REQ;
                when "10"   => sm_err_decode <= ERR_RXD_CMPLT_ABRT;
                when "11"   => sm_err_decode <= ERR_RXD_TIMEOUT;
                when others => sm_err_decode <= ERR_RXD_OK;
              end case;

            elsif pd_pdm_master_cpld_i = '1' then

              outstanding_data <= outstanding_data - 1;
              -- TBD worth checking what the stat value is ?? i.e. if non "00"

              if pd_pdm_data_last_i = '1' and outstanding_data = to_unsigned(0,9) then
                cmd_state <= ST_IDLE;
                ar_sm_rdy <= '1';

                if pd_pdm_hdr_last_i = '0' then
                  sm_err_decode <= ERR_EXPECTED_HDR_LST;
                end if;

                assert pd_pdm_hdr_last_i = '1'
                  report "Was expecting the phy chip to assert HDR_LAST with the final completion!"
                    severity error;
              end if;

              -- I am not sure if this is reachable, but seems like a reasonable check
              if     pd_pdm_data_last_i =  '1'
                 and pd_pdm_hdr_last_i  =  '1'
                 and outstanding_data   /= to_unsigned(0,9)
              then
                sm_err_decode <= ERR_EXPECTED_MORE_DATA;
              end if;

            end if;

          end if;

          --TBD: do we implement a timeout???    so we dont hang the AXI bus if there is an issue....

        when ST_ERR => -- in this state we MUST return the corect ammount of read data!!
          if rdata_fifo_has_space = '1' then
            outstanding_data <= outstanding_data - 1;
          end if;

          if outstanding_data = 0 then
            cmd_state <= ST_IDLE;
            ar_sm_rdy <= '1';
          end if;


        when others =>   -- matches reset condition... expect code coverage = unreachable!
          cmd_state        <= ST_IDLE;
          axr_arb_req_o    <= '0';
          axr_arb_valid_o  <= '0';
          axr_arb_dframe_o <= '0';
          tid_req_o        <= '0';
          ar_sm_rdy        <= '0';

      end case;

      if r_valid = '1' and r_ready_i = '0' then
        -- should never happen; BUT due to the way the fifo resynchronises the writepointer
        --  it is not impossible!!!
        cmd_state <= ST_ERR;
      end if;

      -- Not implementing RESETas a if elsif end-if      so we dont get a clk_enable inferred!
      if rst_n_i = '0' then
        cmd_state        <= ST_IDLE;
        axr_arb_req_o    <= '0';
        axr_arb_valid_o  <= '0';
        axr_arb_dframe_o <= '0';
        tid_req_o        <= '0';
        ar_sm_rdy        <= '0';
      end if;
    end if;
  end process p_fsm;


--synthesis translate_off
  p_simonly_perf : process (clk_i)
  begin
    if rising_edge(clk_i) then
      if cmd_state = ST_IDLE and ar_sm_vld = '1' and ar_sm_rdy = '0' then
        sm_active_cycles <= sm_active_cycles + 1;
        if rd_fifo_empty_duration_ok = '0' and op_fifo_has_space = '1' then
          sm_idle_backed_off_cycles <= sm_idle_backed_off_cycles + 1;
        end if;
      elsif cmd_state /= ST_IDLE then
        sm_active_cycles <= sm_active_cycles + 1;
      end if;
      if rst_n_i = '0' then
        sm_idle_backed_off_cycles <= 0;
        sm_active_cycles <= 0;
      end if;
    end if;
  end process p_simonly_perf;
--synthesis translate_on


  p_fifo_space : process (clk_i)
  begin
    if rising_edge(clk_i) then
      if cmd_state = ST_IDLE then
        fifo_space_bodge_count <= (others => '1');
      elsif fifo_space_bodge_count /= "0000" then
        fifo_space_bodge_count <= fifo_space_bodge_count - 1;
      end if;

      potential_op_fifo_depth <= unsigned(r_chl_dcfifo_fill_lvl_i)
                               + fifo_space_bodge_count
                               + ar_chl_sm.len
                               + to_unsigned(3, potential_op_fifo_depth'length);

      if potential_op_fifo_depth < to_unsigned(g_RD_FIFO_DEPTH, potential_op_fifo_depth'length) then
        op_fifo_has_space <= '1';
      else
        op_fifo_has_space <= '0';
      end if;

      if rst_n_i = '0' then
        fifo_space_bodge_count <= (others => '0');
      end if;
    end if;
  end process p_fifo_space;




  -- NOTE: we need to be really carefull here the value g_RD_FIFO_SPACE_CMP_VAL used in the
  --       comparison. it is a bit of a "magic" number...
  --       Ideally we would use an axi style FIFO that had "ready" (full almost full etc)
  --       synchronous to the value of fill level, but that is not to the authors knowledge
  --       available in the general-cores lib.
  p_ready : process (clk_i)
  begin
    if rising_edge(clk_i) then
      if unsigned(r_chl_dcfifo_fill_lvl_i)
             < to_unsigned(g_RD_FIFO_SPACE_CMP_VAL,r_chl_dcfifo_fill_lvl_i'length)
      then
        rdata_fifo_has_space <= '1';
      else
        rdata_fifo_has_space <= '0';
      end if;

      -- I beleive it is a model bug; but it "looses" the read completion if rdy is deasserted
      --  at the wrong time....  requiring this fifo to drain and then make sure ready
      --  (as we drive it) isn't deasserted for n cycles before issuing another read "masks"
      --  the issue

      if unsigned(r_chl_dcfifo_fill_lvl_i) = 0 then
        if rd_fifo_empty_cycle_count = 0 then
          rd_fifo_empty_duration_ok <= '1';
        else
          rd_fifo_empty_cycle_count <= rd_fifo_empty_cycle_count - 1;
        end if;
      else
        rd_fifo_empty_cycle_count <= (others => '1');
                     -- 7 cycles of being empty before we can get ready to issue the next command
        rd_fifo_empty_duration_ok <= '0';
      end if;

      if rst_n_i = '0' then
        rd_fifo_empty_cycle_count <= (others => '1');
                     -- 7 cycles of being empty before we can get ready to issue the next command
        rd_fifo_empty_duration_ok <= '0';
      end if;
    end if;
  end process p_ready;


  p2l_rdy_o <= rdata_fifo_has_space;


  p_read_dat_out : process (clk_i)
    variable v_is_last : std_logic;
  begin
    if rising_edge(clk_i) then
      r_valid <= '0';
      if ST_WAIT_DAT = cmd_state then
        if pd_pdm_hdr_cid_i = lcl_cid and pd_pdm_data_valid_i = '1' and pd_pdm_master_cpld_i = '1'
        then
          if ar_chl_sm.r_last = '1' and outstanding_data = 0 and pd_pdm_data_last_i = '1' then
            v_is_last :=  '1';
          else
            v_is_last :=  '0';
          end if;

          r_valid    <= '1';
          r_rec_id_o <= ar_chl_id_sm;
          r_rec_o    <= (data => pd_pdm_data_i,
                         resp => ar_chl_sm.r_resp,
                         last => v_is_last);
        end if;

      elsif ST_ERR = cmd_state then
        if    ar_chl_sm.r_last = '1'
          and outstanding_data = 0
        then
          v_is_last :=  '1';
        else
          v_is_last :=  '0';
        end if;

        r_valid      <= rdata_fifo_has_space;
        r_rec_o.data <= x"deadbeef";         -- not origional, but it may be handy for debugging
        r_rec_o.last <= v_is_last;
        r_rec_id_o   <= ar_chl_id_sm;

        -- we can be in the error state for 2 reasons:
        --    we were asked to respond with error,
        --    or we got an error
        if ar_chl_sm.err = '1' then
          r_rec_o.resp <= ar_chl_sm.r_resp;
        else
          r_rec_o.resp <= "10";       -- response SLVERR (I hope)
        end if;
      end if;

      if rst_n_i = '0' then
        r_valid <= '0';
      end if;
    end if;
  end process p_read_dat_out;

  rx_error_o <= '0';

end architecture rtl;
