--------------------------------------------------------------------------------
-- GN4124 core for PCIe FMC carrier
-- http://www.ohwr.org/projects/gn4124-core
--------------------------------------------------------------------------------
--
-- unit name:   p2l_axi4_rd_preprocessor
--
-- description: Implements a AXI4 peripheral slave write port, its purpose is to format and buffer
--              AXI read transactions ready for the L2P interface.
--
--------------------------------------------------------------------------------
--
-- This source describes Open Hardware and is licensed under the CERN-OHL-W v2.
-- You may redistribute and modify this source and make products using it
-- under the terms of the CERN-OHL-W v2 or future versions (https://ohwr.org/cern_ohl_w_v2.txt).
--
-- This source is distributed WITHOUT ANY EXPRESS OR IMPLIED
-- WARRANTY, INCLUDING OF MERCHANTABILITY, SATISFACTORY
-- QUALITY AND FITNESS FOR A PARTICULAR PURPOSE. Please see
-- the CERN-OHL-W v2 for applicable conditions.
--
--------------------------------------------------------------------------------
-- Copyright CERN 2021
--------------------------------------------------------------------------------
--
--
-- TBD:     implement enable modifications : e.g. splitting Xactions to PCIE_MAX_READ_REQ_SIZE,
--                                              4K boundary checks


library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;

use work.gn4124_axi4_pkg.all;

entity p2l_axi4_rd_preprocessor is
  generic (
    g_EN_AXI_QOS_2_TC_LOOKUP : boolean                      := false;
    g_PCIE_DEAGULT_QOS       : std_logic_vector(2 downto 0) := "000";
    g_ID_WIDTH               : positive                     := 2
  );
  port (
    ---------------------------------------------------------
    clk_i                    : in  std_logic;
    rst_n_i                  : in  std_logic;

    seen_err_o               : out std_logic;
    err_code_o               : out std_logic_vector(2 downto 0);

    enable_no_snoop_i        : in  std_logic;
    vc_i                     : in  std_logic             := '0';

    enable_modifications_i   : in  std_logic             := '0';
    modified_max_size_i      : in  unsigned(7 downto 0);

    ---------------------------------------------------------
    -- AXI4 input interfaces:
    ---------------------------------------------------------
    -- Write address channel
    ar_ready_o               : out std_logic;
    ar_valid_i               : in  std_logic;
    ar_chl_i                 : in  t_gn4124_axi4_ax_chnl;
    ar_chl_id_i              : in  std_logic_vector(g_ID_WIDTH-1 downto 0);

    ctrl_rdy_i               : in  std_logic;
    ctrl_vld_o               : out std_logic;
    ctrl_rec_o               : out t_ar_preprocessed;
    ctrl_rec_id_o            : out std_logic_vector(g_ID_WIDTH-1 downto 0)
  );
end p2l_axi4_rd_preprocessor;


architecture rtl of p2l_axi4_rd_preprocessor is

  type t_proc_state is (
     S_IDLE,
     S_ERR,
     S_SEND_CMD
  );

  signal state : t_proc_state;


  type t_err_code is (
      ERR_NONE,
      ERR_RECOVERABLE_AR_CACHE_ILLEGAL_VALUE,
      ERR_RECOVERABLE_AR_BURST_TYPE_UNSUPPORTED,
      ERR_RECOVERABLE_AR_SIZE_UNSUPPORTED
      );

  signal sm_errcode     : t_err_code;
  signal int_seen_error : std_logic;


  -- signal to enable optional pipelining on input...   & it makes sure it is registered instead of comb!
  signal ar_chl_sm      : t_gn4124_axi4_ax_chnl;
  signal ar_chl_sm_id   : std_logic_vector(g_ID_WIDTH-1 downto 0);
  signal ar_sm_rdy      : std_logic;
  signal ar_sm_vld      : std_logic;
                        
  signal ctrl_vld_int   : std_logic;

begin

  b_ar_in_pl : block
    signal ar_chl_pl      : t_gn4124_axi4_ax_chnl;
    signal ar_chl_id_pl   : std_logic_vector(g_ID_WIDTH-1 downto 0);
    signal clken_ar_plreg : std_logic;
    signal clken_ar_opreg : std_logic;
    signal ar_use_pl      : std_logic;
  begin
    -- logic to give us "standarad AXI" write data control....
    -- if we had VHDL2008 this would just take the generic type and be a single instantiation 
    --  instead of a block!
    axi_ar_pipeline_control : entity work.rl0_pl_stage_flowcontrol_srst
    port map (
      clk_i                 => clk_i,
      rst_n_i               => rst_n_i,
      in_rdy_o              => ar_ready_o,
      in_vld_i              => ar_valid_i,
      out_rdy_i             => ar_sm_rdy,
      out_vld_o             => ar_sm_vld,
      clk_en_pl_reg_o       => clken_ar_plreg,
      clk_en_op_reg_o       => clken_ar_opreg,
      use_dat_from_pl_reg_o => ar_use_pl
    );


    p_ar_register_pipeline : process (clk_i)
    begin
      if rising_edge(clk_i) then
        if clken_ar_plreg = '1' then
          ar_chl_pl    <= ar_chl_i;
          ar_chl_id_pl <= ar_chl_id_i;
        end if;

        if clken_ar_opreg = '1' then
          if ar_use_pl = '1' then
            ar_chl_sm    <= ar_chl_pl;
            ar_chl_sm_id <= ar_chl_id_pl;
          else
            ar_chl_sm    <= ar_chl_i;
            ar_chl_sm_id <= ar_chl_id_i;
          end if;
        end if;
      end if;
    end process p_ar_register_pipeline;
  end block b_ar_in_pl;



  -- catch the first error seen and hold untill we reset the module.
  p_error_hold : process (clk_i)
  begin
    if rising_edge(clk_i) then
      if int_seen_error = '0' then
        if sm_errcode /= ERR_NONE then
          int_seen_error <= '1';
          case sm_errcode is
            when ERR_RECOVERABLE_AR_CACHE_ILLEGAL_VALUE    => err_code_o <= "001";
            when ERR_RECOVERABLE_AR_BURST_TYPE_UNSUPPORTED => err_code_o <= "010";
            when ERR_RECOVERABLE_AR_SIZE_UNSUPPORTED       => err_code_o <= "011";
            when others                                    => err_code_o <= "000";
          end case;
        end if;
      end if;

      if rst_n_i = '0' then
        int_seen_error <= '0';
        err_code_o     <= "000";
      end if;
    end if;
  end process p_error_hold;

  seen_err_o <= int_seen_error;


  -----------------------------------------------------------------------------
  -- PCIe read request FSM
  -----------------------------------------------------------------------------
  p_fsm : process (clk_i)
  begin
    if rising_edge(clk_i) then
      ar_sm_rdy  <= '0';
      sm_errcode <= ERR_NONE;

      case state is
        when S_IDLE =>

          ctrl_rec_o.err     <= '0';
          ctrl_rec_o.addr    <= ar_chl_sm.addr(63 downto 2);
          ctrl_rec_o.tc      <= g_PCIE_DEAGULT_QOS;
          ctrl_rec_o.nosnoop <= f_ar_cache_2_nosnoop(ar_chl_sm.cache, enable_no_snoop_i);
          ctrl_rec_o.vc      <= vc_i;

          if ar_chl_sm.addr(63 downto 32) = x"00000000" then
            ctrl_rec_o.addr_is_64b <= '0';
          else
            ctrl_rec_o.addr_is_64b <= '1';
          end if;

          ctrl_rec_o.len <= ('0' & unsigned(ar_chl_sm.len)) +1;

          -- these should be calculable from AR_ADDR, AR_SIZE, AR_LEN
          case ar_chl_sm.addr(1 downto 0) is
            when "00" =>   ctrl_rec_o.fwbe <= "1111";
            when "01" =>   ctrl_rec_o.fwbe <= "1110";
            when "10" =>   ctrl_rec_o.fwbe <= "1100";
            when others => ctrl_rec_o.fwbe <= "1000";
          end case;

          if ar_chl_sm.len = x"00" then
            ctrl_rec_o.lwbe   <= "0000";
          else
            ctrl_rec_o.lwbe   <= "1111";
          end if;
          ctrl_rec_o.r_last <= '1';  -- FOR now, the initiator is responsible for length
          ctrl_rec_o.r_resp <= "00"; -- OK
          ctrl_rec_id_o     <= ar_chl_sm_id;

          if ar_sm_vld = '1' and ar_sm_rdy = '0' then

             -- Makes sure the cache bits are legal. - if not error...
            if f_ar_cache_is_legal(ar_chl_sm.cache) /= '1' then
              state      <= S_ERR;
              sm_errcode <= ERR_RECOVERABLE_AR_CACHE_ILLEGAL_VALUE;
            elsif ar_chl_sm.burst /= C_GN4124_AXI4_BURST_TYPE_INCR then  -- only support INCR
              state      <= S_ERR;
              sm_errcode <= ERR_RECOVERABLE_AR_BURST_TYPE_UNSUPPORTED;
            elsif ar_chl_sm.size /= "010" then   -- only support 32bit ops
              state     <= S_ERR;
              sm_errcode <= ERR_RECOVERABLE_AR_SIZE_UNSUPPORTED;
            -- TBD: we might want to do burst splitting OR equivalent here as another elsif choice
            else
              state <= S_SEND_CMD;
            end if;
          end if;


        when S_ERR              =>
          state             <= S_SEND_CMD;
          ctrl_rec_o.err    <= '1';
          ctrl_rec_o.r_resp <= "10";    -- SLVERR


        when S_SEND_CMD =>
          ctrl_vld_int <= '1';
          if ctrl_vld_int = '1' and ctrl_rdy_i = '1' then
            ctrl_vld_int <= '0';
            state        <= S_IDLE;
            ar_sm_rdy    <= '1';
          end if;

        when others =>
          state <= S_IDLE;

      end case;

      -- Not implementing RESET as an if elsif end-if      so we dont get a clk_enable inferred!
      if rst_n_i = '0' then
        state           <= S_IDLE;
        ctrl_vld_int    <= '0';
        ar_sm_rdy       <= '0';
        sm_errcode      <= ERR_NONE;
      end if;
    end if;
  end process p_fsm;

  ctrl_vld_o <= ctrl_vld_int;

end architecture rtl;
