--------------------------------------------------------------------------------
-- GN4124 core for PCIe FMC carrier
-- http://www.ohwr.org/projects/gn4124-core
--------------------------------------------------------------------------------
--
-- unit name:   rl0_pl_stage_flowcontrol_srst
--
-- description:  generates AXI_STREAM or AVALON_ST (RL0) flowcontrol signals
--               and provide the control signals to make a pipeline stage of whatever type
--
--
--------------------------------------------------------------------------------
--
-- This source describes Open Hardware and is licensed under the CERN-OHL-W v2.
-- You may redistribute and modify this source and make products using it
-- under the terms of the CERN-OHL-W v2 or future versions (https://ohwr.org/cern_ohl_w_v2.txt).
--
-- This source is distributed WITHOUT ANY EXPRESS OR IMPLIED
-- WARRANTY, INCLUDING OF MERCHANTABILITY, SATISFACTORY
-- QUALITY AND FITNESS FOR A PARTICULAR PURPOSE. Please see
-- the CERN-OHL-W v2 for applicable conditions.
--
--------------------------------------------------------------------------------
-- Copyright CERN 2021
--------------------------------------------------------------------------------
--
-- NOTES:
--         expected usecase:    control for a axi pipeline stage, but as we dont have generic types
--                              in ISE, this at least allows re-use of pre-verified control logic!
--
--      ..........
--          signal internal_pipeline_stage : t_data_or_rec;
--          signal output_data             : t_data_or_rec;
--
--       BEGIN
--     ..................
--          <instantiate the component here >
--        ...................
--
--          p_pl_stage : process (clk) begin
--             if rising_edge(clk) then
--                if clk_en_pl_reg = '1' then
--                   internal_pipeline_stage <= data_to_pipeline;
--                end if;
--
--                if clk_en_op_reg = '1' then
--                   if use_dat_from_pl_reg = '1' then
--                      output_data <= internal_pipeline_stage;
--                   else
--                      output_data <= data_to_pipeline;
--                   end if;
--                end if;
--             end if;
--          end process p_pl_stage;
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;

entity rl0_pl_stage_flowcontrol_srst is
  port (
    clk_i                 : in  std_logic;
    rst_n_i               : in  std_logic;

    in_rdy_o              : out std_logic;
    in_vld_i              : in  std_logic;

    out_rdy_i             : in  std_logic;
    out_vld_o             : out std_logic;

    clk_en_pl_reg_o       : out std_logic;
    clk_en_op_reg_o       : out std_logic;
    use_dat_from_pl_reg_o : out std_logic
  );
end entity rl0_pl_stage_flowcontrol_srst;



architecture rtl of rl0_pl_stage_flowcontrol_srst is
-- TBD: is there a XILINX, Intel(Altera), and microsemi freindly way to not optimise this out....,
--     or do I need to go to a per vendor architecture approach....
  signal pl_reg_is_valid  : std_logic;

  signal in_ready_int     : std_logic;
  signal out_valid_int    : std_logic;
begin

  clk_en_pl_reg_o       <= in_vld_i and in_ready_int;
  use_dat_from_pl_reg_o <= pl_reg_is_valid;

  out_vld_o <= out_valid_int;
  in_rdy_o  <= in_ready_int;

  p_comb_clken : process (out_valid_int, out_rdy_i, in_vld_i)
  begin
    if    (out_valid_int = '1' and out_rdy_i = '1')
       or (out_valid_int = '0' and in_vld_i  = '1')
    then
      clk_en_op_reg_o <= '1';
    else
      clk_en_op_reg_o <= '0';
    end if;
  end process p_comb_clken;


  p_reg_state_tracking : process (clk_i)
  begin
    if rising_edge(clk_i) then
      if rst_n_i = '0' then
        pl_reg_is_valid <= '0';
        in_ready_int    <= '1';
      else
        if pl_reg_is_valid = '1' and out_rdy_i = '1' and out_valid_int = '1' then
          pl_reg_is_valid <= '0';
          in_ready_int    <= '1';
        elsif pl_reg_is_valid = '0' and in_vld_i = '1' and out_valid_int = '1' and out_rdy_i = '0' then
          pl_reg_is_valid <= '1';
          in_ready_int    <= '0';
        end if;
      end if;
    end if;
  end process p_reg_state_tracking;


  p_out_valid_reg : process (clk_i)
  begin
    if rising_edge(clk_i) then
      if rst_n_i = '0' then
        out_valid_int <= '0';
      else
        if out_valid_int = '1' and out_rdy_i = '1' then
          out_valid_int <= '0';
          if pl_reg_is_valid = '1' then
            out_valid_int <= '1';
          elsif pl_reg_is_valid = '0' and in_vld_i = '1' then
            out_valid_int <= '1';
          end if;
        elsif out_valid_int = '0' and in_vld_i = '1' then
          out_valid_int <= '1';
        end if;
      end if;
    end if;
  end process p_out_valid_reg;

end architecture rtl;
