//------------------------------------------------------------------------------
// CERN BE-CO-HT
// GN4124 core for PCIe FMC carrier
// http://www.ohwr.org/projects/gn4124-core
//------------------------------------------------------------------------------
//
// unit name:   main
//
// description: This is a simple example testbench, to demonstrate how to use
// the SystemVerilog BFM of the GN4124 to perform simple accesses over wishbone.
//
// The testbench simply connects the wishbone master of the GN4124 to its own
// DMA configuration wishbone slave and attaches a pre-initialised dummy RAM
// with a wishbone interface to the pipelined DMA interface in order to perform
// a DMA read.
//
//------------------------------------------------------------------------------
// Copyright CERN 2021
//------------------------------------------------------------------------------
// Copyright and related rights are licensed under the Solderpad Hardware
// License, Version 2.0 (the "License"); you may not use this file except
// in compliance with the License. You may obtain a copy of the License at
// http://solderpad.org/licenses/SHL-2.0.
// Unless required by applicable law or agreed to in writing, software,
// hardware and materials distributed under this License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
// or implied. See the License for the specific language governing permissions
// and limitations under the License.
//------------------------------------------------------------------------------


`timescale 1ns/1ps

`include "gn4124_bfm.svh"

import wishbone_pkg::*;

`define UUT_GN4124 DUT.uut.cmp_wrapped_gn4124
module main #(
                bit p_WB_DMA = 1,  // params to define if we test features: WB_DMA
                bit p_AXI_SL = 1   // params to define if we test features: AXI_SL
             );

   reg   clk_125m = 0;
   reg   clk_62m5 = 0;
   logic axi_clk  = 0;    // this is SV after all!

   logic gn4124_irq;

   logic [31:0] gn4124_core_status;
   logic        gn_tb_reset;

   string testphase     = "null:= no phase set yet";
   string test_subphase = "ha!";

   t_wishbone_master_in  wb_in,  wb_dma_in,  wb_mem_in;
   t_wishbone_master_out wb_out, wb_dma_out, wb_mem_out;

   always #4ns clk_125m <= ~clk_125m;
   always #23ns clk_62m5 <= ~clk_62m5;
   always #5ns axi_clk <= ~axi_clk;

   logic rst_125m_n;
   logic rst_62m5_n;

   logic rst_gn4124_n;
   logic clk_gn4124;

   logic wb_dma_clk;
   logic wb_dma_rst_n;

   logic axi_rst_n;

   initial begin
      rst_125m_n  = 1'b0;
      rst_62m5_n  = 1'b0;
      axi_rst_n   = 1'b0;
      gn_tb_reset = 1'b1;
      #80ns;
      rst_125m_n  = 1'b1;
      rst_62m5_n  = 1'b1;
      axi_rst_n   = 1'b1;
   end

   typedef struct {
      logic        ready;
      logic        valid;
      logic [ 1:0] id;
      logic [63:0] addr;
      logic [ 7:0] len;
      logic [ 2:0] size;
      logic [ 1:0] burst;
      logic [ 3:0] cache;
      logic [ 3:0] qos;
   } t_axi4_Ax_chnl;

   typedef struct {
      logic        ready;
      logic        valid;
      logic        last;
      logic [ 3:0] strb;
      logic [31:0] data;
   } t_axi4_W_chnl;

   typedef struct {
      logic        ready;
      logic        valid;
      logic [ 1:0] id;
      logic [ 1:0] resp;
   } t_axi4_B_chnl;

   typedef struct {
      logic        ready;
      logic        valid;
      logic [ 1:0] id;
      logic [ 1:0] resp;
      logic [31:0] data;
      logic        last;
   } t_axi4_R_chnl;



   t_axi4_Ax_chnl dut_if_aw;
   t_axi4_W_chnl  dut_if_w;
   t_axi4_B_chnl  dut_if_b;
   t_axi4_Ax_chnl dut_if_ar;
   t_axi4_R_chnl  dut_if_r;

   typedef struct {
      logic [31:0] dat;
      logic [ 1:0] resp;
   } td_rd_resp;

   typedef struct {
      td_rd_resp rdat_q [$:256];
      logic [ 1:0] id;
      logic [63:0] addr;
      logic [ 7:0] len;
      bit          check_data;
      logic  [1:0] exp_resp;
   } td_axi_rd_to_chk;


   td_axi_rd_to_chk axi_rd_checking_queue [$];
   integer axi_rdat_checked_ok_count;
   integer axi_rdat_checked_bad_count;
   integer last_axi_rd_check_data_words;
   integer axi_rd_Data_check_q_sz;
   event   axi_read_processed;
   event   axi_rd_procvessing_Start;
   event   host_mem_rd_done;


   uint32_t axi4b_backpressure_rand;
   uint32_t axi4b_backpressure_cmp;

   uint32_t axi4r_backpressure_rand;
   uint32_t axi4r_backpressure_cmp;

   // this block randomises the backporessure on the axi4_b_channel and axi4_r_chnnel..
   initial begin
      dut_if_aw.valid         = 1'b0;
      dut_if_w.valid          = 1'b0;
      dut_if_ar.valid         = 1'b0;
      axi4b_backpressure_cmp  = 5;
      axi4b_backpressure_rand = 10;
      axi4r_backpressure_cmp  = 5;
      axi4r_backpressure_rand = 10;

      forever begin
         @(posedge axi_clk);
         #1;
         if (axi4b_backpressure_cmp >= $urandom_range(1,axi4b_backpressure_rand)) begin
            dut_if_b.ready <= 1'b1;
         end else begin
            dut_if_b.ready <= 1'b0;
         end
         if (axi4r_backpressure_cmp >= $urandom_range(1,axi4r_backpressure_rand)) begin
            dut_if_r.ready <= 1'b1;
         end else begin
            dut_if_r.ready <= 1'b0;
         end
      end
   end



   IGN4124PCIMaster i_gn4124 ();

     wrapped_xwb_gn4124_core #(
        .g_WITH_AXI (p_AXI_SL),
        .g_WITH_DMA (p_WB_DMA)
     ) DUT (
          .rst_n_a_i          (i_gn4124.rst_n & gn_tb_reset),
          .p2l_clk_p_i        (i_gn4124.p2l_clk_p),
          .p2l_clk_n_i        (i_gn4124.p2l_clk_n),
          .p2l_data_i         (i_gn4124.p2l_data),
          .p2l_dframe_i       (i_gn4124.p2l_dframe),
          .p2l_valid_i        (i_gn4124.p2l_valid),
          .p2l_rdy_o          (i_gn4124.p2l_rdy),
          .p_wr_req_i         (i_gn4124.p_wr_req),
          .p_wr_rdy_o         (i_gn4124.p_wr_rdy),
          .rx_error_o         (i_gn4124.rx_error),
          .vc_rdy_i           (i_gn4124.vc_rdy),
          .l2p_clk_p_o        (i_gn4124.l2p_clk_p),
          .l2p_clk_n_o        (i_gn4124.l2p_clk_n),
          .l2p_data_o         (i_gn4124.l2p_data),
          .l2p_dframe_o       (i_gn4124.l2p_dframe),
          .l2p_valid_o        (i_gn4124.l2p_valid),
          .l2p_edb_o          (i_gn4124.l2p_edb),
          .l2p_rdy_i          (i_gn4124.l2p_rdy),
          .l_wr_rdy_i         (i_gn4124.l_wr_rdy),
          .p_rd_d_rdy_i       (i_gn4124.p_rd_d_rdy),
          .tx_error_i         (i_gn4124.tx_error),
          .dma_irq_o          (dma_irq),
          .irq_p_i            (1'b0),
          .irq_p_o            (gn4124_irq),
          .status_o           (gn4124_core_status),
          .wb_master_clk_i    (clk_125m),
          .wb_master_rst_n_i  (rst_125m_n),
          .wb_master_i        (wb_in),
          .wb_master_o        (wb_out),
          .wb_dma_cfg_clk_i   (clk_125m),
          .wb_dma_cfg_rst_n_i (rst_125m_n),
          .wb_dma_cfg_i       (wb_out),
          .wb_dma_cfg_o       (wb_in),
          .wb_dma_dat_clk_i   (wb_dma_clk),
          .wb_dma_dat_rst_n_i (wb_dma_rst_n),
          .wb_dma_dat_i       (wb_dma_in),
          .wb_dma_dat_o       (wb_dma_out),
          .axi_rst_n_i        (axi_rst_n),
          .axi_clk_i          (axi_clk),
          .axi4s_aw_ready_o   (dut_if_aw.ready),
          .axi4s_aw_valid_i   (dut_if_aw.valid),
          .axi4s_aw_addr_i    (dut_if_aw.addr),
          .axi4s_aw_id_i      (dut_if_aw.id),
          .axi4s_aw_len_i     (dut_if_aw.len),
          .axi4s_aw_size_i    (dut_if_aw.size),
          .axi4s_aw_burst_i   (dut_if_aw.burst),
          .axi4s_aw_cache_i   (dut_if_aw.cache),
          .axi4s_aw_qos_i     (dut_if_aw.qos),
          .axi4s_w_ready_o    (dut_if_w.ready),
          .axi4s_w_valid_i    (dut_if_w.valid),
          .axi4s_w_data_i     (dut_if_w.data),
          .axi4s_w_strobe_i   (dut_if_w.strb),
          .axi4s_w_last_i     (dut_if_w.last),
          .axi4s_b_ready_i    (dut_if_b.ready),
          .axi4s_b_valid_o    (dut_if_b.valid),
          .axi4s_b_id_o       (dut_if_b.id),
          .axi4s_b_resp_o     (dut_if_b.resp),
          .axi4s_ar_ready_o   (dut_if_ar.ready),
          .axi4s_ar_valid_i   (dut_if_ar.valid),
          .axi4s_ar_addr_i    (dut_if_ar.addr),
          .axi4s_ar_id_i      (dut_if_ar.id),
          .axi4s_ar_len_i     (dut_if_ar.len),
          .axi4s_ar_size_i    (dut_if_ar.size),
          .axi4s_ar_burst_i   (dut_if_ar.burst),
          .axi4s_ar_cache_i   (dut_if_ar.cache),
          .axi4s_ar_qos_i     (dut_if_ar.qos),
          .axi4s_r_ready_i    (dut_if_r.ready),
          .axi4s_r_valid_o    (dut_if_r.valid),
          .axi4s_r_id_o       (dut_if_r.id),
          .axi4s_r_resp_o     (dut_if_r.resp),
          .axi4s_r_data_o     (dut_if_r.data),
          .axi4s_r_last_o     (dut_if_r.last)
          );

   assign wb_dma_clk   = clk_125m;
   assign wb_dma_rst_n = rst_125m_n;

   xwb_dpram #
     (
      .g_size                  (16384),
      .g_slave1_interface_mode (1), // 1 = PIPELINED
      .g_slave2_interface_mode (1),
      .g_slave1_granularity    (1), // 1 = WORD
      .g_slave2_granularity    (1)
      )
   MEM (
        .rst_n_i   (wb_dma_rst_n),
        .clk_sys_i (wb_dma_clk),
        .slave1_i  (wb_dma_out),
        .slave1_o  (wb_dma_in),
        .slave2_i  (wb_mem_out),
        .slave2_o  (wb_mem_in)
        );

   CBusAccessor acc;


   task val_check(string name, uint32_t addr, val, expected, mask);
      if ((val & mask) != (expected & mask))
        begin
           $display();
           $display("Simulation FAILED");
           $fatal(1, "%s error at address 0x%.2x. Expected 0x%.8x, got 0x%.8x, mask 0x%08x",
                  name, addr, expected, val, mask);
        end
   endtask // val_check


   task reg_check(uint32_t addr, expected, mask = 32'hffffffff);
      uint64_t val;
      acc.read(addr, val);
      val_check("Register read-back", addr, val, expected, mask);
   endtask // reg_check


   task mem_check(uint32_t addr, expected);
      uint64_t val;
      i_gn4124.host_mem_read(addr, val);
      val_check("Memory read-back", addr, val, expected, 32'hffffffff);
   endtask // reg_check

   // this task checks that the IRQ is asserted (and DMA state is idle)
   task check_irq_status;
      //  Check irq status
      reg_check('h04, 'h04, 'h0f);
      if (dma_irq != 1'b1)
        $fatal(1, "dma irq should be 1");
   endtask


   task clear_irq;
      acc.write('h04, 'h04);
      reg_check('h04, 'h00, 32'h0f);
      if (dma_irq != 1'b0)
        $fatal(1, "dma irq should be 0");
   endtask





   task t_setup_dma( input [63:0] host_addr,
                     input [31:0] local_addr,
                     input [31:0] length,
                    // input [31:0] attributes,
                     input [63:0] next_chained_descr_addr,
                     input bit more_in_chain,
                     input bit w_nrd_sel);
      automatic logic [31:0] attributes;
      attributes = 'h0;
      attributes[1] = more_in_chain;    // doc has these the wrong way round!!!
      attributes[0] = w_nrd_sel;        // doc has these the wrong way round!!!
      acc.write('h08, local_addr[ 0+:32]);                // carrier local addr
      acc.write('h0c, host_addr[ 0+:32]);                 // hstartL
      acc.write('h10, host_addr[32+:32]);                 // hstartH
      acc.write('h14, length);                            // count
      acc.write('h18, next_chained_descr_addr[ 0+:32]);   // nextL
      acc.write('h1C, next_chained_descr_addr[32+:32]);   // nextH
      acc.write('h20, attributes);                              // attrib
   endtask


   task t_start_dma();
      acc.write('h00, 'h01); // start
//      bits 3:2 are byte_ordering
//      bit 1 is "abort"
// bit 0 is tart
   endtask


   task t_add_chaining_descriptor(
                     input [63:0] host_addr,
                     input [31:0] local_addr,
                     input [31:0] length,
                     //input [31:0] attributes,
                     input [63:0] next_chained_descr_addr,
                     input [63:0] start_address_in_host_memory,   // note the verlog wrapper adds 'h2000_0000
                     input        more_in_chain,
                     input        w_nrd_sel);
      logic [31:0] attributes;
      attributes = 'h0;
      attributes[0] = w_nrd_sel;
      attributes[1] = more_in_chain;
      i_gn4124.host_mem_write_noaddr_twiddles(start_address_in_host_memory + 'h0000, local_addr); // remote address
      i_gn4124.host_mem_write_noaddr_twiddles(start_address_in_host_memory + 'h0004, host_addr[ 0+:32]); // hstartL
      i_gn4124.host_mem_write_noaddr_twiddles(start_address_in_host_memory + 'h0008, host_addr[32+:32]); // hstartH
      i_gn4124.host_mem_write_noaddr_twiddles(start_address_in_host_memory + 'h000C, length); // count
      i_gn4124.host_mem_write_noaddr_twiddles(start_address_in_host_memory + 'h0010, next_chained_descr_addr[ 0+:32]); // nextL
      i_gn4124.host_mem_write_noaddr_twiddles(start_address_in_host_memory + 'h0014, next_chained_descr_addr[32+:32]); // nextH
      i_gn4124.host_mem_write_noaddr_twiddles(start_address_in_host_memory + 'h0018, attributes); // attrib
   endtask




   task automatic t_get_device_mem(ref uint32_t wrd_addr, output [31:0] data);
      data[ 0+:8] = MEM.U_DPRAM.gen_splitram.U_RAM_SPLIT.ram0[wrd_addr];
      data[ 8+:8] = MEM.U_DPRAM.gen_splitram.U_RAM_SPLIT.ram1[wrd_addr];
      data[16+:8] = MEM.U_DPRAM.gen_splitram.U_RAM_SPLIT.ram2[wrd_addr];
      data[24+:8] = MEM.U_DPRAM.gen_splitram.U_RAM_SPLIT.ram3[wrd_addr];
   endtask


   task automatic t_blat_device_mem(input uint32_t wrd_addr, input [31:0] data);
     MEM.U_DPRAM.gen_splitram.U_RAM_SPLIT.ram0[wrd_addr] = data[ 0+:8];
     MEM.U_DPRAM.gen_splitram.U_RAM_SPLIT.ram1[wrd_addr] = data[ 8+:8];
     MEM.U_DPRAM.gen_splitram.U_RAM_SPLIT.ram2[wrd_addr] = data[16+:8];
     MEM.U_DPRAM.gen_splitram.U_RAM_SPLIT.ram3[wrd_addr] = data[24+:8];
   endtask



   task automatic t_compare_memories(input [63:0]    pcie_start_addr,   // note needs to be 8B aligned
                                     input [31:0]    wb_start_addr,
                                     input  int      words_to_compare,
                                     output uint32_t bad_comparisons,
                                     output uint32_t good_comparisons,
                                     input  bit      verbose,
                                     input  bit      no_twiddles = 1'b0);
   uint32_t     addr_cnt;
   uint64_t     system_dat;
   logic [31:0] wb_mem_dat;
   logic [31:0] difference;
   bit          error_detected;
   uint32_t     wb_addr;
   uint64_t     pcie_addr;

   bad_comparisons  = 0;
   good_comparisons = 0;

   if (verbose) begin
      $write("Checking Expected memory contents against what is in the wishbone target memory\n");
      $write("________________________________________________________________________________\n");
      $write("| Addr (byte) | wishbone   | mem content |     PCIE_ADDR      |  error |  cnt  |\n");
      $write("|-------------+------------+-------------+--------------------+--------+-------|\n");
   end

   for (addr_cnt = 'h00; addr_cnt < words_to_compare; addr_cnt += 1)begin
      wb_addr   = wb_start_addr/4   + addr_cnt ;
      pcie_addr = pcie_start_addr + (addr_cnt * 4);

      t_get_device_mem(       wb_addr,  wb_mem_dat);

      if (1'b1 == no_twiddles ) begin
         i_gn4124.host_mem_read_noaddr_twiddles(pcie_addr, system_dat);
      end else begin
         i_gn4124.host_mem_read(pcie_addr, system_dat);
      end

      difference = wb_mem_dat ^ system_dat[0+:32];

       if (difference !== 32'h0) begin
         error_detected = 1'b1;
      end else begin
         error_detected = 1'b0;
      end

      if (1'b1 ==  error_detected) begin
         bad_comparisons++;
      end else begin
         good_comparisons++;
      end

      if (verbose) begin
         $write("| 0x:%08x | 0x%08x | 0x%08x  | 0x%016x |   %1b    | %05d |",
                  4* wb_addr, wb_mem_dat,
                                      system_dat[31:0],
                                                  pcie_addr,
                                                             error_detected,
                                                                       addr_cnt
               );
         if (1'b1 ==  error_detected) begin
            $write(" *** mismatch 0x%08x\n", difference);
         end else begin
            $write("\n");
         end
      end
   end

   if (verbose) begin
      $write("------------------------------------------------------------------------------\n ");
   end
   endtask






   task t_send_aw( input logic [ 1:0] id,
                   input logic [63:0] addr,
                   input logic [ 7:0] len,
                   input logic [ 2:0] size  = 3'h2,
                   input logic [ 1:0] burst = 2'h1,
                   input logic [ 3:0] cache = 4'h0,
                   input logic [ 3:0] qos   = 4'h0
                 );
                 @(posedge axi_clk);
                 #1;
      dut_if_aw.valid <= 1'b1;
      dut_if_aw.id    <= id;
      dut_if_aw.addr  <= addr;
      dut_if_aw.len   <= len;
      dut_if_aw.size  <= size;
      dut_if_aw.burst <= burst;
      dut_if_aw.cache <= cache;
      dut_if_aw.qos   <= qos;

      @(posedge axi_clk);
      while (dut_if_aw.ready !== 1'b1) begin
         @(posedge axi_clk);
      end
      #1;
      dut_if_aw.valid <= 1'b0;
   endtask


   task t_send_w(input logic       last,
                 input logic [3:0] strb,
                 input logic[31:0] data
                );
      dut_if_w.valid <= 1'b1;
      dut_if_w.last  <= last;
      dut_if_w.strb  <= strb;
      dut_if_w.data  <= data;
      @(posedge axi_clk);
      while (1'b0 === dut_if_w.ready) begin
         @(posedge axi_clk);
       //$display("spinning at %t time", $time);
      end
      dut_if_w.valid <= 1'b0;
  //   $display("done_wdat_phs");
   endtask


   task t_get_b(output logic [1:0] resp,
                output logic [1:0] id
               );
      @(posedge axi_clk);
      while ((dut_if_b.valid !== 1'b1) || (dut_if_b.ready !== 1'b1)) begin
         @(posedge axi_clk);
      end
         resp = dut_if_b.resp;
         id   = dut_if_b.id;

   endtask


   task t_send_ar( input logic [ 1:0] id,
                   input logic [63:0] addr,
                   input logic [ 7:0] len,
                   input logic [ 2:0] size,
                   input logic [ 1:0] burst,
                   input logic [ 3:0] cache,
                   input logic [ 3:0] qos  = 4'h0
                 );
                 @(posedge axi_clk);
                 #1;
      dut_if_ar.valid <= 1'b1;
      dut_if_ar.id    <= id;
      dut_if_ar.addr  <= addr;
      dut_if_ar.len   <= len;
      dut_if_ar.size  <= size;
      dut_if_ar.burst <= burst;
      dut_if_ar.cache <= cache;
      dut_if_ar.qos   <= qos;
      @(posedge axi_clk);
      while (dut_if_ar.ready !== 1'b1) begin
         @(posedge axi_clk);
      end
      #1;
      dut_if_ar.valid <= 1'b0;
   endtask



   event start_host_mem_read;

   initial begin
      td_axi_rd_to_chk rd_being_checked;

      axi_rdat_checked_ok_count  = 0;
      axi_rdat_checked_bad_count = 0;

      forever begin
         if (axi_rd_checking_queue.size() == 0) begin
            @(posedge axi_clk);
         end else begin
            logic [63:0] mem_addr;
            integer index_being_checked;
            logic [1:0] merged_resp;
            bit seen_err_in_this_read;

            -> axi_rd_procvessing_Start;
            //$display("starting checking");
            seen_err_in_this_read = 1'b0;
            rd_being_checked = axi_rd_checking_queue[0];

            if (rd_being_checked.len+1 !== rd_being_checked.rdat_q.size()) begin
                        $display("Unexpected number of rdat words, expected 0d:%03d, got 0d:%03d", rd_being_checked.len+1, rd_being_checked.rdat_q.size());
                        $fatal();
            end

           // TBD: how do we handle the resp value changing during the read to check...    is this appropriate?
           index_being_checked = 0;
           merged_resp = 2'h0;
           while (index_being_checked < rd_being_checked.rdat_q.size()) begin
              merged_resp = merged_resp | rd_being_checked.rdat_q[index_being_checked].resp;
              index_being_checked++;
           end

            if (rd_being_checked.exp_resp !== merged_resp) begin
               $display("Unexpected r_resp, expected 0d:%01d, got 0d:%01d (merged)", rd_being_checked.exp_resp, merged_resp);
               $fatal();
            end



           if (1'b1 == rd_being_checked.check_data) begin

               mem_addr           = rd_being_checked.addr;
               index_being_checked = 0;
               // now check the read data matches what is in the host memory...
               while (index_being_checked !== rd_being_checked.rdat_q.size()) begin
                  uint64_t  host_mem_moddeled_dat;
                  -> start_host_mem_read;
                  i_gn4124.host_mem_read_noaddr_twiddles(mem_addr , host_mem_moddeled_dat);
                  -> host_mem_rd_done;
                  if (host_mem_moddeled_dat[31:0] !==  rd_being_checked.rdat_q[index_being_checked].dat) begin
                           $display("bad_Axi_rdat_data, expected 0x:%08x, got 0d:%08x for addr 0x:%016x", host_mem_moddeled_dat[31:0], rd_being_checked.rdat_q[index_being_checked].dat, mem_addr);
                           $fatal();
                // end else begin
                //    $display("checked work %d of %d was OK",index_being_checked, rd_being_checked.rdat_q.size());
                  end

                  mem_addr = mem_addr +4;
                  index_being_checked++;
               end
               //$display("completed data_check");
            end else begin
              // $display("skipped data_check");
            end


            while (0 !== rd_being_checked.rdat_q.size()) begin
               void'(rd_being_checked.rdat_q.pop_front());
            end

            last_axi_rd_check_data_words = rd_being_checked.len+1;

            void'(axi_rd_checking_queue.pop_front());

            if (1'b1 == seen_err_in_this_read) begin
               axi_rdat_checked_bad_count++;
          //     $display("axi read was bad!");
            end else begin
               axi_rdat_checked_ok_count ++;
           //    $display("axi read was good!");
            end
            -> axi_read_processed;

         end
      end
   end


   initial begin
      forever begin
         @(posedge axi_clk);
         axi_rd_Data_check_q_sz <= axi_rd_checking_queue.size();
      end
   end




   task automatic t_check_axi_rd_Dat(input logic [ 1:0] id,
                                     input logic [63:0] addr,
                                     input logic [ 7:0] len,
                                     input logic  [1:0] expected_resp = 2'h0,
                                     input bit          check_data    = 1'b1,
                                     input integer      timeout
                                     );
     // logic [31: 0] rdat_q [$:256];    // the [$:256] sets a max size bound...
      bit got_last;

      td_axi_rd_to_chk collected_rd;
      td_rd_resp rd_dat_struct;

      got_last = 1'b0;

      collected_rd.id         = id;
      collected_rd.addr       = addr;
      collected_rd.len        = len;
      collected_rd.check_data = check_data;
      collected_rd.exp_resp   = expected_resp;

      fork
         // timeout
         begin
            int timeout_count = 0;

            while ((timeout_count++ < timeout) & (1'b0 == got_last)) begin
               @(posedge axi_clk);
            end
            if (1'b0 == got_last) begin
                  $display("timed out while waiting for a the axi_R_channel.last bit to be asserted during a transfer! there are %d items in the queue", collected_rd.rdat_q.size());
                  $fatal();
            end
         end


         // pop rdata_items into queue untill last is asserted
         begin
         //   $display("started_polling");
            while (got_last == 1'b0) begin
               @(posedge axi_clk);
               if ( (1'b1 === dut_if_r.valid) & (1'b1 === dut_if_r.ready)) begin

                  rd_dat_struct.dat  = dut_if_r.data;
                  rd_dat_struct.resp = dut_if_r.resp;
                  if (id !== dut_if_r.id) begin
                     $display("Unexpected read_id, expected 0b:%02b, got 0b:%02b", id, dut_if_r.id);
                     $fatal();
                  end
                  // TODO check RESP code!
                  collected_rd.rdat_q.push_back(rd_dat_struct);
                  if (1'b1 == dut_if_r.last) begin
                     got_last = 1'b1;
                  end
                  //$display("processing_data_this_clk");
               end else begin
                  //$display("no data available ,skipping a clk");
               end
            end
          //  $display("finished_polling_rdat");
         end
      join_any


      axi_rd_checking_queue.push_back(collected_rd);
   endtask


   task t_assert_gennum_reset();
      if (gn4124_core_status[0] === 1'b1) begin
         @(negedge i_gn4124.p2l_clk_p);
         gn_tb_reset  = 1'b0;
         rst_125m_n   = 1'b0;
         rst_62m5_n   = 1'b0;
         axi_rst_n    = 1'b0;
         repeat (5) @(posedge i_gn4124.p2l_clk_p);
         @(negedge i_gn4124.p2l_clk_p);
         // check the genum pll is unlocked!
         if (gn4124_core_status[0] !== 1'b0) begin
            $display("hmm, TB expected that the gennum PLL should have been unlocked!");
            $fatal();
         end
         gn_tb_reset  = 1'b1;

         repeat (25) @(posedge i_gn4124.p2l_clk_p);

         // make sure we see some clocks to reset the logic!
         rst_125m_n   = 1'b1;
         rst_62m5_n   = 1'b1;
         axi_rst_n    = 1'b1;

         if (gn4124_core_status[0] !== 1'b1) begin
            $display("hmm, TB expected that the gennum PLL should have been locked again!");
            $fatal();
         end

         repeat (4) @(posedge i_gn4124.p2l_clk_p);
      end else begin
        $display("hmm, TB expected that the gennum PLL should have been locked!");
        $fatal();
      end

   endtask



  `include "testcases.svh"


// ************************************************************************************************
// * Main body of the testbench that acutally drives and tests the DUT
// ************************************************************************************************
   initial begin
      automatic int ntest = 1;
      automatic int tests = 0;

      // set the total number of tests
      if (p_WB_DMA === 1) begin
         tests += 15;
      end
      if (p_AXI_SL === 1) begin
         tests += 5;
      end
      if ((p_WB_DMA === 1) & (p_AXI_SL === 1)) begin
         tests += 1;
      end

      @(posedge i_gn4124.ready);

      acc = i_gn4124.get_accessor();

      acc.set_default_xfer_size(4);

      @(posedge clk_125m);

      // *******************************************************************************************
      //  IMPORTANT NOTES ON TESTCASES!
      // *******************************************************************************************
      // Each test is contained in the testcases.svh file, and SHOULD run atomically without needing
      // memory contents or configuration from previous tests.
      // Ideally this would also include configuring the BFM for each test & resets etc but that
      //    was a stage too far!

      // basic tests...
          //   TODO: add some infra / generate so that we havea target to read / write to if DMA isn't present!
          // we really should set something up to validate that the core works with DMA && AXI turned off!!


      // Wishbone DMA tests
      if (1 === p_WB_DMA) begin
        t_test_simple_wishbone_rw( ntest, tests);     // TESTS DMA registers...
        t_test_unchecked_dma_read_aborted(ntest, tests);
        t_test_256B_dma_write(ntest, tests);
        t_test_2x128B_reads(ntest, tests);
        t_test_256B_read(ntest, tests);
        t_test_dma_rd_with_wishbone_stalls(ntest, tests);
        t_test_8KiB_chained_DMA_WRITE(ntest, tests);
        t_test_8KiB_Byteswaps(ntest, tests);
        t_test_32bit_addr_overflow_rd(ntest, tests);
        t_test_mixed_chaining(ntest, tests);
        t_test_read_chaining(ntest, tests);
        t_test_4K_boundary_rw(ntest, tests);
      end


      // AXI tests
      if (1 === p_AXI_SL) begin
        t_axi_w_directed_tests(ntest, tests);
        t_axi_w_b2b_tests(ntest, tests);
        t_axi_w_error_tests(ntest, tests);
        t_axi_r_directed_tests(ntest, tests);
        t_axi_r_error_tests(ntest, tests);
        t_test_axi_RD_backpressure(ntest, tests);
      end

      if ((1 === p_WB_DMA) & (1 === p_AXI_SL)) begin
        t_all_interfaces_active(ntest, tests);
      end

// disabled tests WB
      //t_test_32bit_addr_overflow_wr(ntest, tests);
          /// TBD this test should NOT pass!:
          //            it demonstrates a 4KB Addressing boundary crossed!!!! (PCIE SPEC VIOLATION)
          //            and a 32-bit => 64bit addr.... (part of the read needs a 64B addr cycle)
          // This can be masked in Software (due to sensible addresses => DMA),
          //           but it was fixed for the other (DMA RD) direction...



      $display();
      $display("Simulation PASSED");

      $finish;
   end // initial begin

endmodule // main
