//------------------------------------------------------------------------------
// CERN BE-CO-HT
// GN4124 core for PCIe FMC carrier
// http://www.ohwr.org/projects/gn4124-core
//------------------------------------------------------------------------------
//
// unit name: "testcases"
//
// description: This file contains the testaces intended to be 'included into the
//  main.sv testbench,
//
//------------------------------------------------------------------------------
// Copyright CERN 2021
//------------------------------------------------------------------------------
// This source describes Open Hardware and is licensed under the CERN-OHL-W v2.
// You may redistribute and modify this source and make products using it
// under the terms of the CERN-OHL-W v2 or future versions (https://ohwr.org/cern_ohl_w_v2.txt).
//
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED
// WARRANTY, INCLUDING OF MERCHANTABILITY, SATISFACTORY
// QUALITY AND FITNESS FOR A PARTICULAR PURPOSE. Please see
// the CERN-OHL-W v2 for applicable conditions.
//
//------------------------------------------------------------------------------




// -------------------------------------------
//  TESTCASE:    Simple wishbone RW rest.
//      Purpose:
//            Sanity checking that we have access to wihbone CSR functioning...
//      Method:
//         Use 2 of the addr registers in the DMA control register space as a scratchpad to test read and writes
//         Passes if readback matches expected value.
// --------------------------------------------
   task automatic t_test_simple_wishbone_rw(ref   int testno,
                                            input int num_tests
                                           );
      uint32_t addr, val, expected;

      @(posedge clk_125m);

      // ---------------------------------
      $write("Test %0d/%0d: simple read/write accesses over Wishbone: ",
             testno++, num_tests);
      testphase = "simple read/write accesses over Wishbone";

      // Verify simple read/writes over wishbone
      reg_check('h0, 'h0);

      // use registers as scratchpad to test RW accesses...
      //    EXTEND to validate BE functionality..
      acc.write('h0c, 'hffacce55);
      acc.write('h10, 'h1badcafe);

      reg_check('h0c, 'hffacce55);
      reg_check('h10, 'h1badcafe);

      // Reset all DMA config registers
      for (addr = 'h00; addr <= 'h20; addr += 4)
        begin
           acc.write(addr, 'h0);
        end

      reg_check('h0c, 'h0);
      reg_check('h10, 'h0);

      repeat(2) @(posedge clk_125m);

      $write("PASS\n");
   endtask



// -------------------------------------------
//  TESTCASE:    t_TEST_unchecked_dma_read_aborted
//      Purpose:
//            Sanity checking DMA reads and abort functinoality.
//      Method:
//          NOTE: NOT SELF CHECKING!!!!!!!!!!!!!
// --------------------------------------------
   task automatic t_test_unchecked_dma_read_aborted(ref   int ntest,
                                                    input int tests
                                                   );
      testphase = "128B read over DMA, abort after first read: ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);

      if (dma_irq != 1'b0)
        $fatal(1, "dma irq should be 0");

      t_setup_dma(.host_addr               ('h2000_0000),
                  .local_addr              ('h00),
                  .length                  ('h80),
                  .more_in_chain           (0),
                  .w_nrd_sel               (0),
                  .next_chained_descr_addr ('hdeadbeefbeefbeef));
      t_start_dma();

      // wait for transfer to start
      @(posedge i_gn4124.l2p_valid);

      repeat(2) @(posedge clk_125m);

      // Test abort feature
      acc.write('h00, 'h02);
      reg_check('h04, 'h03, 32'h0f);
      acc.write('h00, 'h00);

      repeat(2) @(posedge clk_125m);

      $write("PASS\n");

   endtask



// -------------------------------------------
//  TESTCASE:    t_test_256B_dma_write
//      Purpose:
//           BASic self-checking DMA write (Card to sys_mem)
//      Method:
//          NOTE: self cehcking of memory contents...
// --------------------------------------------
   task automatic t_test_256B_dma_write(ref   int ntest,
                                         input int tests
                                         );
      uint32_t addr,expected, from_mem ;
      bit      error_detected;
      uint32_t badcomp, goodcomp;

      logic [31:0] rand_wdat = $urandom();


   //   rand_wdat = 32'h80000020;

      testphase = "test_256B_dma_write: ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);


      // Setup data in BFM memory
      for (addr = 'h00; addr < 'h40; addr += 1)
        i_gn4124.host_mem_write(4 * addr, rand_wdat - addr);

       //  Setup & Start the DMA
      t_setup_dma(.host_addr               ('h2000_0000),
                  .local_addr              ('h00),
                  .length                  ('h100),
                  .more_in_chain           (0),
                  .w_nrd_sel               (1),
                  .next_chained_descr_addr ('hdeadbeefbeefbeef));

      t_start_dma();

      @(posedge dma_irq);
      check_irq_status;
      clear_irq;

      repeat(4) @(posedge clk_125m);

      error_detected = 1'b0;
      for (addr = 'h00; addr < 'h40; addr += 1) begin
           expected = rand_wdat - addr;
           t_get_device_mem(addr, from_mem);
           if (from_mem !== expected) begin
              error_detected = 1'b1;
            end
       end

       if (error_detected)
         t_compare_memories('h00000000, 'h0, 64, badcomp, goodcomp, 1);

       if (error_detected != 1'b1) begin
            $write("PASS \n");
       end else begin
            $write("FAIL  -- IRQ was OK But data integrity failed!!!\n");
       end
      #1us;
   endtask








// -------------------------------------------
//  TESTCASE:    t_test_2x128B_reads
//      Purpose:
//           BASic self-checking DMA read (sys_mem to card)
//      Method:
//          NOTE: self checking of memory contents...
// --------------------------------------------
   task automatic t_test_2x128B_reads(ref   int ntest,
                                         input int tests
                                         );
      uint32_t addr,expected, from_mem ;
      bit      error_detected;
      logic [31:0] rand_wdat = $urandom();


      testphase = "test_2x128B reads: ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);

     // rand_wdat = 32'h80000020;

      // TBD: make this "even more" randomised data?
      for (addr = 'h00; addr < 'h40; addr += 1)
        t_blat_device_mem(addr, rand_wdat - addr);

      t_add_chaining_descriptor(.host_addr                   ('h2000_0080),
                                .local_addr                  ('h80),
                                .length                      ('h80),
                                .next_chained_descr_addr     ('h0),
                                .start_address_in_host_memory('h20020000),
                                .more_in_chain               (1'b0),
                                .w_nrd_sel                   (1'b0));

      t_setup_dma(.host_addr               ('h2000_0000),
                  .local_addr              ('h00),
                  .length                  ('h80),
                  .more_in_chain           (1),
                  .w_nrd_sel               (0),
                  .next_chained_descr_addr ('h20020000));

      t_start_dma();

      @(posedge dma_irq);

      check_irq_status;
      clear_irq;

      for (addr = 'h00; addr < 'h40; addr += 1)
        begin
           expected = rand_wdat - addr;
           mem_check(4 * addr, expected);         /// TODO: is this the correct type of memory checking!!!!
        end

      repeat(4) @(posedge clk_125m);

      $write("PASS\n");
   endtask




// -------------------------------------------
//  TESTCASE:    t_test_256B_read
//      Purpose:
//           BASic self-checking DMA read (sys_mem to card)
//      Method:
//          NOTE: self checking of memory contents...
// --------------------------------------------
   task automatic t_test_256B_read(ref   int ntest,
                                         input int tests
                                         );
      uint32_t addr,expected, from_mem ;
      bit      error_detected;

      logic [31:0] rand_wdat = $urandom();


      testphase = "test_256B_read : ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);

      for (addr = 'h00; addr < 'h40; addr += 1) begin
        t_blat_device_mem(addr, rand_wdat - addr*7);
      end

      t_setup_dma(.host_addr               ('h2000_0000),
                  .local_addr              ('h00),
                  .length                  ('h100),
                  .more_in_chain           (0),
                  .w_nrd_sel               (0),
                  .next_chained_descr_addr ('hffeff_0123_4567_19a0));

      t_start_dma();

      @(posedge dma_irq);

      check_irq_status;
      clear_irq;

      for (addr = 'h00; addr < 'h40; addr += 1)
        begin
           expected = rand_wdat - addr*7;
           mem_check(4 * addr, expected);
        end

      repeat(4) @(posedge clk_125m);

      $write("PASS\n");
   endtask




// -------------------------------------------
//  TESTCASE:    t_test_dma_rd_with_wishbone_stalls
//      Purpose:
//            check the core WRT stalls on the wichbone bus
//      SELF_CHECKING:
//            GROSS check - targets are pre-used and data is non-unique
//      Method:
//      Notes:
//
// --------------------------------------------
task automatic t_test_dma_rd_with_wishbone_stalls(ref   int ntest,
                                                  input int tests
                                                 );

      uint32_t addr, val, expected;
      int stall_cycle[14];

      logic [31:0] rand_wdat = $urandom();

      testphase = "256B DMA reads with wishbone stalling: ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);


      for (addr = 'h00; addr < 'h40; addr += 1) begin
        t_blat_device_mem(addr, rand_wdat - addr);
      end


      t_setup_dma(.host_addr               ('h2000_0000),
                  .local_addr              ('h00),
                  .length                  ('h100),
                  .more_in_chain           (0),
                  .w_nrd_sel               (0),
                  .next_chained_descr_addr ('h20020000));

      stall_cycle = '{0,1,2,14,15,16,17,30,31,32,33,61,62,63};
      foreach(stall_cycle[i])
        begin
           //acc.write('h00, 'h01); // start

         $sformat(testphase, "256B DMA reads with wishbone stalling:  (stall_cycle = %0d)", stall_cycle[i]);

           t_start_dma();

           @(posedge wb_dma_out.stb);
           repeat(stall_cycle[i]) @(posedge clk_125m);
           force `UUT_GN4124.dma_stall_i = 1'b1;
           @(posedge clk_125m);
           force `UUT_GN4124.dma_ack_i = 1'b0;
           @(posedge clk_125m);
           release `UUT_GN4124.dma_stall_i;
           @(posedge clk_125m);
           release `UUT_GN4124.dma_ack_i;

           @(posedge dma_irq);

           check_irq_status;
           clear_irq;

           for (addr = 'h00; addr < 'h40; addr += 1)  begin
                expected = rand_wdat - addr;
                mem_check(4 * addr, expected);
           end

           repeat(4) @(posedge clk_125m);
        end

      $write("PASS\n");
endtask




// -------------------------------------------
//  TESTCASE:    t_test_2x4KiB_chained_DMA_WRITE
//      Purpose:
//            legacy: checks write chains...
//      SELF_CHECKING:
//            NONE - no timeout    does not even wait to confirm transfer finished...
//      Method:
//      Notes:
//
// --------------------------------------------
task automatic t_test_2x4KiB_chained_DMA_WRITE(ref   int ntest,
                                                  input int tests
                                                 );

      uint32_t addr;
      uint32_t val;
      uint32_t expected;

      // ---------------------------------
      testphase = "2x4KiB chained DMA write: ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);

      // Setup data in BFM memory
      for (addr = 'h00; addr < 'h800; addr += 1)
        i_gn4124.host_mem_write(4 * addr, 32'h80000020 - addr);

      t_add_chaining_descriptor(.host_addr                    ('h2000_1000),
                                .local_addr                   ('h1000),
                                .length                       ('h1000),
                                .next_chained_descr_addr      ('h0),
                                .start_address_in_host_memory ('h2002_0000),
                                .more_in_chain                (1'b0),
                                .w_nrd_sel                    (1'b1));

      t_setup_dma(.host_addr               ('h2000_0000),
                  .local_addr              ('h00),
                  .length                  ('h1000),
                  .more_in_chain           (0),
                  .w_nrd_sel               (1),
                  .next_chained_descr_addr ('h2002_0000));

      t_start_dma();

      @(posedge dma_irq);

      check_irq_status;
      clear_irq;

      // cehck data

      repeat(4) @(posedge clk_125m);

      $write("PASS\n");

      // wait for WB transfer to finish
      #5us;


endtask






 // -------------------------------------------
//  TESTCASE:    t_test_8KiB_chained_DMA_WRITE
//      Purpose:
//            legacy: checks write chains...
//      SELF_CHECKING:
//            NONE - no timeout    does not even wait to confirm transfer finished...
//      Method:
//      Notes:
//
// --------------------------------------------
task automatic t_test_8KiB_chained_DMA_WRITE(ref   int ntest,
                                                  input int tests
                                                 );

      uint32_t addr, val, expected, from_mem;
      bit      error_detected;
      uint32_t badcomp, goodcomp;

      logic [31:0] rand_wdat = $urandom();

      // ---------------------------------
      testphase = "8KiB chained DMA write: ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);


       // Setup data in BFM memory
      for (addr = 'h00; addr < 'h800; addr += 1) begin
        i_gn4124.host_mem_write(4 * addr, rand_wdat - addr);
      end


      t_add_chaining_descriptor(.host_addr                   (64'h20001800),
                                .local_addr                  (32'h1800),
                                .length                      ('h800),
                                .next_chained_descr_addr     (32'hDEADBEEf),
                                .start_address_in_host_memory('h2002_0000),
                                .more_in_chain               (1'b0),
                                .w_nrd_sel                   (1'b1));

      t_setup_dma(.host_addr               (64'h2000_0000),
                  .local_addr              ('h00),
                  .length                  ('h1800),
                  .more_in_chain           (1),
                  .w_nrd_sel               (1),
                  .next_chained_descr_addr (64'h20020000));

      t_start_dma();

      @(posedge dma_irq);

      check_irq_status;
      clear_irq;

      repeat(4) @(posedge clk_125m);

      // do I need to wait longer here???

      // Check data
      error_detected = 1'b0;
      for (addr = 'h00; addr < 'h800; addr += 1) begin
           expected = rand_wdat - addr;
           t_get_device_mem(addr, from_mem);
           if (from_mem !== expected) begin
              error_detected = 1'b1;
            end
       end

       if (error_detected)
         t_compare_memories('h00000000, 'h0, 2048, badcomp, goodcomp, 1);

       if (error_detected != 1'b1) begin
            $write("PASS \n");
       end else begin
            $write("FAIL  -- IRQ was OK But data integrity failed!!!\n");
       end

      #1us;
  endtask





// -------------------------------------------
//  TESTCASE:    t_test_8KiB_Byteswaps
//      Purpose:
//            tests byteswap behaviour
//      Method:
//          .....
// --------------------------------------------
   task automatic t_test_8KiB_Byteswaps(ref   int ntest,
                                        input int tests
                                       );
      uint32_t addr, val, expected, from_mem;
      bit      error_detected;
      uint32_t badcomp, goodcomp;

      // Ramdomise the write data for this testcase, that way a previous test can't cause a false +ve..
      logic [31:0] rand_wdat = $urandom();

      for (int i = 0; i < 4; i++) begin

         $sformat(testphase, "8KiB DMA read (byte swap = %0d) (rand = 0x%08x) : ",
                                                         i,     rand_wdat);
         $write("Test %0d/%0d: %s", ntest++, tests, testphase);


         for (addr = 'h00; addr < 'h800; addr += 1) begin
           t_blat_device_mem(addr, rand_wdat - addr);
         end


         t_setup_dma(.host_addr               ('h2000_0000),
                     .local_addr              ('h00),
                     .length                  ('h2000),
                     .more_in_chain           (0),
                     .w_nrd_sel               (0),
                     .next_chained_descr_addr ('hffeff_0123_4567_19a0));

         acc.write('h00, (i << 2) | 'h01); // start BUT byteswaps are applied...

         @(posedge dma_irq);

         check_irq_status;


         for (addr = 'h00; addr < 'h800; addr += 1)
           begin
              expected = rand_wdat - addr;
              if (i == 1)
                expected = {<<8{expected}};
              else if (i == 2)
                expected = {<<16{expected}};
              else if (i == 3)
                expected = {<<16{{<<8{expected}}}};
              mem_check(4 * addr, expected);
           end

         clear_irq;

         repeat(4) @(posedge clk_125m);

         $write("PASS\n");
          rand_wdat = $urandom();

         #1us;
      end
   endtask


// -------------------------------------------
//  TESTCASE:    t_test_mixed_chaining
//      Purpose:
//          The purpose of this testcase is to detect wishbone R->W and W->R turnarround issues from
//          the dma_controller state machine
//      Method:
//          NOTE: SELF_CHECKING: YES
//          We programa long-ish chain of alternating reads and writes, then apply backpressure on
//          the wishbone interface for long enough to cause it to fail...
//          as of 20/01/2021 it is failing (expected) ...
// --------------------------------------------
   task automatic t_test_mixed_chaining(ref   int ntest,
                                        input int tests
                                       );

      uint32_t     addr;
      uint32_t     expected;
      uint32_t     badcomp;
      uint32_t     goodcomp;
      logic [31:0] rand_wdat = $urandom();
      bit          dma_done;
      integer      cycles;
      uint32_t     op_len_bytes = 128;
      uint64_t     wdat;
      uint64_t     start_addr;

      testphase = "t_test_mixed_chaining: ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);

      // initialise memory
      for (addr = 'h00; addr < op_len_bytes/4; addr += 1) begin
        t_blat_device_mem(addr, rand_wdat - addr);
      end



      if (dma_irq != 1'b0) begin
        $fatal(1, "dma irq should be 0");
      end


      // First DMA copies tha data into system memory at offset 0x1000
      t_setup_dma(.host_addr               ('h2000_1000),
                    .local_addr              ('h00),
                    .length                  (op_len_bytes),
                    .more_in_chain           (1),
                    .w_nrd_sel               (0),
                    .next_chained_descr_addr ('h20020000));

      // Second DMA Transfer Wipes the region the data came from.
      t_add_chaining_descriptor(.start_address_in_host_memory(64'h2002_0000), .next_chained_descr_addr('h2002_0020), .length(op_len_bytes),   .more_in_chain(1'b1), .w_nrd_sel(1'b1), .host_addr(64'h2000_0000), .local_addr('h0000));

      t_add_chaining_descriptor(.start_address_in_host_memory(64'h2002_0020), .next_chained_descr_addr('h2002_0040), .length(op_len_bytes/4), .more_in_chain(1'b1), .w_nrd_sel(1'b1), .host_addr(64'h2000_1000), .local_addr('h00));
      t_add_chaining_descriptor(.start_address_in_host_memory(64'h2002_0040), .next_chained_descr_addr('h2002_0060), .length(op_len_bytes/4), .more_in_chain(1'b1), .w_nrd_sel(1'b0), .host_addr(64'h2000_1000), .local_addr('h00));
      t_add_chaining_descriptor(.start_address_in_host_memory(64'h2002_0060), .next_chained_descr_addr('h2002_0080), .length(op_len_bytes/4), .more_in_chain(1'b1), .w_nrd_sel(1'b1), .host_addr(64'h2000_1000 +   op_len_bytes/4 ), .local_addr('h00 +   op_len_bytes/4));
      t_add_chaining_descriptor(.start_address_in_host_memory(64'h2002_0080), .next_chained_descr_addr('h2002_00a0), .length(op_len_bytes/4), .more_in_chain(1'b1), .w_nrd_sel(1'b0), .host_addr(64'h2000_1000 +   op_len_bytes/4 ), .local_addr('h00 +   op_len_bytes/4));
      t_add_chaining_descriptor(.start_address_in_host_memory(64'h2002_00a0), .next_chained_descr_addr('h2002_00c0), .length(op_len_bytes/4), .more_in_chain(1'b1), .w_nrd_sel(1'b1), .host_addr(64'h2000_1000 +   op_len_bytes/2 ), .local_addr('h00 +   op_len_bytes/2));
      t_add_chaining_descriptor(.start_address_in_host_memory(64'h2002_00c0), .next_chained_descr_addr('h2002_00e0), .length(op_len_bytes/4), .more_in_chain(1'b1), .w_nrd_sel(1'b0), .host_addr(64'h2000_1000 +   op_len_bytes/2 ), .local_addr('h00 +   op_len_bytes/2));
      t_add_chaining_descriptor(.start_address_in_host_memory(64'h2002_00e0), .next_chained_descr_addr('h2002_0100), .length(op_len_bytes/4), .more_in_chain(1'b1), .w_nrd_sel(1'b1), .host_addr(64'h2000_1000 + 3*op_len_bytes/4 ), .local_addr('h00 + 3*op_len_bytes/4));
      t_add_chaining_descriptor(.start_address_in_host_memory(64'h2002_0100), .next_chained_descr_addr('hdead_beed), .length(op_len_bytes/4), .more_in_chain(1'b0), .w_nrd_sel(1'b0), .host_addr(64'h2000_1000 + 3*op_len_bytes/4 ), .local_addr('h00 + 3*op_len_bytes/4));


      dma_done = 1'b0;
      fork
         begin
            t_start_dma();
            @(posedge dma_irq);
            dma_done = 1'b1;
            check_irq_status;
            clear_irq;
         end
         while (1'b0 == dma_done) begin
            // Wait for a transfer to be done....     but to simulate something like DDR refresh timing backpressure lets wait a while...
            // Trying to backport this, I hope the semantics are right.....
            @(posedge clk_125m);
            if (1'b1 == `UUT_GN4124.dma_ctrl_direction_d) begin
                @(posedge `UUT_GN4124.dma_cyc_o);
                @(posedge clk_125m);
                force `UUT_GN4124.dma_stall_i = 1'b1;
                @(posedge clk_125m);
                force `UUT_GN4124.dma_ack_i = 1'b0;
                repeat(125) @(posedge clk_125m);
                release `UUT_GN4124.dma_stall_i;
                @(posedge clk_125m);
                release `UUT_GN4124.dma_ack_i;
                @(negedge `UUT_GN4124.dma_ctrl_direction_d);
            end // if
         end // while
      join



      // allow wishbone to finish
     //#1us;
     // t_compare_memories('h00000000, 'h0, op_len_bytes/4, badcomp, goodcomp, 1);

      for (addr = 'h00; addr < op_len_bytes/4; addr += 1)
        begin
           expected = rand_wdat - addr;
           t_get_device_mem(addr, wdat);
           if (expected !== wdat) begin
              $write("\n  Exected 0x%08x, got 0x%08x, for addr 0x%04x \n", expected, wdat, addr);
              $fatal();
           end
        end

        // compare other mem regions as well!!

      repeat(2) @(posedge clk_125m);


      $write("PASS\n");
   endtask



// -------------------------------------------
//  TESTCASE:    t_test_32bit_addr_overflow_wr
//      Purpose:
//            Test a 32bit Overflow for the host address.
//      Method:
//            setup the 2 test infrastructure BAR's on different consecutive addrs,
//             (and lets be evil and make one on a 32bit bar and one a 64bit BAR...)
//              set up a small DMA that crosses this boundary and then check the data...
//      NOTES: TBD: this test passes, but the request corsses a 4KB boundary
//                   (actually it is a 32 -> 64 bit addr boundary).
//                 I epxect it should FAIL however the BFM is not flagging any errors.....   TBD
// --------------------------------------------
   task automatic t_test_32bit_addr_overflow_wr(ref   int ntest,
                                                input int tests
                                               );

      uint32_t     addr;
      uint32_t     expected;
      uint64_t     wdat;
      uint64_t     pcie_addr;
      uint64_t     rdat;
      automatic logic [31:0] rand_wdat = $urandom();

      testphase = "t_test_32bit_addr_overflow_wr: ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);


      test_subphase = "config_bars";
      i_gn4124.t_configure_bar(0, 'h100000000, 'h20000000);
      i_gn4124.t_configure_bar(1, 'h0fffff000, 'h00001000);


      test_subphase = "init test mem";
      for (addr = 'h00; addr < 16; addr += 1) begin
         wdat = 64'h0;
         wdat[0+:32] = rand_wdat - addr;
         pcie_addr      = 64'hffff_fff0 + addr*4;
         i_gn4124.host_mem_write_noaddr_twiddles(pcie_addr, wdat);
      end

      test_subphase = "config_dma";
      t_setup_dma(.host_addr               ('hffff_fff0),
                  .local_addr              ('h00),
                  .length                  ('h20),
                  .more_in_chain           (0),
                  .w_nrd_sel               (1),
                  .next_chained_descr_addr ('h20020000));

      test_subphase = "do_dma";
      t_start_dma();
      @(posedge dma_irq);
      check_irq_status;
      clear_irq;

      test_subphase = "Check_result";

      //t_compare_memories('hffff_fff0, 'h0, 8, addr, expected, 1'b1, 1'b1);


      // do data check
      for (addr = 'h00; addr < 8; addr++)
        begin
           expected = rand_wdat - addr;
           t_get_device_mem(addr, rdat);
           //$write("\n  Exected 0x%08x, got 0x%08x, for addr 0x%04x \n", expected, rdat, addr);
           if (expected !== rdat[31:0]) begin
              $write("\n ERROR: Exected 0x%08x, got 0x%08x, for addr 0x%04x \n", expected, rdat, addr);
              $fatal();
           end
        end


      test_subphase = "Revert BAR configs";
      i_gn4124.t_detault_bar_config();


      repeat(2) @(posedge clk_125m);
      $write("PASS\n");
   endtask




// -------------------------------------------
//  TESTCASE:    t_test_32bit_addr_overflow_rd
//      Purpose:
//            Test a 32bit Overflow for the host address.
//      Method:
//            setup the 2 test infrastructure BAR's on different consecutive addrs,
//             (and lets be evil and make one on a 32bit bar and one a 64bit BAR...)
//              set up a small DMA that crosses this boundary and then check the data...
// --------------------------------------------
   task automatic t_test_32bit_addr_overflow_rd(ref   int ntest,
                                                input int tests
                                               );

      uint32_t     addr;
      uint32_t     expected;
      uint64_t     wdat;
      uint64_t     pcie_addr;
      uint64_t     rdat;

      uint64_t    pcie_start_addr = 64'hffff_ffe8;
      automatic logic [31:0] rand_wdat = $urandom();

      testphase = "t_test_32bit_addr_overflow_rd: ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);


      test_subphase = "config_bars";
      i_gn4124.t_configure_bar(0, 'h100000000, 'h20000000);
      i_gn4124.t_configure_bar(1, 'h0fffff000, 'h00001000);


      test_subphase = "init test mem";
      for (addr = 'h00; addr < 16; addr += 1) begin
         wdat = 64'h0;
         wdat[0+:32] = rand_wdat - addr;
         t_blat_device_mem(addr, wdat[31:0]);
      end

      test_subphase = "config_dma";
      t_setup_dma(.host_addr               (pcie_start_addr),
                  .local_addr              ('h00),
                  .length                  ('h40),
                  .more_in_chain           (0),
                  .w_nrd_sel               (0),
                  .next_chained_descr_addr ('h20020000));

      test_subphase = "do_dma";
      t_start_dma();
      @(posedge dma_irq);
      check_irq_status;
      clear_irq;

      test_subphase = "Check_result";

   //   t_compare_memories(pcie_start_addr, 'h0, 16, addr, expected, 1'b1, 1'b1);


      // do data check
      for (addr = 'h00; addr < 16; addr++)
        begin
           expected = rand_wdat - addr;
           pcie_addr = pcie_start_addr + addr*4;
           i_gn4124.host_mem_read_noaddr_twiddles(pcie_addr, rdat);
          // $write("\n  Exected 0x%08x, got 0x%08x, for addr 0x%04x \n", expected, rdat, addr);
           if (expected !== rdat[31:0]) begin
              $write("\n ERROR: Exected 0x%08x, got 0x%08x, for addr 0x%04x \n", expected, rdat, addr);
              $fatal();
           end
        end


      test_subphase = "Revert BAR configs";
      i_gn4124.t_detault_bar_config();

      repeat(2) @(posedge clk_125m);
      $write("PASS\n");
   endtask







// -------------------------------------------
//  TESTCASE:    t_test_4K_boundary_rw
//      Purpose:
//           BASic self-checking DMA read (sys_mem to card)
//           The address is designed to cross a PCIE 4K boundary. this should be split into multiple packets!
//      Method:
//          NOTE: NOT self checking of memory contents...
// --------------------------------------------
   task automatic t_test_4K_boundary_rw(ref   int ntest,
                                         input int tests
                                         );
      uint32_t addr,expected, from_mem ;
      bit      error_detected;
      uint32_t badcomp, goodcomp;

      logic [31:0] rand_wdat = $urandom();


      testphase = "t_test_4K_boundary_rw : ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);

      for (addr = 'h00; addr < 'h10; addr += 1) begin              // TBD: do we make this 'more' randomised data!
        t_blat_device_mem(addr, rand_wdat - addr*7);
      end

        t_setup_dma(.host_addr               ('h2000_0FF8),
                    .local_addr              ('h00),
                    .length                  ('h10),
                    .more_in_chain           (1),
                    .w_nrd_sel               (0),
                    .next_chained_descr_addr ('h20020000));

         t_add_chaining_descriptor(.start_address_in_host_memory(64'h2002_0000),
                                   .next_chained_descr_addr('h2002_0000),
                                   .length(16),
                                   .more_in_chain(1'b0),
                                   .w_nrd_sel(1'b1),
                                   .host_addr(64'h2000_0ff8),
                                   .local_addr('h0010)
                                  );

       t_start_dma();

      @(posedge dma_irq);

      check_irq_status;
      clear_irq;

    // for (addr = 'h00; addr < 4; addr += 1)
    //   begin
    //      expected = rand_wdat - addr*7;
    //      mem_check(4 * addr, expected);
    //   end
    //
    // for (addr = 'h00; addr < 4; addr += 1)
    //   begin
    //      expected = rand_wdat - addr*7;
    //      mem_check(4 * addr + 16, expected);
    //   end


      repeat(4) @(posedge clk_125m);

      $write("PASS\n");
   endtask


// -------------------------------------------
//  TESTCASE:    t_test_read_chaining
//      Purpose:
//            Sanity checking my changes for read chaining with a longer chain...
//      Method:
//          NOTE: SELF_CHECKING: TBD
//          all about getting multple writes to be sdone
//
// --------------------------------------------
   task automatic t_test_read_chaining(ref   int ntest,
                                       input int tests
                                      );

      uint32_t addr,expected;
      uint32_t badcomp, goodcomp;


      logic [31:0] rand_wdat = $urandom();
 //   rand_wdat = 32'h80000020;


      testphase = "t_test_read_chaining: ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);

      // Setup data in BFM memory
      for (addr = 'h00; addr < 'h20; addr += 1) begin
        t_blat_device_mem(addr, rand_wdat - addr);
      end   // BUT Ideally  still need to set it in the target mem not the host mem...

      if (dma_irq != 1'b0) begin
        $fatal(1, "dma irq should be 0");
      end


      t_add_chaining_descriptor(.host_addr                    ('h2000_0010),
                                .local_addr                   ('h10),
                                .length                       ('h18),
                                .more_in_chain                (1'b1),
                                .w_nrd_sel                    (1'b0),
                                .next_chained_descr_addr      ('h2002_0020),
                                .start_address_in_host_memory ('h2002_0000));

      t_add_chaining_descriptor(.host_addr                    ('h2000_0028),
                                .local_addr                   ('h28),
                                .length                       ('h10),
                                .more_in_chain                (1'b1),
                                .w_nrd_sel                    (1'b0),
                                .next_chained_descr_addr      ('h2002_0040),
                                .start_address_in_host_memory ('h2002_0020));

      t_add_chaining_descriptor(.host_addr                   ('h2000_0030),
                                .local_addr                  ('h30),
                                .length                      ('h10),
                                .more_in_chain               (1'b1),
                                .w_nrd_sel                   (1'b0),
                                .next_chained_descr_addr     ('h2002_0060),
                                .start_address_in_host_memory('h2002_0040));

      t_add_chaining_descriptor(.host_addr                   ('h2000_0040),
                                .local_addr                  ('h40),
                                .length                      ('h40),
                                .more_in_chain               (1'b0),
                                .w_nrd_sel                   (1'b0),
                                .next_chained_descr_addr     ('hffff0),
                                .start_address_in_host_memory('h2002_0060));

        t_setup_dma(.host_addr               ('h2000_0000),
                    .local_addr              ('h00),
                    .length                  ('h300),
                    .more_in_chain           (1),
                    .w_nrd_sel               (0),
                    .next_chained_descr_addr ('h2002_0000));


       t_start_dma();

       // TODO: implement a timeout here.....
       @(posedge dma_irq);

      check_irq_status;
      clear_irq;


      //t_compare_memories('h00000000, 'h0, 32, badcomp, goodcomp, 1);

      for (addr = 'h00; addr < 'h20; addr += 1)
        begin
           expected = rand_wdat - addr;
           mem_check(4 * addr, expected);
        end

      repeat(2) @(posedge clk_125m);

      $write("PASS\n");
   endtask




 // This isn't a testcase it is just a building block....
   task automatic t_send_axi4_wr_rand(input logic [ 1:0] id,
                                      input logic [63:0] start_addr,
                                      input int          length,
                                      input int          timeout,
                                      input bit          force_error = 1'b0,
                                      input logic [3:0]  fwbe        = 4'hf,
                                      input bit          verbose     = 1'b0
                                     );
      logic [31:0] wdat_arry [*];
      logic [ 3:0] first_strb;
      logic [ 3:0] last_strb;

      logic [1:0] b_resp;
      logic [1:0] b_id;

     logic     got_b;
      uint64_t  host_mem_moddeled_dat;


     logic  detected_error;
     int    error_count = 0;

     logic [31:0] first_word_in_target_before_write;
     logic [31:0] last_word_in_target_before_write;


     // first_strb = $urandom_range(0,15);
     // last_strb  = $urandom_range(0,15);

     //-> enter_task;

     last_strb  = 4'hf;
     first_strb = fwbe;

     got_b = 1'b0;


         i_gn4124.host_mem_read_noaddr_twiddles((start_addr/4) * 4, host_mem_moddeled_dat);
         first_word_in_target_before_write = host_mem_moddeled_dat[0+:32];
         i_gn4124.host_mem_read_noaddr_twiddles(((start_addr/4) + length) * 4, host_mem_moddeled_dat);
         last_word_in_target_before_write  = host_mem_moddeled_dat[0+:32];
      // make first strobe legal, based on start_Addr
   //   for int bitno = 0; bitno <4; bitno++) begin
   //      if (bitno >= start_addr[0+:2] ) begin
   //         first_strb[bitno] = 1'b1;
   //      end else begin
   //         first_strb[bitno] = 1'b0;
   //      end
   //   end

      for (int i = 0; i <= length; i++) begin
         wdat_arry[i] = $urandom();
      end

      // " nested forks, the outside if a join_any  for a timeout
      fork
         begin // Fistf part of Join_any : this is the write OP
            fork  // run 3 threads simultaneousley: sending wdat, AW and waiting for next B_resp
               begin
                  if (1'b1 == verbose) $display("fork start send_w");
                  for (int wphs = 0; wphs <= length; wphs++) begin
                     if (0 == wphs) begin
                        t_send_w(.last( (wphs==length) ? 1'b1 : 1'b0), .strb(first_strb), .data(wdat_arry[wphs]));
                     end else if (length == wphs) begin
                        t_send_w(.last( (wphs==length) ? 1'b1 : 1'b0), .strb(last_strb), .data(wdat_arry[wphs]));
                     end else begin
                        t_send_w(.last( (wphs==length) ? 1'b1 : 1'b0), .strb(4'hf), .data(wdat_arry[wphs]));
                     end
                  end
                  if (1'b1 == verbose) $display("fork end send_w");
               end

               begin
                  if (1'b1 == verbose) $display("fork start send_aw");
                  t_send_aw(.id(id), .addr(start_addr), .len(length[0+:8]), .size(3'h2), .burst(2'h1), .cache(4'h0));
                  if (1'b1 == verbose) $display("fork end send_aw");
               end

               begin
                  if (1'b1 == verbose) $display("fork start get_b");
                  t_get_b(.resp(b_resp), .id(b_id));
                  got_b = 1'b1;
                  if (1'b1 == verbose) $display("fork end get_b\n *************\n");
               end
            join
         end

         begin      // specond part of Join_any; this is the timout...
            int timeout_count = 0;

            while ((timeout_count++ < timeout) & (1'b0 == got_b)) begin
               @(posedge axi_clk);
            end
            if (1'b0 == got_b) begin
                  $display("timed out while waiting for a b_resp");
                  $fatal();
            end
         end

      join_any
    // -> finished_forking;

     if (b_resp !== 2'b00) begin
        $display("got b_resp = ERROR"); $fatal();
     end

     if (b_id !== id) begin
        $display("got b_id that was different to the supplied value: ERROR, expected 0x%02x, gtot: 0x%02x", id, b_id); $fatal();
     end

      // wait for a little to alloow the BFM to complete...
   //  repeat(20)  @(posedge `UUT_GN4124.gen_with_dma.uut.preproc_inst.clk);

     detected_error = 1'b0;

     // // inject an error to test this code!
     if (1'b1 == force_error) begin
        host_mem_moddeled_dat = 64'h1234567890abcdef;
        i_gn4124.host_mem_write_noaddr_twiddles(start_addr + 4, host_mem_moddeled_dat);
     end


     // change expected data based on fwbe
     wdat_arry[0][ 0+:8] = first_strb[0] ? wdat_arry[0][ 0+:8] : first_word_in_target_before_write[ 0+:8];
     wdat_arry[0][ 8+:8] = first_strb[1] ? wdat_arry[0][ 8+:8] : first_word_in_target_before_write[ 8+:8];
     wdat_arry[0][16+:8] = first_strb[2] ? wdat_arry[0][16+:8] : first_word_in_target_before_write[16+:8];
     wdat_arry[0][24+:8] = first_strb[3] ? wdat_arry[0][24+:8] : first_word_in_target_before_write[24+:8];



    @(posedge axi_clk);

      // check all locations  for data mismatches!!    (note ignores impact of W STrobes!!)!
     for (int i = 0; i <= length; i++) begin
         i_gn4124.host_mem_read_noaddr_twiddles(start_addr + 4*i, host_mem_moddeled_dat);
        if (wdat_arry[i]  !== host_mem_moddeled_dat[0+:32]) begin
         detected_error = 1'b1;
         error_count ++;
       end
      end



     // if detected_error then print the range!!!
      if (1'b1 == detected_error) begin
         for (int i = 0; i <= length; i++) begin
             i_gn4124.host_mem_read_noaddr_twiddles(start_addr + 4*i, host_mem_moddeled_dat);
           $write(" AXI_W_MASTER_DATA_CHECK : addr:0x%016x  expected: 0x%08x  got:  0x%08x  good:%01b\n", start_addr + 4*i, wdat_arry[i], host_mem_moddeled_dat[0+:32], wdat_arry[i]=== host_mem_moddeled_dat[0+:32]);
         end
         if (error_count === 0 + force_error) begin
            $write(" AXI_W_MASTER_DATA_CHECK : error count was 0d%d, ane error injection was %01b, THIS_IS_ACCEPTABLE\n", error_count, force_error);
         end else begin
            $write(" AXI_W_MASTER_DATA_CHECK : error count was 0d%d, ane error injection was %01b, TEST: FAIL\n", error_count, force_error);
            $fatal();
         end
      end

      if ((1'b1 == force_error) & !(1'b1 == detected_error)) begin
         $write(" AXI_W_MASTER_DATA_CHECK : an error was supposed to be inserted, but none was detected,  please check...   ERR_VAL = 0x%016x\n", host_mem_moddeled_dat);
         for (int i = 0; i <= length; i++) begin
             i_gn4124.host_mem_read_noaddr_twiddles(start_addr + 4*i, host_mem_moddeled_dat);
           $write(" AXI_W_MASTER_DATA_CHECK : addr:0x%016x  expected: 0x%08x  got:  0x%08x  good:%01b\n", start_addr + 4*i, wdat_arry[i], host_mem_moddeled_dat[0+:32], wdat_arry[i]=== host_mem_moddeled_dat[0+:32]);
         end
         $fatal();
      end
    // -> finished_Task;
   endtask







   task automatic t_axi_read_and_check( input logic [ 1:0] id,
                                        input logic [63:0] addr,
                                        input logic [ 7:0] len,
                                        input logic [ 2:0] size    = 3'h2,
                                        input logic [ 1:0] burst   = 2'h1,
                                        input logic [ 3:0] cache   = 4'h0,
                                        input integer      timeout = 1024
                                      );
     integer   current_axi_good;
     integer   current_axi_bad;
     uint64_t  host_mem_moddeled_dat;



     if (axi_rd_checking_queue.size !== 0) begin
        $display("Queue should be empty if tou are caling this task!");
        $fatal();
     end

     // initialise BFM target mem with random data...
     for (int i = 0; i <= len; i++) begin
        host_mem_moddeled_dat[31:0] = $urandom();
        i_gn4124.host_mem_write_noaddr_twiddles(addr + 4*i, host_mem_moddeled_dat);
     end


     current_axi_good = axi_rdat_checked_ok_count;
     current_axi_bad  = axi_rdat_checked_bad_count;

     fork
      t_send_ar(.id(   id),
                .addr( addr),
                .len(  len),
                .size( size),
                .burst(burst),
                .cache(cache)
               );

      t_check_axi_rd_Dat(.id(     id),
                         .addr(   addr),
                         .len(    len),
                         .timeout(timeout)
                        );

     join
     // wait for the event to indicate that the event has been processed and checked against the BFM
     @axi_read_processed;

     if (axi_rdat_checked_ok_count !== current_axi_good + 1) begin
        $display("unexpected non-increment to axi_rdat_checked_ok_count");
     end;

   endtask


 task automatic t_queued_axi_reads(input int num_tests);
   typedef struct {
      logic [ 1:0] id;
      logic [63:0] start_addr;
      logic  [7:0] len;
   }t_xaction;

   t_xaction test_create;

   uint64_t host_mem_moddeled_dat;

   t_xaction xaction_q [$];
   t_xaction checking_q [$];


   int posted_cmds;
   int checked_rds;
   int received_read_pkts;

   bit commands_all_sent;

   // todo:initialise memory
   for (integer mem_offset = 0;  mem_offset < 500; mem_offset++) begin
      host_mem_moddeled_dat[ 0+:32] = $urandom();
      host_mem_moddeled_dat[32+:32] = $urandom();
      i_gn4124.host_mem_write_noaddr_twiddles(64'h2000_0000 + 4*mem_offset, host_mem_moddeled_dat);
   end

   // create a read
   for (integer i = 0; i<num_tests; i++) begin
     test_create.id = $urandom_range(0,3);
     test_create.len = $urandom_range(0,255);

     // pick an addr in bar 0
     test_create.start_addr = 64'h2000_0000 + 4*$urandom_range(0,150);

     xaction_q.push_back(test_create);
   end


   commands_all_sent = 1'b0;

   posted_cmds = 0;
   received_read_pkts = 0;
   checked_rds = 0;

   fork
      begin
         t_xaction test_cmd;
         while (xaction_q.size() !== 0) begin
            test_cmd = xaction_q.pop_front();
            t_send_ar(.id(   test_cmd.id),
                      .addr( test_cmd.start_addr),
                      .len(  test_cmd.len),
                      .size( 3'h2),
                      .burst(2'h1),
                      .cache(4'h0)
                     );
            checking_q.push_back(test_cmd);
            posted_cmds++;
        //    $display("issued cmd: id 0x:%01x, addr:0x%016x, len 0x%02x", test_cmd.id, test_cmd.start_addr, test_cmd.len);
         end
         commands_all_sent = 1'b1;
        // $display("have now posted %d commands", posted_cmds);
      end

      begin
         t_xaction check_cmd;
         while ((commands_all_sent == 1'b0) || (checking_q.size() != 0)) begin
            if (checking_q.size() == 0) begin
               @(posedge axi_clk);
            end else begin
               check_cmd = checking_q.pop_front();
            //   $display("checking cmd: id 0x:%01x, addr:0x%016x, len 0x%02x", check_cmd.id, check_cmd.start_addr, check_cmd.len);
               t_check_axi_rd_Dat(.id(     check_cmd.id),
                                  .addr(   check_cmd.start_addr),
                                  .len(    check_cmd.len),
                                  .timeout(250 + 4*check_cmd.len)
                                 );
            //   $display("Collected rd data resp %d", received_read_pkts);
               received_read_pkts++;
            end
         end
      end


      begin
         integer read_procesing_timeout;
         bit     got_enough_reads;
         integer read_count;

         read_procesing_timeout = 0;
         got_enough_reads       = 0;
         read_count             = 0;
         fork
            begin
               while ((read_procesing_timeout < 5000) && (1'b0 == got_enough_reads)) begin
                  @(posedge axi_clk);
                  read_procesing_timeout++;
               end
            end

            begin
               while (read_count < num_tests) begin
                  @(axi_read_processed);
                  read_count++;
                  read_procesing_timeout = 0;
                  checked_rds++;
               end
               got_enough_reads = 1'b1;

            end
         join_any
      end
   join

   if (posted_cmds === checked_rds) begin
   //   $display("t_queued_axi_reads have issued %d reads and got the same number of correct responses",checked_rds );
   end else begin
       $display("t_queued_axi_reads have issued %d reads but got a different number of responses = %d ",posted_cmds, checked_rds );
       $fatal();
   end

endtask

// -------------------------------------------
//  TESTCASE:    t_axi_r_directed_tests
//      Purpose:
//           BASic directed tests for AXI_R functinoality
//      Method:
//                in development
// --------------------------------------------
task automatic t_axi_r_directed_tests(ref   int ntest,
                                      input int tests);
      testphase = "Directed tests AXI4 slave (read): ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);


      t_axi_read_and_check(.id(   2'h0),
                           .addr( 64'h20001344),
                           .len(  3),
                           .size( 3'h2),
                           .burst(2'h1),
                           .cache(4'h0),
                           .timeout(128)
                          );

                          // let's use the default values now..

      //    hit a 64b target!
      t_axi_read_and_check(.id(2'h1), .addr(64'h400001234), .len(3) );

      // try different lengths
      t_axi_read_and_check(.id(2'h1), .addr(64'h400001234), .len(0));
      t_axi_read_and_check(.id(2'h1), .addr(64'h400001234), .len(1));
      t_axi_read_and_check(.id(2'h1), .addr(64'h400001234), .len(2));
      t_axi_read_and_check(.id(2'h1), .addr(64'h400001234), .len(254));
      t_axi_read_and_check(.id(2'h1), .addr(64'h400001234), .len(255));

      // make sure we have twiddled the bottom 2 addr bits so they can impact the fwbe!
      for (int i = 0; i < 4; i++) begin
         t_axi_read_and_check(.id(2'h1), .addr(64'h400001234 + i), .len(0));
      end


       t_queued_axi_reads(7);

        $write("PASS\n");

      // make testcase seperation obvious...
      #5us;

endtask



// -------------------------------------------
//  TESTCASE:    t_axi_w_directed_tests
//      Purpose:
//           BASic directed tests for AXI_W functinoality
//      Method:
//          NOTE: self checking of memory contents...
//
// --------------------------------------------
task automatic t_axi_w_directed_tests(ref   int ntest,
                                      input int tests);

     int addr_offset;


      testphase = "Directed tests AXI4 slave (write): ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);

     t_send_axi4_wr_rand( .id(2'h0), .start_addr(64'h020001234), .length(3), .timeout(125));
     t_send_axi4_wr_rand( .id(2'h1), .start_addr(64'h020573210), .length(7), .timeout(125));
     t_send_axi4_wr_rand( .id(2'h2), .start_addr(64'h020001234), .length(1), .timeout(125));
     t_send_axi4_wr_rand( .id(2'h3), .start_addr(64'h020001234), .length(0), .timeout(125));

     // test fwbe on length 2 ops, addr changes accordingly
     for (addr_offset = 0; addr_offset <4; addr_offset++) begin
       case (addr_offset)
          0:      t_send_axi4_wr_rand( .id(2'h0), .start_addr(64'h20001244 + addr_offset), .length(1), .timeout(125), .fwbe(4'hf));
          1:      t_send_axi4_wr_rand( .id(2'h0), .start_addr(64'h20001254 + addr_offset), .length(1), .timeout(125), .fwbe(4'he));
          2:      t_send_axi4_wr_rand( .id(2'h0), .start_addr(64'h20001264 + addr_offset), .length(1), .timeout(125), .fwbe(4'hc));
          3:      t_send_axi4_wr_rand( .id(2'h0), .start_addr(64'h20001274 + addr_offset), .length(1), .timeout(125), .fwbe(4'h8));
       endcase
     end


     // test fwbe on length 2 ops, no base addr change
     for (addr_offset = 0; addr_offset <4; addr_offset++) begin
       case (addr_offset)
          0:      t_send_axi4_wr_rand( .id(2'h2), .start_addr(64'h20001344), .length(1), .timeout(125), .fwbe(4'hf));
          1:      t_send_axi4_wr_rand( .id(2'h2), .start_addr(64'h20001354), .length(1), .timeout(125), .fwbe(4'he));
          2:      t_send_axi4_wr_rand( .id(2'h2), .start_addr(64'h20001364), .length(1), .timeout(125), .fwbe(4'hc));
          3:      t_send_axi4_wr_rand( .id(2'h2), .start_addr(64'h20001374), .length(1), .timeout(125), .fwbe(4'h8));
       endcase
     end

     // test fwbe on single flit ops..
     for (addr_offset = 0; addr_offset <4; addr_offset++) begin
       case (addr_offset)
          0:      t_send_axi4_wr_rand( .id(2'h1), .start_addr(64'h20001284 + addr_offset), .length(0), .timeout(125), .fwbe(4'hf));
          1:      t_send_axi4_wr_rand( .id(2'h1), .start_addr(64'h20001294 + addr_offset), .length(0), .timeout(125), .fwbe(4'he));
          2:      t_send_axi4_wr_rand( .id(2'h1), .start_addr(64'h200012a4 + addr_offset), .length(0), .timeout(125), .fwbe(4'hc));
          3:      t_send_axi4_wr_rand( .id(2'h1), .start_addr(64'h200012b4 + addr_offset), .length(0), .timeout(125), .fwbe(4'h8));
       endcase
     end

     // test a Long operation (1KB, max AXI AWLEN=255 will support)
      t_send_axi4_wr_rand( .id(2'h3), .start_addr(64'h020001000), .length(255), .timeout(600));

     // try writing to BFM_BAR0 instead of BFM_BAR1
     t_send_axi4_wr_rand( .id(2'h0), .start_addr(64'h400001234), .length(5), .timeout(125));


     $write("PASS\n");

endtask



// -------------------------------------------
//  TESTCASE:    t_axi_w_b2b_tests
//      Purpose:
//           BASic testing of back2back writes
//      Method:
//          NOTES: .......
// --------------------------------------------
task automatic t_axi_w_b2b_tests(ref   int ntest,
                                 input int tests);

      int addr_offset;
      logic [31:0] wdat_arry [*];

      bit dataphases_done;
      uint64_t host_mem_moddeled_dat;

      testphase = "Directed tests AXI4 slave (write) back2back: ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);

      for (int i = 0; i <= 12; i++) begin
         wdat_arry[i] = $urandom();
      end

      addr_offset     = 0;
      dataphases_done = 0;
      fork

          begin
              // first phase: back to back commands, then data arrives..
              fork
                 begin
                    t_send_aw(.id(0), .addr(64'h20000000), .len(1), .size(3'h2), .burst(2'h1), .cache(4'h0));
                    t_send_aw(.id(1), .addr(64'h20000008), .len(1), .size(3'h2), .burst(2'h1), .cache(4'h0));
                    t_send_aw(.id(2), .addr(64'h20000010), .len(1), .size(3'h2), .burst(2'h1), .cache(4'h0));
                //    t_send_aw(.id(id), .addr(64'h20000018), .len(1), .size(3'h2), .burst(2'h1), .cache(4'h0));
                //    t_send_aw(.id(id), .addr(64'h20000020), .len(1), .size(3'h2), .burst(2'h1), .cache(4'h0));
                //    t_send_aw(.id(id), .addr(64'h20000028), .len(1), .size(3'h2), .burst(2'h1), .cache(4'h0));
                 end

                 begin
                    repeat (20) @(posedge axi_clk);
                    for (int write_no = 0; write_no <3; write_no ++) begin
                       for (int wphs = 0; wphs <= 1; wphs++) begin
                          t_send_w(.last( (wphs==1) ? 1'b1 : 1'b0), .strb(4'hf), .data(wdat_arry[addr_offset++]));
                       end
                    end
                 end
              join

               #500ns;

              fork
                 begin
                    repeat (20) @(posedge axi_clk);
                    t_send_aw(.id(3), .addr(64'h20000018), .len(1), .size(3'h2), .burst(2'h1), .cache(4'h0));
                    repeat (20) @(posedge axi_clk);
                    t_send_aw(.id(2), .addr(64'h20000020), .len(1), .size(3'h2), .burst(2'h1), .cache(4'h0));
                    repeat (20) @(posedge axi_clk);
                    t_send_aw(.id(1), .addr(64'h20000028), .len(1), .size(3'h2), .burst(2'h1), .cache(4'h0));
                 end

                 begin
                    for (int write_no = 0; write_no <3; write_no ++) begin
                       for (int wphs = 0; wphs <= 1; wphs++) begin
                          t_send_w(.last( (wphs==1) ? 1'b1 : 1'b0), .strb(4'hf), .data(wdat_arry[addr_offset++]));
                       end
                    end
                 end
              join
           end

           // get the requisite number of good bresp
           begin
              logic [1:0] b_resp;
              logic [1:0] b_id;
              logic [1:0] expected_id;
              for(int expected_responses = 0; expected_responses<6; expected_responses++) begin
                 t_get_b(.resp(b_resp), .id(b_id));

                 if (b_resp !== 2'h0) begin
                    $display("bad bresp");
                    $fatal();
                 end

                 case (expected_responses)
                    0: expected_id = 0;
                    1: expected_id = 1;
                    2: expected_id = 2;
                    3: expected_id = 3;
                    4: expected_id = 2;
                    5: expected_id = 1;
                    default: begin $display("unexpected value");
                                   $fatal();
                             end
                 endcase

                 if (b_id !== expected_id) begin
                    $display("bad b_id");
                    $fatal();
                 end

              end
              dataphases_done = 1;
           end

           // timeout
           begin
              integer timeout_count;
              timeout_count = 0;
              while ((timeout_count < 50) & (1'b0 == dataphases_done)) begin
                 @(posedge axi_clk);
              end
              if (1'b0 == dataphases_done) begin
                 $display("timeout");
                 $fatal();
              end
           end
      join_any


      #100ns;
      for (int i = 0; i < 12; i++) begin
         i_gn4124.host_mem_read_noaddr_twiddles(64'h20000000 + 4*i, host_mem_moddeled_dat);
         if (wdat_arry[i]  !== host_mem_moddeled_dat[0+:32]) begin
            $display("bad data compare");
            $fatal();
         //end else begin
         //    $display("good data compare %d", i);
         end
      end

      #1us;

     $write("PASS\n");
endtask


 task automatic t_subtest_insufficient_wr_dat();

       logic [1:0] b_resp;
       logic [1:0] b_id;

       logic [63:0] target_mem_content;
       uint64_t host_mem_moddeled_dat;

      test_subphase = "try writing unsufficient data";

      i_gn4124.host_mem_read_noaddr_twiddles(64'h20000000,  host_mem_moddeled_dat);
      target_mem_content = host_mem_moddeled_dat;

      fork
        t_send_aw(.id(3), .addr(64'h20000000), .len(1), .size(3'h2), .burst(2'h1), .cache(4'h0));
        t_send_w(.last(1'b1), .strb(4'hf), .data(32'h0));
        t_get_b(.resp(b_resp), .id(b_id));
      join

      if (b_id !== 2'h3) begin
        $display("bad ID on B channel"); $fatal();
      end

      if (b_resp !== 2'b10) begin
         $display("expected an error on bresp"); $fatal();
      end

      #100ns;

      i_gn4124.host_mem_read_noaddr_twiddles(64'h20000000,  host_mem_moddeled_dat);
      if (target_mem_content !== host_mem_moddeled_dat) begin
         $display("t_subtest_insufficient_wr_dat() :: B_RESP indicates the write failed; origional, new :: 0x%016x, 0x%016x", target_mem_content, host_mem_moddeled_dat);
      end

endtask


task automatic t_subtest_axi_wr_extra_wr_dat(input logic [63:0] axi_addr,
                                             input logic [ 1:0] id,
                                             input logic  [7:0] axi_len,
                                             input int          extra_words,
                                             input logic [1:0]  expected_b_resp,
                                             input bit          do_data_check = 1
                                            );

       logic [1:0] b_resp;
       logic [1:0] b_id;

       logic [31:0] wdat_arry [*];
       uint64_t host_mem_moddeled_dat;
       int      write_offset;

      $sformat(test_subphase, "too much W_data: axi_len %d, extra_words %d", axi_len, extra_words);

      //   $display(" t_subtest_axi_wr_extra_wr_dat() :: check the mem contents a cmd with for axi_len 0d:%03d", axi_len);

      // get the remote memory contents so we can ccheck they are unaltered later...
      if (1'b1 == do_data_check) begin
         for (int len = 0; axi_len >= len; len++) begin
            i_gn4124.host_mem_read_noaddr_twiddles(axi_addr + 4*len, host_mem_moddeled_dat);
            wdat_arry[len] = host_mem_moddeled_dat[31:0];
          //  $display("reading data initially we see offset 0x%02x with 0x%08x", len, wdat_arry[len]);
         end
      end

      write_offset = 0;
      fork
        t_send_aw(.id(id), .addr(axi_addr), .len(axi_len), .size(3'h2), .burst(2'h1), .cache(4'h0));
        begin
           for (int wordno = axi_len + extra_words; wordno >= 0; wordno--) begin
              logic [31:0] newdat;
              newdat = $urandom();
              if (0 == wordno) begin
                 t_send_w(.last(1'b1), .strb(4'hf), .data(newdat));
              end else begin
                 t_send_w(.last(1'b0), .strb(4'hf), .data(newdat));
              end

              if (2'h0 == expected_b_resp) begin
               //  $display("becasue this is expected to be a doog write; offset 0x%02x = 0x%08x becomes  0x%08x", write_offset, wdat_arry[wordno], newdat[31:0]);
                 wdat_arry[write_offset++] = newdat;
              end
           end
        end
        t_get_b(.resp(b_resp), .id(b_id));
      join


      if (b_id !== id) begin
        $display("bad ID on B channel");
        $fatal();
      end

      if (b_resp !== expected_b_resp) begin
         $display("expected an error on bresp, expected 0d:%0d, got 0d:%0d", expected_b_resp, b_resp);
         $fatal();
      end

      #100ns;

      if (1'b1 == do_data_check) begin
         for (int len = 0; axi_len >= len; len++) begin
            i_gn4124.host_mem_read_noaddr_twiddles(axi_addr + 4*len, host_mem_moddeled_dat);
            if (wdat_arry[len] !== host_mem_moddeled_dat[31:0]) begin
               $display("*****: t_subtest_axi_wr_extra_wr_dat() :: check bad for offset %0d.  expected :got 0x:%08x 0x%08x", len, host_mem_moddeled_dat[31:0], wdat_arry[len]);
               $fatal();
            end else begin
            //   $display("*****: t_subtest_axi_wr_extra_wr_dat() :: check good for offset %0d", len);
            end
         end
      end else begin
      end

endtask


// -------------------------------------------
//  TESTCASE:    t_axi_w_error_tests
//      Purpose:
//           BASic testing ofor AXI4 error conditions...
//      Method:
//          NOTES: to cover:   wrong ammount of data
//                             bad command types
//                             bad size
//                             illegal cache values
// --------------------------------------------
task automatic t_axi_w_error_tests(ref   int ntest,
                                  input int tests);

      int addr_offset;
      logic [31:0] wdat_arry [*];

      bit dataphases_done;
      uint64_t host_mem_moddeled_dat;

     logic [1:0] b_id;
     logic [1:0] b_resp;

      testphase = "Directed tests AXI4 error (write): ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);

      t_subtest_insufficient_wr_dat();

      t_subtest_axi_wr_extra_wr_dat(.axi_addr(64'h20000000), .id(2'h0),
                                    .axi_len(0), .extra_words(1), .expected_b_resp(2'h2));

      t_assert_gennum_reset();

      t_subtest_axi_wr_extra_wr_dat(.axi_addr(64'h20000000), .id(2'h1),
                                    .axi_len(0), .extra_words(0), .expected_b_resp(2'h0));

      t_subtest_axi_wr_extra_wr_dat(.axi_addr(64'h20000000), .id(2'h2),
                                    .axi_len(1), .extra_words(1), .expected_b_resp(2'h2));

      t_assert_gennum_reset();

      t_subtest_axi_wr_extra_wr_dat(.axi_addr(64'h20000000), .id(2'h3),
                                    .axi_len(1), .extra_words(0), .expected_b_resp(2'h0));

      t_subtest_axi_wr_extra_wr_dat(.axi_addr(64'h20000000), .id(2'h2),
                                    .axi_len(1), .extra_words(2), .expected_b_resp(2'h2));

      t_assert_gennum_reset();

      t_subtest_axi_wr_extra_wr_dat(.axi_addr(64'h20000000), .id(2'h3),
                                    .axi_len(254), .extra_words(2), .expected_b_resp(2'h2));

      t_assert_gennum_reset();

      t_subtest_axi_wr_extra_wr_dat(.axi_addr(64'h20000000), .id(2'h0),
                                    .axi_len(255), .extra_words(1), .expected_b_resp(2'h2));

      t_assert_gennum_reset();

      t_subtest_axi_wr_extra_wr_dat(.axi_addr(64'h20000000), .id(2'h1),
                                    .axi_len(255), .extra_words(2), .expected_b_resp(2'h2));

      t_assert_gennum_reset();

      test_subphase = "fail on first/only word byte_enables";
      for (int addr = 0; addr < 4; addr++) begin
         for (int i = 0; i < 16; i++) begin
            t_send_aw(.id(0), .addr(64'h20000000 + addr), .len(0), .size(3'h2), .burst(2'h1), .cache(4'h0));
            t_send_w(.last( 1'b1), .strb(i[0+:4]), .data($urandom()));
            t_get_b(.resp(b_resp), .id(b_id));
            if (i[0+:4] == 4'h0) begin
               if (b_resp != 2'h2) begin
                  $display("1 got a response of 0x%01x for first_word strobes of 0b:%04b and bottom addr bits 0x:%01x ", b_resp, i, addr);
                  $fatal();
               end
            end else begin
               logic [3:0] mask;
               case (addr[1:0])
                 1:       mask = 4'b0001;
                 2:       mask = 4'b0011;
                 3:       mask = 4'b0111;
                 default: mask = 4'b0000;
               endcase

               mask = mask & i[0+:4];

               if (mask != 4'h0) begin  // a strobe bit below that indicated by addr[1:0] is asserted.
                  if (b_resp != 2'h2) begin
                     $display("2 got a response of 0x%01x for first_word strobes of 0b:%04b and bottom addr bits 0x:%01x mask = 0b:%04b ", b_resp, i, addr, mask);
                     $fatal();
                  end
               end else begin
                  if (b_resp != 2'h0) begin
                     $display("3 got a response of 0x%01x for first_word strobes of 0b:%04b and bottom addr bits 0x:%01x.", b_resp, i, addr);
                  end
               end
            end
            repeat (2) @(posedge axi_clk);
            t_assert_gennum_reset();
            repeat (2) @(posedge axi_clk);
         end // for
      end // for



      t_assert_gennum_reset();

      test_subphase = "fail intermediate strobe assertions";
      for (int phase = 0; phase < 5; phase++) begin
      logic [4:0] strb_val;
         t_send_aw(.id(0), .addr(64'h20000000), .len(2), .size(3'h2), .burst(2'h1), .cache(4'h0));
         strb_val = 5'h1f;
         strb_val[phase] = 1'b0;
         t_send_w(.last( 1'b0), .strb(4'hf), .data($urandom()));
         t_send_w(.last( 1'b0), .strb(strb_val[0+:4]), .data($urandom()));
         t_send_w(.last( 1'b1), .strb(4'hf), .data($urandom()));
         t_get_b(.resp(b_resp), .id(b_id));

         if (strb_val[3:0] == 4'hf) begin
             if (b_resp != 2'h0) begin
                 $display(" got a response of 0x%01x for strobes of 0b:%05b in the middle of a burst ..", b_resp, strb_val);
                 $fatal();
             end;
         end else begin
             if (b_resp != 2'h2) begin
               $display(" got a response of 0x%01x for strobes of 0b:%05b in the middle of a burst .", b_resp, strb_val);
               $fatal();
             end;
         end
      end


      t_assert_gennum_reset();


      test_subphase = "last word strobe assertions";
      for (int i = 0; i < 16; i++) begin
         t_send_aw(.id(0), .addr(64'h20000000 ), .len(2), .size(3'h2), .burst(2'h1), .cache(4'h0));
         t_send_w(.last( 1'b0), .strb(4'hf), .data($urandom()));
         t_send_w(.last( 1'b0), .strb(4'hf), .data($urandom()));
         t_send_w(.last( 1'b1), .strb(i[0+:4]), .data($urandom()));
         t_get_b(.resp(b_resp), .id(b_id));
         if (   (i[0+:4] == 4'b1111)
             || (i[0+:4] == 4'b0111)
             || (i[0+:4] == 4'b0011)
             || (i[0+:4] == 4'b0001)
            ) begin
            if (b_resp != 2'h0) begin
                $display(" got a response of 0x%01x for last word strobes of 0b:%04b ..", b_resp, i);
                $fatal();
            end;
         end else begin
            if (b_resp != 2'h2) begin
                $display(" got a response of 0x%01x for last word strobes of 0b:%04b.", b_resp, i);
                $fatal();
            end;
         end
      end


      t_assert_gennum_reset();


      test_subphase = "try different burst values";
      for (int i = 0; i < 4; i++) begin
            t_send_aw(.id(0), .addr(64'h20000000), .len(0), .size(3'h2), .burst(i[1:0]), .cache(4'h0));
            t_send_w(.last( 1'b1), .strb(4'hf), .data($urandom()));
            t_get_b(.resp(b_resp), .id(b_id));
            if ((2'h0 == b_resp) && (2'h1 == i)) begin
               //$display("Tried sending a burst value of 0d:%0d, got a response 0x:%01x ..",  i, b_resp);
            end else begin
              if (2'h2 !== b_resp) begin
                 $display("Tried sending a burst value of 0d:%0d, got a response 0x:%01x .",  i, b_resp);
                 $fatal();
              end
            end

      end

      t_assert_gennum_reset();

      test_subphase = "try different size values";
      for (int i = 0; i < 8; i++) begin
            t_send_aw(.id(0), .addr(64'h20000000), .len(0), .size(i), .burst(2'h1), .cache(4'h0));
            t_send_w(.last( 1'b1), .strb(4'hf), .data($urandom()));
            t_get_b(.resp(b_resp), .id(b_id));
            if ((2'h0 == b_resp) && (2 == i)) begin
              //$display("Tried sending a size value of 0d:%0d, got a response 0x:%01x ..",  i, b_resp);
            end else begin
              if (2'h2 !== b_resp) begin
                     $display("Tried sending a size value of 0d:%0d, got a response 0x:%01x esponse...",  i, b_resp);
                 $fatal();
              end
            end
      end

      t_assert_gennum_reset();

      test_subphase = "try different cache values";
      for (int i = 0; i < 16; i++) begin
            t_send_aw(.id(0), .addr(64'h20000000), .len(0), .size(3'h2), .burst(2'h1), .cache(i));
            t_send_w(.last( 1'b1), .strb(4'hf), .data($urandom()));
            t_get_b(.resp(b_resp), .id(b_id));
            if (   (i[0+:4] == 4'b0100)
                || (i[0+:4] == 4'b0101)
                || (i[0+:4] == 4'b1000)
                || (i[0+:4] == 4'b1001)
                || (i[0+:4] == 4'b1100)
                || (i[0+:4] == 4'b1101)
               ) begin
               if (b_resp != 2'h2) begin
                   $display("Tried sending a cache value of 0d:%0d, got a response 0x:%01x ..",  i, b_resp);
                   $fatal();
               end;
            end else begin
               if (b_resp != 2'h0) begin
                  $display("Tried sending a cache value of 0d:%0d, got a response 0x:%01x ..",  i, b_resp);
                   $fatal();
               end;
            end
      end


      t_assert_gennum_reset();

      test_subphase = "bad size various lengths, checking discards";
      // here I am using size to force a discard, and varying length
      for (int i = 0; i < 8; i++) begin
            t_send_aw(.id(0), .addr(64'h20000000), .len(i), .size(3'h7), .burst(2'h1), .cache(4'h0));
            for (int j = 0;  j <= i; j++) begin
               if (j == i) begin
                  t_send_w(.last( 1'b1), .strb(4'hf), .data($urandom()));
               end else begin
                  t_send_w(.last( 1'b0), .strb(4'hf), .data($urandom()));
               end
            end
            t_get_b(.resp(b_resp), .id(b_id));
            if (2'h2 !== b_resp) begin
                     $display("Tried sending a length value of 0d:%0d, got a response 0x:%01x ..",  i, b_resp);
                 $fatal();
            end
      end


      t_assert_gennum_reset();

      // then finally put through a good write!
      t_subtest_axi_wr_extra_wr_dat(.axi_addr(64'h20000000), .id(2'h1),
                                    .axi_len(5), .extra_words(0), .expected_b_resp(2'h0)
                                   );

      t_assert_gennum_reset();

      test_subphase = "try bad_strobes_longer_writes";
      for (int i = 0; i < 3; i++) begin
        t_send_aw(.id(0), .addr(64'h20000000), .len(2), .size(3'h2), .burst(2'h1), .cache(4'hf));
        t_send_w(.last( 1'b0), .strb(4'h1 << i), .data($urandom()));
        t_send_w(.last( 1'b0), .strb(4'hf), .data($urandom()));
        t_send_w(.last( 1'b1), .strb(4'hf), .data($urandom()));
        t_get_b(.resp(b_resp), .id(b_id));
        if (b_resp != 2'h2) begin
           $display("Tried sending a bad final strobed on a 3 word write, got a response 0x:%01x ..",  b_resp);
           $fatal();
        end;
      end

      t_assert_gennum_reset();

     // TODO: add a more thorough test for QWORD.... Strobe behaviours!!!
      test_subphase = "try a legal aligned qword write";
      t_send_aw(.id(0), .addr(64'h20000000), .len(1), .size(3'h2), .burst(2'h1), .cache(4'hf));
      t_send_w(.last( 1'b0), .strb(4'ha), .data($urandom()));
      t_send_w(.last( 1'b1), .strb(4'ha), .data($urandom()));
      t_get_b(.resp(b_resp), .id(b_id));
      if (b_resp != 2'h0) begin
         $display("Tried sending an aligned qword write, got a response 0x:%01x ..", b_resp);
         $fatal();
      end

      t_assert_gennum_reset();

      test_subphase = "try a illlegal un-aligned qword write";
      t_send_aw(.id(0), .addr(64'h20000004), .len(1), .size(3'h1), .burst(2'h1), .cache(4'hf));
      t_send_w(.last( 1'b0), .strb(4'h4), .data($urandom()));
      t_send_w(.last( 1'b1), .strb(4'hf), .data($urandom()));
      t_get_b(.resp(b_resp), .id(b_id));
      if (b_resp != 2'h2) begin
         $display("Tried sending an illlegal unaligned qword write, got a response 0x:%01x ..",  b_resp);
         $fatal();
      end


      test_subphase = "finishing up";
      #1us;

     $write("PASS\n");

endtask




// -------------------------------------------
//  TESTCASE:    t_axi_r_error_tests
//      Purpose:
//           BASic testing ofor AXI4 error conditions...
//      Method:
//          NOTES: to cover:
//                             bad command types
//                             bad size
//                             illegal cache values
// --------------------------------------------
task automatic t_axi_r_error_tests(ref   int ntest,
                                  input int tests);


     logic [1:0] id;
     logic [1:0] exp_resp;
     logic [63:0] addr;
     logic [7:0] len;
     logic [3:0] t_cache;
     logic [1:0] t_burst;
     logic [2:0] t_size;


      testphase = "Directed tests AXI4 error (read): ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);

      id       = 2'h0;
      addr     = 64'h2000_0000;
      len      = 8'h1;




      for (int  lp_count = 0; lp_count < 16 *8 *4 ; lp_count++) begin
            t_cache = lp_count[5+:4];
            t_burst = lp_count[3+:2];
            t_size  = lp_count[0+:3];

            exp_resp = 2'b00;
            if (t_burst != 2'h1) begin
               exp_resp = 2'h2;
            end
            if (3'h2 != t_size[2:0]) begin
               exp_resp = 2'h2;
            end
            if (   (t_cache == 4'h4)
                || (t_cache == 4'h5)
                || (t_cache == 4'h8)
                || (t_cache == 4'h9)
                || (t_cache == 4'hc)
                || (t_cache == 4'hd)
               ) begin
               exp_resp = 2'h2;
            end

            t_send_ar(         .id(id), .addr(addr), .len(len), .size(t_size), .burst(t_burst), .cache(t_cache));
            t_check_axi_rd_Dat(.id(id), .addr(addr), .len(len), .timeout(250), .expected_resp(exp_resp), .check_data(1'b0));

            t_assert_gennum_reset();
            @(posedge axi_clk);

      end
      test_subphase = "finishing up";
      #1us;

     $write("PASS\n");

endtask


// -------------------------------------------
//  TESTCASE:    t_all_interfaces_active
//      Purpose:
//           check concurrent AXI_RD, AXI_WR, WB_DMA and WB_MASTER accesses
//      Method:
//            TODO: is it sensible to make it self-check data integrity? or is this beter done in real HW ?
//                  make sure AXI Writes are posted, rather than waiting for completion   MORE PErformance
// --------------------------------------------
task automatic t_all_interfaces_active(ref   int ntest,
                                       input int tests);


   typedef struct {
      logic [ 1:0] id;
      logic [63:0] start_addr;
      logic  [7:0] len;
   }t_xaction;


   t_xaction axi_rd_xaction_q [$];
   t_xaction axi_rd_checking_q [$];
   t_xaction test_create;

   bit read_commands_all_sent;

   int posted_rd_cmds;

      bit axi_rds_done;
      bit axi_writes_done;
      bit wbm_done;

      testphase = "All interfaces!!! : ";
      $write("Test %0d/%0d: %s", ntest++, tests, testphase);



      if (dma_irq != 1'b0) begin
        $fatal(1, "dma irq should be 0");
      end


      // the chain is a perpetual loop, and we need to abort the transfer....
      t_add_chaining_descriptor(.host_addr                    ('h04_0000_0000),
                                .local_addr                   ('h00),
                                .length                       ('h100),
                                .more_in_chain                (1'b1),
                                .w_nrd_sel                    (1'b0),
                                .next_chained_descr_addr      ('h04_0002_0020),
                                .start_address_in_host_memory ('h04_0002_0000));

      t_add_chaining_descriptor(.host_addr                    ('h04_0000_0000),
                                .local_addr                   ('h00),
                                .length                       ('h100),
                                .more_in_chain                (1'b1),
                                .w_nrd_sel                    (1'b1),
                                .next_chained_descr_addr      ('h04_0002_0000),
                                .start_address_in_host_memory ('h04_0002_0020));


        t_setup_dma(.host_addr               ('h2000_0000),
                    .local_addr              ('h00),
                    .length                  ('h300),
                    .more_in_chain           (1),
                    .w_nrd_sel               (0),
                    .next_chained_descr_addr ('h04_0002_0000));



   for (integer i = 0; i<15; i++) begin
     test_create.id = $urandom_range(0,3);
     test_create.len = $urandom_range(0,255);

     // pick an addr in bar 0
     test_create.start_addr = 64'h2000_0000 + 4*$urandom_range(0,150);

     axi_rd_xaction_q.push_back(test_create);
   end

   read_commands_all_sent = 1'b0;
   posted_rd_cmds = 0;

   axi_rds_done    = 1'b0;
   axi_writes_done = 1'b0;
   wbm_done        = 1'b0;

   t_start_dma();

      fork
         // AXI_RDS
         begin
              fork
                 begin
                    t_xaction test_cmd;
                    while (axi_rd_xaction_q.size() !== 0) begin
                       test_cmd = axi_rd_xaction_q.pop_front();
                       t_send_ar(.id(   test_cmd.id),
                                 .addr( test_cmd.start_addr),
                                 .len(  test_cmd.len),
                                 .size( 3'h2),
                                 .burst(2'h1),
                                 .cache(4'h0)
                                );
                       axi_rd_checking_q.push_back(test_cmd);
                       posted_rd_cmds++;
                   //    $display("issued cmd: id 0x:%01x, addr:0x%016x, len 0x%02x", test_cmd.id, test_cmd.start_addr, test_cmd.len);
                    end
                    read_commands_all_sent = 1'b1;
                   // $display("have now posted %d commands", posted_cmds);
                 end

                 begin
                    t_xaction check_cmd;
                    while ((read_commands_all_sent == 1'b0) || (axi_rd_checking_q.size() != 0)) begin
                       if (axi_rd_checking_q.size() == 0) begin
                          @(posedge axi_clk);
                       end else begin
                          check_cmd = axi_rd_checking_q.pop_front();
                       //   $display("checking cmd: id 0x:%01x, addr:0x%016x, len 0x%02x", check_cmd.id, check_cmd.start_addr, check_cmd.len);
                          t_check_axi_rd_Dat(.id(     check_cmd.id),
                                             .addr(   check_cmd.start_addr),
                                             .len(    check_cmd.len),
                                             .timeout(1000 + 4*check_cmd.len),
                                             .check_data(1'b0)
                                            );
                       //   $display("Collected rd data resp %d", received_read_pkts);
                        //  received_read_pkts++;
                       end
                    end
                 end
              join
          // $display("axi reads completed");
           axi_rds_done = 1'b1;
         end


         // AXI_WRITES   : todo: improve this; it doesn't really hammer things...
         begin

            for (int i = 0; i < 20; i++) begin
               int len;
               int id;
               len = $urandom_range(0, 255);
               id  = $urandom_range(0, 3);
                t_subtest_axi_wr_extra_wr_dat(.axi_addr(64'h20000000), .id(id),
                                              .axi_len(len), .extra_words(0), .expected_b_resp(2'h0),
                                              .do_data_check(1'b0)
                                             );
            end
          //  $display("axi writes completed");
            axi_writes_done = 1'b1;
         end


         //wb_dma
         begin
            while ((axi_rds_done == 1'b0) | (0 == axi_writes_done) | (1'b0 == wbm_done)) begin
            #1us;
            end

            // send an WB_DMA abort
            acc.write('h00, 'h02);
            #1us; // give time for it to finish gracefully

            reg_check('h04, 'h03); // // check resp code!
            acc.write('h00, 'h00);
         end



         //WB_Host , add traffic to load it up...
         begin
            uint64_t rdat;
            #500ns;
            rdat = 'hffacce00;
            for (int i = 0; i < 50; i++) begin

               acc.write('h1c,  rdat);
               acc.read('h1c,   rdat);
               rdat[16+:16] = 'hdead;
               acc.write('h1c, rdat);
               acc.read('h1c,   rdat);
               rdat++;
               rdat[16+:16] = 'hffac;
            end
          //  $display("wishbone_host_ops_done");
            wbm_done = 1'b1;
         end
      join


     $write("PASS\n");


endtask



// -------------------------------------------
//  TESTCASE:    t_test_axi_RD_backpressure
//      Purpose:
//           check the systemn works when AXI R and B channel excerpt bacbkpressure...
//      Method:
//           WORK IN PROGRESS!!!!
// --------------------------------------------
task automatic t_test_axi_RD_backpressure(ref   int ntest,
                                          input int tests);

   uint64_t  BAR_0_ADDR;
   integer   received_rd_last_assertions;

   BAR_0_ADDR = 64'h2000_0000;


   testphase = "t_test_axi_RD_backpressure: ";
   $write("Test %0d/%0d: %s", ntest++, tests, testphase);

   test_subphase = "Change AXI_Backpressure values";
   axi4b_backpressure_cmp = 0;
   axi4r_backpressure_cmp = 0;

   repeat (20) @(posedge axi_clk);

   test_subphase = "Post enough reads to fill the read buffer and make sure that we see backpressure on AR_RDY";


// the first reads will leave data in the read response buffer, that should allow me to wait for the next cmd...
   t_send_ar(.id(0),
             .addr(BAR_0_ADDR),
             .len(8'hff),
             .size(3'h2),
             .burst(2'h1),
             .cache(4'h0));

   t_send_ar(.id(1),
             .addr(BAR_0_ADDR),
             .len(8'h80),
             .size(3'h2),
             .burst(2'h1),
             .cache(4'h0));

   t_send_ar(.id(2),
             .addr(BAR_0_ADDR),
             .len(8'hf3),
             .size(3'h2),
             .burst(2'h1),
             .cache(4'h0));

   test_subphase = "Push lots of small read requests for fill up the preprog -> initiator fifo";
   for (int i = 0; i < 16; i++) begin
   //for (int i = 0; i < 17; i++) begin
      t_send_ar(.id(i + 3),
                .addr(BAR_0_ADDR),
                .len(i[7:0]),
                .size(3'h2),
                .burst(2'h1),
                .cache(4'h0));
   end



   test_subphase = "Check the AR interface is still able to accept a command ";
   repeat (500) @(posedge axi_clk);    // make sure there is enought time to forward the request_
   repeat (400) begin
      assert (dut_if_ar.ready == 1'b1) else begin
         $display("unexpected status"); $fatal();
      end
      @(posedge axi_clk);
   end


   test_subphase = "Push one more AR_CMD then check that ar_Rdy is ZERO";
   repeat (10) @(posedge axi_clk);
      t_send_ar(.id('hffff),
                .addr(BAR_0_ADDR),
                .len(2),
                .size(3'h2),
                .burst(2'h1),
                .cache(4'h0));
   @(posedge axi_clk);
   repeat (200) begin
      assert (dut_if_ar.ready == 1'b0) else begin
         $display("unexpected status"); $fatal();
      end
      @(posedge axi_clk);
   end


   //let things drain down and check everything is OK...
   test_subphase = "Leave things for a little bit...   (inspect)";
   repeat (100) @(posedge axi_clk);



   test_subphase = "Change AXI_Backpressure values, and count back in packets..";
   repeat (5) @(posedge axi_clk);
   axi4b_backpressure_cmp = 9;
   axi4r_backpressure_cmp = 9;

   received_rd_last_assertions = 0;
   //while (received_rd_last_assertions < 21) begin
   while (received_rd_last_assertions < 20) begin
      bit saw_a_last = 1'b0;
      int cycle_count = 0;
      while (saw_a_last == 1'b0) begin
         @(posedge axi_clk);
         if ((1'b1 == dut_if_r.ready) & (1'b1 == dut_if_r.valid) & (1'b1 == dut_if_r.last) ) begin
           saw_a_last = 1'b1;
         end
         cycle_count++;
         if (cycle_count > 1000) begin
             $display("timeout"); $fatal();
         end
      end
      received_rd_last_assertions++;
     // $display("got %02d AXI_R_CHANNEL packet Last signalled", received_rd_last_assertions);
   end



   test_subphase = "got enough read responses, timout then next test";
   repeat (600)  @(posedge axi_clk);
   axi4b_backpressure_cmp = 5;
   axi4r_backpressure_cmp = 5;

   test_subphase = "done ";
   $write("PASS\n");
   @(posedge axi_clk);
endtask
